#!/usr/bin/env chai
;;;; -*- mode: chai -*-

;; _prelude_init.ch -- first file loaded by prelude.ch
;; October 1, 2008
;;
;; Copyright (C) 2007-2008 Devin Papineau
;;
;; This software is provided 'as-is', without any express or implied
;; warranty.  In no event will the authors be held liable for any damages
;; arising from the use of this software.
;;
;; Permission is granted to anyone to use this software for any purpose,
;; including commercial applications, and to alter it and redistribute it
;; freely, subject to the following restrictions:
;;
;; 1. The origin of this software must not be misrepresented; you must not
;;    claim that you wrote the original software. If you use this software
;;    in a product, an acknowledgment in the product documentation would be
;;    appreciated but is not required.
;; 2. Altered source versions must be plainly marked as such, and must not be
;;    misrepresented as being the original software.
;; 3. This notice may not be removed or altered from any source distribution.
;;
;; Devin Papineau jdmpapin@uwaterloo.ca

;;; Concatenation operator
defop: "++", "+"

defmacro: (#|op_++:| a, b) {
  [#send, a, #concat, b]
}
;;; ----


oncompile: { List.define_method(#map, nil) }


;;; Code quoting
defuprec: 0
defuop: "~~", 0
defuop: "~.", "@"
defuop: "~*", "~."

oncompile: {
    #|expand_~~| = fn: expr {
        if: expr.is_a?(Symbol) {
            [#quote, expr]
        } :: not expr.is_a?(List) {
            expr
        } :: expr(0) == #"uop_~." {
            expr(1)
        } :: else {
            my map = []
            my i = 0
            while: i < expr.size {
                my x = expr(i)
                if: x.is_a?(List) and x(0) == #"uop_~*" {
                    map.push(x(1))
                } :: else {
                    map.push([#list, #|expand_~~|(x)])
                }
                i = i + 1
            }

            my a = [#list]
            i = 0
            while: i < map.size {
                my b = map(i)
                a = [#send, a, #concat, b]
                i = i + 1
            }
            a
        }
    }


    ;; def[] will do this for you later
    (slot: #|expand_~~|, name) = #"expand_~~"
}

defmacro: (#|uop_~~:| expr) { #|expand_~~|(expr) }
;;; ----


;;; Make raise an operator
defmacro: (uop_raise: e) {
    ~~ (raise: ~.e)
}

defuop: "raise", "my"
;;; ----


;;; Python-style iteration protocol
oncompile: {
    StopIterationException = Class(Exception, #StopIterationException)

    List.define_method(#iter, fn: self {
        my i = -1
        my size = self.size

        fn: {
            i = i + 1
            if: i >= size {
                raise StopIterationException()
            }
            self(i)
        }
    })

    Range.define_method(#iter, fn: self {
        my i = (slot: self, lo)
        my end = (slot: self, hi)

        fn: {
            if: i > end {
                raise StopIterationException()
            }

            my result = i
            i = i.succ
            result
        }
    })
}
;;; ----


;;; For-each loops
defop: "in", "=="  ;; we define the expansion later

defmacro: (for: *args) {
    if: args.size < 2 { raise ParseError("too few args to for{}") }

    my body = args.pop
    if: args.size > 1 {
        my first = args.shift
        ~~ for: ~.first {
            for: ~*args { ~.body }
        }
    } :: else {
        my cond = args(0)
        if: !cond.is_a?(List) or cond(0) != #op_in or cond.size != 3 {
            raise ParseError("bad for clause")
        }

        my x = cond(1)
        my seq = cond(2)
        my o = gensym()
        my iter = gensym()

        ~~ let: ~.o, ~.iter = ~.seq.iter {
            while: (try: {~.o = ~.iter(); true} :: StopIterationException {}) {
                my ~.x = ~.o
                ~.body
            }
        }
    }
}
;;; ----


;;; Macro for defining functions
defmacro: (def: call, body) {
    if: not call.is_a?(List) or call(0) != #call {
        raise ParseError("first argument to def{} is not a function call")
    }

    my args = call.dup
    args.shift ;; lose the #call
    my name = args.shift

    ~~ progn: {
        ~.name = fn: ~*args { ~.body }
        (slot: ~.name, name) = (quote: ~.name)
        (slot: ~.name, containing_module) = (current_module:)
        nil
    }
}
;;; ----


;;; Macro for defining classes
defmacro: (class: name, super, body = nil) {
    if: not body {
        body = super
        super = #Object
    }

    ~~ progn: {
        try: {
            ~.name
        } :: NameError {
            ~.name = Class(~.super, quote: ~.name)
            (slot: ~.name, containing_module) = (current_module:)
        }

        let: self = ~.name { def_means_method: { ~.body } }
    }
}

;;; This takes any "toplevel" (under at most a progn[]) def[] forms
;;; and turns them into _defm[] forms.
defmacro: (def_means_method: body) {
    my replace_def = fn: expr {
        if: expr.is_a?(List) and expr(0) == #def {
            my args = expr.dup
            args.shift
            ~~ _defm: ~*args
        } :: else {
            expr
        }
    }

    if: body.is_a?(List) and body(0) == #progn {
        my replaced = []
        for: bexpr in body { replaced.push(replace_def(bexpr)) }
        replaced
    } :: else {
        replace_def(body)
    }
}
;;; ---

;;; Macro for defining methods. Not for explicit use.
defmacro: (_defm: call, body) {
    my method_name
    my args

    if: call.is_a?(Symbol) {
        method_name = call
        args = []
    } :: not call.is_a?(List) or call(0) != #call or call.size <= 1 {
        raise ParseError("_defm{} expects a function call or identifier")
    } :: else {
        args = call.dup
        args.shift ;; lose #call
        method_name = args.shift
    }

    ~~ self.define_method((quote: ~.method_name), fn: self, ~*args { ~.body })
}
;;; ----


;;; Define Range#map so that some of this syntax can start to work.
;;; This gets redefined later (from Iterable), but that's OK.
oncompile: {
    Range.define_method(#map, fn: self, f {
        my result = []
        for: x in self { result.push(f(x)) }
        result
    })
}
;;; ----


;;; Unary dot: .x should be the same as self.x
defmacro: (#|uop_.:| method_call) {
    ~~ self . ~.method_call
}

;;; In calls
defcall: #|uop_.|, f, args {
    ~~ self . ~.(f(1)) (~*args)
}

;;; In setf
defsetf: #|uop_.|, lval, rval {
    ~~ self . ~.(lval(1)) = ~.rval
}
;;; ----


;;; Unary at sign: @x becomes slot [self, x]
defmacro: (#|uop_@:| member) {
    ~~ slot: self, ~.member
}

;;; In setf
defsetf: #|uop_@|, lval, rval {
    ~~ (slot: self, ~.(lval(1))) = ~.rval
}
;;; ----


;;; Explicit returns
class: ReturnException, Exception { }
class: AlreadyReturnedException, Exception { }

defmacro: (leave: return, body = nil) {
    if: not body {
        body = return
        return = #return
    }

    my ret_exc = gensym()
    my ret_val = gensym()
    my ret_fn = gensym()

    ~~ let: ~.ret_val, ~.ret_exc = ReturnException() {
        let: (~.ret_fn = fn: val = nil {
                 ~.ret_val = val
                 raise ~.ret_exc
             }),
             (~.return = fn: val = nil { ~.ret_fn(val) })
        {
            try: {
                ~.ret_val = ~.body
            } :: ReturnException => e {
                if: ~.ret_exc != e { raise e }
            } :: ensure {
                ~.ret_fn = fn: val = nil { raise AlreadyReturnedException() }
                ~.ret_val
            }
        }
    }
}
;;; ----


;;; fnot: not on the function level
def: fnot(f) {
    fn: *args { !f.apply(args) }
}
;;; ----


;;; A class to avoid defining a class every time I want to return
;;; "something iterable where #iter is this code."
class: FnIter {
    def: _init_(f) { @f = f }
    def: iter { @f() }

    def: inspect {
        let: lex, name, klass, mod {
            lex = slot: @f, lexical_parent
            name = if: lex { slot: lex, name } :: { "[orphaned]" }

            klass = lex && (slot: lex, method_of)
            if: klass { name = klass.inspect ++ "#" ++ name }

            mod = lex && (slot: klass || lex, containing_module)
            if: mod and mod != (current_module:) {
                name = mod.to_s ++ ":" ++ name
            }

            "#<" ++ name ++ " iterator>"
        }
    }

    def: pretty_print(indent = 0, limit = 80) {
        " ".repeat(indent) ++ .inspect
    }
}

;;; Macro to look nicer
defmacro: (iter: body) {
    ~~ FnIter(fn: { ~.body })
}
;;; ----


;;; Mixin for iterable things, expects subscribing classes to implement
;;; the iteration protocol
;;; ----


;;; These classes can be iterated over...
;;for: klass in [List, Range, FnIter] { Iterable.subscribe(klass) }
;;; ----


;;; Redefine List#concat! so that it can take any iterable
;;; This allows me to make it use Iterable#map without it dying
;;; on the resulting FnIter object.
oncompile: {
    class: List {
        my old_concat! = .method_implementation_unsafe(#concat!)

        def: concat!(xs) {
            if: xs.is_a?(List) {
                old_concat!(self, xs)
            } :: else {
                for: x in xs { .push(x) }
            }
        }
    }
}
;;; ----


;;; Now that forcing List to use Iterable#map won't cause things to
;;; break...
;;for: klass in [List, Range] {
;;    Iterable.force_method(klass, #map)
;;    Iterable.force_method(klass, #reduce)
;;}
;;; ----


;;; (x in xs) expands to xs.include?(x)
defmacro: (op_in: x, xs) {
    ~~ (~.xs.include?(~.x))
}
;;; ----


;;; Syntax for dictionaries
defmacro: (d: *pairs) {
    my d = gensym()
    my assigns = []

    for: pair in pairs {
        if: !pair.is_a?(List) or pair(0) != #"op_=>" {
            raise ParseError("d[] expects only => pairs")
        }

        my k = pair(1)
        my v = pair(2)
        assigns.push(~~ (~.d(~.k) = ~.v))
    }

    ~~ let: ~.d = Dict() {
        ~*assigns
        ~.d
    }
}

;;; Can't expand =>, that's silly
defmacro: (#|op_=>:| a, b) {
    raise ParseError("stray => pair operator")
}
;;; ----


;;; Define some Dict methods
class: Dict {
    ;; inspect is not overridden in an intelligent way in C
    def: inspect_raw {
        my result = "d["
        my first = true

        for: k in .keys {
            if: first { first = false } :: { result.concat!(", ") }
            result.concat!(k.inspect ++ " => " ++ self(k).inspect)
        }
        result.concat!("]")
        result
    }
}
;;; ----


;;; Sequence unpacking in assignment can have errors
oncompile: {
    class: SequenceUnpackException, Exception { }
}

;;; This is how we assign to lists
defsetf: list, lval, rval {
    my lvals = lval.dup
    lvals.shift

    my seq = gensym()
    my next_fn = gensym()

    my assignments = []
    for: lval in lvals { assignments.push(~~ (~.lval = ~.next_fn())) }

    ~~ let: ~.next_fn, ~.seq = ~.rval {
        try: {
            ~.next_fn = ~.seq.iter
        } :: NoMethodError => e {
            raise SequenceUnpackException(e.message)
        }

        try: {
            progn: { ~*assignments }
        } :: StopIterationException {
            raise SequenceUnpackException("sequence too short: " ++ ~.seq.inspect)
        }

        ~.seq
    }
}

;;; my [x, [y,z]] binds x, y and z
defmybindings: list, lvals {
    lvals = lvals.dup
    lvals.shift

    my result = []
    for: l in lvals { result = result ++ my_bindings_for(l) }
    result
}
;;; ----


;;; In order to implement +=, -=, etc., without evaluating possibly
;;; side-effecting subexpressions (eg. something_destructive().x += 3),
;;; we need a way to extract the subexpressions and pre-evaluate them.
oncompile: {
    subexpression_functions = d[]

    def: extract_subexprs(expr) {
        if: not expr.is_a?(List) {
            [[], expr]
        } :: else {
            my f = subexpression_functions(expr(0))
            if: not f {
                raise ParseError("no subexpr function for " ++ expr(0).inspect)
            }
            f(expr)
        }
    }
}

;;; Allow the definition for what counts as a subexpression for a
;;; given form
defmacro: (defsubexprs: form_name, var, body) {
    ~~ oncompile: {
        subexpression_functions(quote: ~.form_name) = fn: ~.var { ~.body }
    }
}

;;; Utility function that makes defsubexprs[] declarations really short
;;; in general.
oncompile: {
    def: extract_subexprs_with_mask(expr, mask) {
        my assignments = []
        my out_expr = []

        for: i in (0..expr.size - 1) {
            if: mask(i) {
                my tmpnam = gensym()
                assignments.push(~~ (my ~.tmpnam = ~.(expr(i))))
                out_expr.push(tmpnam)
            } :: else {
                out_expr.push(expr(i))
            }
        }

        [assignments.reverse, out_expr]
    }
}
;;; ----


;;;; Take out the subexpressions of call[]
defsubexprs: call, expr {
    if: defcalled?(expr(1)) {
        extract_subexpr(macroexpand1(expr))
    } :: else {
        my mask = [false] ++ List(expr.size-1, fn: _ { true })
        extract_subexprs_with_mask(expr, mask)
    }
}

;;; Take out the subexpressions of send[]
defsubexprs: send, expr {
    my mask = [false, true, false] ++ List(expr.size-3, fn: _ { true })
    extract_subexprs_with_mask(expr, mask)
}
;;; ----


;;; Shortcut assignment operators like +=, -=, etc.
oncompile: {
    def: expand_shortcut_assignment(op, lval, rval) {
        my [subexpr_assigns, extracted] = extract_subexprs(lval)
        my opsym = ("op_" ++ op).intern

        ~~ progn: {
            ~*subexpr_assigns
            ~.extracted = ~.([opsym, extracted, rval])
        }
    }
}

;; For addition
defop: "+=", "="
defmacro: (#|op_+=:| x,y) { expand_shortcut_assignment("+",x,y) }

;; For subtraction
defop: "-=", "="
defmacro: (#|op_-=:| x,y) { expand_shortcut_assignment("-",x,y) }

;; Multiplication
defop: "*=", "="
defmacro: (#|op_*=:| x,y) { expand_shortcut_assignment("*",x,y) }

;; Division
defop: "/=", "="
defmacro: (#|op_/=:| x,y) { expand_shortcut_assignment("/",x,y) }

;; Integer division
defop: "//=", "="
defmacro: (#|op_//=:| x,y) { expand_shortcut_assignment("//",x,y) }

;; Modulo
defop: "%=", "="
defmacro: (#|op_%=:| x,y) { expand_shortcut_assignment("%",x,y) }

;; Power
defop: "**=", "="
defmacro: (#|op_**=:| x,y) { expand_shortcut_assignment("**",x,y) }

;; Bitwise xor
defop: "^=", "="
defmacro: (#|op_^=:| x,y) { expand_shortcut_assignment("^",x,y) }

;; Bitwise and
defop: "&=", "="
defmacro: (#|op_&=:| x,y) { expand_shortcut_assignment("&",x,y) }

;; Bitwise or
defop: "|=", "="
defmacro: (#|op_\|=:| x,y) { expand_shortcut_assignment("|",x,y) }

;; Shift left
defop: "<<=", "="
defmacro: (#|op_<<=:| x,y) { expand_shortcut_assignment("<<",x,y) }

;; Shift right
defop: ">>=", "="
defmacro: (#|op_>>=:| x,y) { expand_shortcut_assignment(">>",x,y) }

;; Concatenate
defop: "++=", "="
defmacro: (#|op_++=:| x,y) { expand_shortcut_assignment("++",x,y) }
;;; ----


;;; Regular expression syntax
oncompile: {
    _regexen = d[]

    deftagstr: re, text { ~~ re: ~.text }

    defmacro: (re: text) {
        my sym = text.intern
        if: not _regexen.has_key?(sym) {
            _regexen(sym) = Regex(text)
        }
        ~~ _regexen(quote: ~.sym)
    }
}

;;; Regex methods
class: Regex {
    ;; #inspect should return something in that syntax
    def: inspect { "re\"" ++ .source ++ "\"" }

    ;; calling a regex should match; then you can treat it as a predicate
    def: _call_(s) { .match(s) }
}
;;; ----


;;; String methods
class: String {
    ;; iterate over the occurrences of matching text to a regex
    def: scan(re) {
        iter: {
            my i = 0
            fn: {
                my md
                if: i >= .size or !(md = re.match(self, i)) {
                    raise StopIterationException()
                } :: else {
                    i = md.match_indices(0)(1) + 1
                    md
                }
            }
        }
    }

    def: chars {
        iter: {
            my scan = .scan(re"[\000-\377]").iter
            fn: { scan().match }
        }
    }

    def: lines {
        iter: {
            my scan = .scan(re"^.*$(\r?\n)?").iter
            fn: { scan().match }
        }
    }

    ;; regex replacement
    def: gsub(re, with) {
        my [i, s, md] = [0, "", nil]
        if: with.is_a?(String) { with = with._to_replacement_func }

        while: i < .size and md = re.match(self, i) {
            my [start, end] = md.match_indices(0)
            if: start > 0 {
                s.concat!(.substr(i, start - 1))
            }

            s.concat!(with(md))
            if: end >= start {
                i = end + 1
            } :: else {
                ;; zero-width match
                s.concat!(self(start))
                i = start + 1
            }
        }

        s.concat!(.substr(i, -1))
        s
    }

    def: gsub!(re, with) {
        .become!(.gsub(re, with))
        self
    }

    ;; split on a regex (default whitespace)
    def: split(re = re"\s+") {
        my last_end = -1
        my result = []
        my md

        while: md = re.match(self, last_end + 1) {
            my [start, end] = md.match_indices(0)

            if: start != 0 {
                result.push(self.substr(last_end + 1, start - 1))
            }

            last_end = end
        }

        if: last_end < self.size - 1 {
            result.push(self.substr(last_end + 1, -1))
        }

        result
    }
}
;;; ----


;;; Range should inspect in a way that doesn't suck as well
class: Range {
    def: inspect_raw {
        @lo.inspect ++ ".." ++ @hi.inspect
    }
}
;;; ----


;;; nicer importing
defmacro: (import: *args) {
    if: args.empty? {
        raise ParseError("malformed import (import:)")
    }

    if: args(-1).is_a?(List) and args(-1)(0) == #call {
        ;; import: foo, bar (from: module)
        my [call_, name, [from_, module]] = args.pop
        if: from_ != #from {
            raise ParseError("malformed import - looks like import-from but isn't")
        }

        args.push(name)
        for: arg in args {
            if: not arg.is_a?(Symbol) {
                raise ParseError("malformed import-from - not all just names")
            }
        }

        if: not module.is_a?(Symbol) {
            raise ParseError("malformed import-from - non-identifier module")
        }

        ~~ _import_from: ~.module, ~*args
    } :: args(-1).is_a?(List) and args(-1)(0) == #"op_=" and
         args(-1)(2).is_a?(List) and args(-1)(2)(0) == #call
    {
        raise ParseError("malformed import-from - not all just names")
    } :: args.size == 1 and args(0).is_a?(List) and args(0)(0) == #"uop_*" {
        ;; import: * (from: module)
        my [star_, [from_, module]] = args(0)
        if: from_ != #from {
            raise ParseError("malformed import-all - got " ++
                             from_ ++ " instead of from")
        }

        if: not module.is_a?(Symbol) {
            raise ParseError("malformed import-all - non-identifier module")
        }

        ~~ _import_from: ~.module
    } :: else {
        ;; normal import
        module = args.shift

        if: not args.empty? {
            ~~ progn: {
                import: ~.module
                import: ~*args
            }
        } :: module.is_a?(Symbol) {
            ~~ import: ~.module = ~.module
        } :: not module.is_a?(List) or module(0) != #"op_=" {
            raise ParseError("malformed import, got " ++ module.inspect)
        } :: else {
            module = module.dup
            module.shift

            my [name, realname] = module

            if: not name.is_a?(Symbol) && realname.is_a?(Symbol) {
                raise ParseError("malformed import, got " ++ module.inspect)
            }

            ~~ _import: ~.realname, ~.name
        }
    }
}

defmacro: (_import_from: m, *names) {
    if: names.empty? {
        ~~ _import_all: ~.m
    } :: else {
        my each_one = []
        for: name in names {
            each_one.push(~~ (_import_name: ~.m, ~.name))
        }

        ~~ progn: { ~*each_one }
    }
}
;;; ----
