#!/usr/bin/env chai
;;; -*- mode: chai -*-

;; _prelude_init2.ch -- second file loaded by prelude.ch
;; October 1, 2008
;;
;; Copyright (C) 2007-2008 Devin Papineau
;;
;; This software is provided 'as-is', without any express or implied
;; warranty.  In no event will the authors be held liable for any damages
;; arising from the use of this software.
;;
;; Permission is granted to anyone to use this software for any purpose,
;; including commercial applications, and to alter it and redistribute it
;; freely, subject to the following restrictions:
;;
;; 1. The origin of this software must not be misrepresented; you must not
;;    claim that you wrote the original software. If you use this software
;;    in a product, an acknowledgment in the product documentation would be
;;    appreciated but is not required.
;; 2. Altered source versions must be plainly marked as such, and must not be
;;    misrepresented as being the original software.
;; 3. This notice may not be removed or altered from any source distribution.
;;
;; Devin Papineau jdmpapin@uwaterloo.ca

;;; "Assignment functions," stealing assignment's extensions
defmacro: (afn: *args) {
    if: args.size < 1 {
        raise ParseError("afn{} expects at least a body")
    }

    args = args.dup
    my body = args.pop
    my internal_names = []
    for: x in args { internal_names.push(if: x.is_a?(Symbol), x, gensym()) }

    my names_assigns = []
    for: i in 0..(args.size - 1) {
        my [arg, name] = [args(i), internal_names(i)]
        if: arg == name {
            names_assigns.push([name])
        } :: else {
            names_assigns.push([name, ~~ (my ~.arg = ~.name)])
        }
    }

    my formal_params = []
    for: x in names_assigns { formal_params.push(x(0)) }

    my stmts = []
    for: x in names_assigns {
        if: x.size != 1 { stmts.push(x(1)) }
    }

    ~~ fn: ~*formal_params {
        progn: {
            ~*stmts
            ~.body
        }
    }
}

;;; (x -> x*x) is the same as afn [x, x*x]
;;; (*[x,y,z] -> x*y*z) is the same as afn [x,y,z, x*y*z]
defrightprec: 450
defop: "->", 450

defmacro: (#|op_->:| arg, body) {
    if: arg.is_a?(List) and arg(0) == #"uop_*" and
        arg(1).is_a?(List) and arg(1)(0) == #list {
        my args = arg(1).dup
        args.shift
        ~~ afn: ~*args { ~.body }
    } :: else {
        ~~ afn: ~.arg { ~.body }
    }
}
;;; ----


;;; Useful methods for match data objects
class: RegexMatchData {
    def: match_indices(i) {
        try: {
            my i2 = i * 2
            [@offsets(i2), @offsets(i2 + 1) - 1]
        } :: OutOfBoundsError {
            raise OutOfBoundsError("index " ++ i.to_s ++ " out of bounds")
        }
    }

    def: _call_(i) {
        my [start, end] = .match_indices(i)
        @text.slice(start, end)
    }

    def: match { self(0) }
    def: pre_match { @text.slice(0, @offsets(0) - 1) }
    def: post_match { @text.slice(@offsets(1), -1) }
}
;;; ----


;;;
class: String {
    def: _to_replacement_func {
        my [last_end, f] = [-1, md -> ""]
        for: backref_md in .scan(re"\\(\d|&)") {
            my [backref_start, backref_end] = backref_md.match_indices(0)
            my text = if: backref_start == 0 { "" }
                      :: else { .substr(last_end + 1, backref_start - 1) }

            my index = if: backref_md(1) == "&", 0, backref_md(1).to_i

            [last_end, f] = [backref_end,
                             md -> let: s = f(md) {
                                 s.concat!(text)
                                 s.concat!(md(index))
                                 s
                             }]
        }


        my text = .substr(last_end + 1, .size - 1)
        fn: md {
            my s = f(md)
            s.concat!(text)
            s
        }
    }
}
;;; ----


;;; raw strings, act a bit like ruby's single-quoted ones
deftagstr: r, text {
    text.gsub(re"\\([\"\\])", "\\1")
}

;;; word arrays, like %w()
deftagstr: w, text {
    ~~ [~*(eval("\"" ++ text ++ "\"").split)]
}
;;; ----


;;; Pretty printing, and fixing inspection
_inspect_stack = []

for: klass in [Number, Symbol, Boolean, NilClass, File, Exception, Class] {
    class: klass {
        def: pretty_print(indent = 0, limit = 80) {
            " ".repeat(indent) ++ .inspect
        }
    }
}

class: String {
    def: pretty_print(indent = 0, limit = 80) {
        if: @pretty_string {
            " ".repeat(indent) ++ self
        } :: else {
            " ".repeat(indent) ++ .inspect
        }
    }
}

class: List {
    def: pretty_print_raw(indent = 0, limit = 80) {
        my spaces = " ".repeat(indent)
        my one_line = spaces ++ .inspect_up_one_level

        if: one_line.size <= limit or .empty? {
            one_line
        } :: else {
            my r = spaces ++ "[\n"

            my first = true
            for: x in self {
                if: first { first = false } :: { r.concat!(",\n") }
                r.concat!(x.pretty_print(indent + 2, limit))
            }

            r.concat!("\n" ++ spaces ++ "]")
            r
        }
    }
}

class: Dict {
    def: pretty_print_raw(indent = 0, limit = 80) {
        my spaces = " ".repeat(indent)
        my one_line = spaces ++ .inspect_up_one_level

        if: one_line.size <= limit or .empty? {
            one_line
        } :: else {
            my format_pair = fn: k {
                my ins = k.inspect
                my one_line = spaces ++ "  " ++ ins ++ " => " ++ self(k).inspect

                if: one_line.size <= limit {
                    one_line
                } :: else {
                    my r = k.pretty_print(indent + 2, limit).dup
                    r.concat!(" =>\n")
                    r.concat!(self(k).pretty_print(indent + 4, limit))
                    r
                }
            }

            my keys = .keys
            my r = spaces ++ "d[\n" ++ format_pair(keys.shift)
            for: k in keys { r.concat!(",\n" ++ format_pair(k)) }
            r.concat!("\n" ++ spaces ++ "]")
            r
        }
    }
}

class: Object {
    def: inspect {
        my found = false
        for: obj in _inspect_stack {
            if: self.eq?(obj) { found = true }
        }

        if: found {
            "#<...>"
        } :: else {
            _inspect_stack.push(self)
            my r = .inspect_raw
            _inspect_stack.pop
            r
        }
    }

    def: inspect_up_one_level {
        if: _inspect_stack.empty? {
            .inspect
        } :: else {
            my save = _inspect_stack.pop
            my r = .inspect
            _inspect_stack.push(save)
            r
        }
    }

    def: pretty_print(indent = 0, limit = 80) {
        my found = false
        for: obj in _inspect_stack { if: self.eq?(obj) { found = true } }

        if: found {
            " ".repeat(indent) ++ "#<...>"
        } :: else {
            _inspect_stack.push(self)
            my r = .pretty_print_raw(indent, limit)
            _inspect_stack.pop
            r
        }
    }

    def: pretty_print_raw(indent = 0, limit = 80) {
        my spaces = " ".repeat(indent)
        my one_line = spaces ++ .inspect_up_one_level

        if: one_line.size <= limit {
            one_line
        } :: else {
            my format_pair
            def: format_pair(k) {
                my v = (varslot: self, k)
                my one_line = spaces ++ "  " ++ k.to_s ++ "=" ++ v.inspect

                if: one_line.size <= limit {
                    one_line
                } :: else {
                    my r = spaces ++ "  " ++ k.to_s ++ "=\n"
                    r.concat!(v.pretty_print(indent + 4, limit))
                    r
                }
            }

            my md = re" 0x[a-fA-F0-9]+".match(one_line)
            my addy = if: md { md(0) } :: { "" }

            my r = spaces ++ "#<" ++ .class.inspect ++ addy
            for: var in .slots {
                r.concat!("\n")
                r.concat!(format_pair(var))
            }
            r.concat!("\n" ++ spaces ++ ">")
            r
        }
    }
}
;;;
