#!/usr/bin/env chai
;;;; -*- mode: chai -*-

;; sourcify.ch -- source code printer
;; September 24, 2008
;;
;; Copyright (C) 2007-2008 Devin Papineau
;;
;; This software is provided 'as-is', without any express or implied
;; warranty.  In no event will the authors be held liable for any damages
;; arising from the use of this software.
;;
;; Permission is granted to anyone to use this software for any purpose,
;; including commercial applications, and to alter it and redistribute it
;; freely, subject to the following restrictions:
;;
;; 1. The origin of this software must not be misrepresented; you must not
;;    claim that you wrote the original software. If you use this software
;;    in a product, an acknowledgment in the product documentation would be
;;    appreciated but is not required.
;; 2. Altered source versions must be plainly marked as such, and must not be
;;    misrepresented as being the original software.
;; 3. This notice may not be removed or altered from any source distribution.
;;
;; Devin Papineau jdmpapin@uwaterloo.ca

#|\nto_code_brace_rules| = Dict(_ -> fn: *args { false })
defmacro: (def_brace_rule: form, i, body) {
    ~~ (
        #|\nto_code_brace_rules|(quote: ~.form) = fn: ~.form, ~.i { ~.body }
    )
}

def_brace_rule: progn,    i  { true }
def_brace_rule: _progn,   i  { true }
def_brace_rule: if,       i  { i % 2 == 0 or i == if.size - 1 }
def_brace_rule: def,      i  { i > 1 }
def_brace_rule: defmacro, i  { i > 1 }
def_brace_rule: fn,       i  { i == fn.size - 1 }
def_brace_rule: for,      i  { i == for.size - 1 }
def_brace_rule: while,    i  { i == while.size - 1 }
def_brace_rule: class,    i  { i == class.size - 1 }
def_brace_rule: mixin,    i  { i == mixin.size - 1 }
def_brace_rule: try,      i  { i % 2 == 1 }
def_brace_rule: let,      i  { i == let.size - 1 }
def_brace_rule: whilet,   i  { i == whilet.size - 1 }
def_brace_rule: wrapif,   i  { i > 1 }
def_brace_rule: with_file,i  { i == with_file.size - 1 }
def_brace_rule: leave,    i  { i == leave.size - 1 }

def_brace_rule: def_means_method, i  { true }
def_brace_rule: unwind_protect,   i  { true }

#|\ncall_paren_exceptions| = d[]
defmacro: (def_call_paren_exc: name) {
    ~~ #|\ncall_paren_exceptions|(quote: ~.name) = true
}

def_call_paren_exc: call
def_call_paren_exc: list
def_call_paren_exc: d
def_call_paren_exc: quote

#|\nno_space_ops| = d[]
defmacro: (def_no_space_op: name) {
    ~~ #|\nno_space_ops|(~.(name.to_s)) = true
}

def_no_space_op: "."

;;; Convert a symbol (identifier) to source code.
def: symbol_to_code(sym, indent) {
    my s = sym.inspect

    if: re"^#\"".match(s) {
        "#|" ++ s.substr(2, -2) ++ "|"
    } :: else {
        sym.to_s
    }
}

;;; Convert a self-evaluating literal (number or string) to source code
def: selfeval_to_code(e, indent) { e.inspect }

;;; Local defs for list_to_code()
let: limit, op_need_parens?, arg_need_parens?, too_long?, emit_arglist {
    limit = 80

    def: op_need_parens?(mine, me_u, it) {
        my md
        if: not it.is_a?(List) {
            false
        } :: md = re"^u?op_".match(it(0).to_s) or it(0) == #call {
            my its_op = if: md {
                md.post_match.intern
            } :: else {
                #__call_op__
            }

            mine = precedence_of(#_main, mine, me_u)
            my its = precedence_of(#_main, its_op,
                                   if: it(0) == #call { false }
                                   :: it.size == 2)
            its < mine
        } :: else {
            !#|\ncall_paren_exceptions|(it(0))
        }
    }

    def: arg_need_parens?(e) {
       e.is_a?(List) and !#|\ncall_paren_exceptions|(e(0)) and
            !re"^u?op_"(e(0).to_s)
    }

    def: too_long?(line) { line.size > limit }

    def: emit_arglist(onto_str, args, indent, flush_left = nil) {
        my first = true
        my broke_lines = false

        if: not flush_left { flush_left = onto_str.size }

        my r = ""
        leave: try_with_breaks {
            for: arg in args {
                if: not first { r.concat!(",") }

                my at_pos = re"^.*\Z".match(r)(0).size + indent+flush_left
                if: not first { at_pos += 1 }

                my paren = ""
                wrapif: arg_need_parens?(arg) {
                    paren = "("
                    at_pos += 1
                } :: {
                    my arg_code = sourcify(arg, at_pos)
                    if: not first and arg_code.size + at_pos > limit {
                        broke_lines = true
                        try_with_breaks()
                    } :: !first {
                        r.concat!(" ")
                    }

                    r.concat!(paren)
                    r.concat!(arg_code)
                    if: re"\n".match(arg_code) {
                        broke_lines = true
                        try_with_breaks()
                    }
                } :: {
                    r.concat!(")")
                }

                first = false
            }
        }

        ;; If there are newlines *anywhere*, start over, breaking
        ;; lines after every single argument.
        if: broke_lines {
            r = ""
            first = true
            my left_spaces = " ".repeat(indent + flush_left)

            for: arg in args {
                if: not first {
                    r.concat!(",\n")
                    r.concat!(left_spaces)
                }

                my at_pos = indent + flush_left
                wrapif: arg_need_parens?(arg) {
                    r.concat!("(")
                    at_pos += 1
                } :: {
                    r.concat!(sourcify(arg, at_pos))
                } :: {
                    r.concat!(")")
                }

                first = false
            }
        }

        onto_str.concat!(r)
        broke_lines
    }

    ;;; This is global.
    ;;; Convert a list (special form) to source code.
    def: list_to_code(lst, indent) {
        if: lst.size < 1 {
            raise ArgumentError("empty lists aren't valid code")
        }

        if: not lst(0).is_a?(Symbol) {
            raise ArgumentError("lists whose first element isn't a " ++
                                "symbol aren't valid code")
        }

        my code = lst.dup
        my tag = code.shift

        my s = sourcify(tag)
        if: re"#\|".match(s) {
            s = s.substr(0, -2) ++ ":|"
        } :: else {
            s.concat!(":")
        }

        if: re"^u?op_".match(tag.to_s) {
            my op = tag.to_s.gsub(re"^u?op_", "")
            if: code.size == 1 {
                my s = op.dup
                if: re"[A-Za-z0-9_]".match(op) {
                    s.concat!(" ")
                }

                if: op_need_parens?(op.intern, true, code(0)) {
                    s ++ "(" ++ sourcify(code(0), indent + 1 + s.size) ++ ")"
                } :: else {
                    s ++ sourcify(code(0), indent + s.size)
                }
            } :: else {
                my r = if: op_need_parens?(op.intern, false, code(0)) {
                    "(" ++ sourcify(code(0), indent + 1) ++ ")"
                } :: else {
                    sourcify(code(0), indent)
                }

                if: #|\nno_space_ops|(op) {
                    r.concat!(op)
                } :: else {
                    r.concat!(" " ++ op ++ " ")
                }

                my right_text = ""
                my at = indent
                wrapif: op_need_parens?(op.intern, false, code(1)) {
                    right_text = "("
                    at += 1
                } :: {
                    at += re"^.*\Z".match(r)(0).size
                    my right = sourcify(code(1), at)
                    if: right.size + at > limit and
                        indent + 4 + right_text.size < at
                    {
                        at = indent + 4
                        r.concat!("\n" ++ " ".repeat(at))
                        at += right_text.size
                        right = sourcify(code(1), at)
                    }

                    right_text.concat!(right)
                } :: {
                    right_text.concat!(")")
                }

                r.concat!(right_text)
                r
            }
        } :: tag == #call {
            my r = if: op_need_parens?(#__call_op__, false, code(0)) {
                       "(" ++ sourcify(code(0), indent + 1) ++ ")"
                   } :: else {
                       sourcify(code(0), indent)
                   }
            r.concat!("(")
            emit_arglist(r, code.slice(1, -1), indent)
            r.concat!(")")
            r
        } :: tag == #list or tag == #d {
            my r = (if: tag == #list, "", "d") ++ "["
            emit_arglist(r, code, indent)
            r.concat!("]")
            r
        } :: tag == #quote and code(0).is_a?(Symbol) {
            code(0).inspect
        } :: tag == #progn || tag == #_progn and code.size > 1 {
            sourcify([tag, [#progn] ++ lst.slice(1, -1)], indent)
        } :: else {
            my r = s.dup
            if: code.empty? {
                "(" ++ r ++ ")"
            } :: else {
                my last_braces = false
                my ruined = false
                my broke_lines = false
                my arg_index = 1

                my need_braces? = fn: a, i {
                    (a.is_a?(List) and a(0) == #progn) or
                        #|\nto_code_brace_rules|(tag)(lst, i)
                }

                r.concat!(" ")
                while: arg_index < lst.size {
                    if: last_braces { r.concat!(" :: ") }

                    my spaces = " ".repeat(indent)
                    my arg = lst(arg_index)

                    if: need_braces?(arg, arg_index) {
                        if: !arg.is_a?(List) or arg(0) != #progn {
                            arg = [#progn, arg]
                        }

                        if: broke_lines {
                            r.concat!("\n")
                            r.concat!(spaces)
                        } :: arg_index != 1 and !last_braces {
                            r.concat!(" ")
                        }

                        r.concat!("{\n")
                        for: stmt in arg.slice(1, -1) {
                            r.concat!(spaces)
                            r.concat!("    ")
                            r.concat!(sourcify(stmt, indent + 4))
                            r.concat!("\n")
                        }
                        r.concat!(spaces ++ "}")

                        last_braces = true
                        ruined = true
                        broke_lines = false
                        arg_index += 1
                    } :: else {
                        my last = arg_index + 1
                        while: last < lst.size and !need_braces?(lst(last), last) {
                            last += 1
                        }
                        last -= 1

                        if: emit_arglist(r, lst.slice(arg_index, last),
                                         (if: ruined, 0, indent),
                                         re"^.*\Z".match(r)(0).size)
                        {
                            broke_lines = true
                        }

                        last_braces = false
                        arg_index = last + 1
                    }
                }

                r
            }
        }
    }
}

;;; The controller function. Convert some ast representation to source code.
def: sourcify(e, indent = 0) {
    my result = _conv(e.class)(e, indent).dup
    (slot: result, pretty_string) = true
    result
}

_conv = Dict(unknown_class -> fn: e, indent {
    raise Exception("not a valid ast: " ++ e.inspect)
})

_conv(Fixnum) = selfeval_to_code
_conv(Float) = selfeval_to_code
_conv(String) = selfeval_to_code
_conv(Symbol) = symbol_to_code
_conv(List) = list_to_code
