/* Bytecode.c -- internal bytecode interpreter of chai
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "Bytecode.h"
#include "mem.h"
#include "value.h"
#include "obj.h"
#include "sym.h"
#include "sym_literals.h"
#include "assorted.h"
#include "Exception.h"
#include "Module.h"
#include "ch_signal.h"
#include <inttypes.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

typedef struct {
    ch_class_t *klass;
    ch_value_t name, impl;
} method_cache_t;

typedef struct {
    size_t len;
    uint16_t *code;

    ch_value_t module;

    size_t nstrings;
    char **strings;
    ch_value_t *symbols;
    uint32_t *strlens;
    method_cache_t *method_cache;
    int n_methcached;
    ch_value_t **globals_cache;
    int n_globalscached;
} bytecode_t;

static void bytecode_initData(ch_obj_t* obj)
{
    bytecode_t* bc = (bytecode_t*)obj->data;
    bc->len = 0;
    bc->code = NULL;
    bc->nstrings = 0;
    bc->strings = NULL;
    bc->strlens = NULL;
    bc->symbols = NULL;
    bc->module = kNil;
}

static void bytecode_dup(bytecode_t* dest, bytecode_t* src)
{
    dest->len = src->len;
    dest->code = src->code;
    dest->nstrings = src->nstrings;
    dest->strings = src->strings;
    dest->strlens = src->strlens;
    dest->module = src->module;
}

void ch_Bytecode_init()
{
    ch_globalRoot(&ch_cBytecode.obj);
    ch_cBytecode = ch_class_create(SYM_Bytecode, ch_cObject);
    unsigned int bcRepr = ch_genRepr(sizeof(bytecode_t),
                                     bytecode_initData,
                                     (ch_dupFn_t)bytecode_dup);

    ((ch_class_t*)ch_cBytecode.obj->data)->reprID = bcRepr;
}

ch_value_t ch_bytecode_new(size_t size,
                           uint16_t* code,
                           size_t nstrings,
                           char** strings,
                           uint32_t* strlens,
                           ch_value_t module)
{
    ch_value_t result = ch_obj_create(ch_cBytecode);
    bytecode_t* bc = (bytecode_t*)result.obj->data;

    bc->len = size;
    bc->code = code;
    bc->nstrings = nstrings;
    bc->strings = strings;
    bc->strlens = strlens;
    bc->module = module;

    ch_value_t* symbols = ch_malloc(sizeof(ch_value_t) * nstrings);
    bc->symbols = symbols;

    int i;
    for (i = 0; i < nstrings; i++)
        bc->symbols[i] = ch_internSizedString(strlens[i], strings[i]);

    /* Overestimate number of sends. */
    int n_sends = 0;
    for (i = 0; i < bc->len; i++)
        if (bc->code[i] == OP_SEND)
            n_sends++;

    bc->method_cache = ch_malloc(n_sends * sizeof(method_cache_t));
    bc->n_methcached = 0;

    /* Overestimate number of globals */
    int n_globals = 0;
    for (i = 0; i < bc->len; i++)
        if (bc->code[i] == OP_GGET || bc->code[i] == OP_GSET)
            n_globals++;

    bc->globals_cache = ch_malloc(n_globals * sizeof(ch_value_t*));
    bc->n_globalscached = 0;

    return result;
}

#if 0 // TODO: remove
ch_value_t ch_bytecode_load(const char* filename)
{
    FILE* fp = fopen(filename, "rb");
    if(!fp) { fatal("%s: %s", filename, strerror(errno)); }

    int c1, c2;
    c1 = fgetc(fp);
    c2 = fgetc(fp);

    if (c1 == EOF || c2 == EOF)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: premature end of file", filename);

    if(c1 == '#' && c2 == '!') {
        char buffer[BUFSIZ];
        fgets(buffer, sizeof(buffer), fp); /* eat shebang line */
    } else {
        ungetc(c2, fp);
        ungetc(c1, fp);
    }

    chc_header_t header;
    if (fread(&header, sizeof(header), 1, fp) < 1)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: premature end of file in header", filename);

    if (header.magic != CHC_MAGIC)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: incorrect magic number for chai bytecode: 0x%x",
                  filename,
                  header.magic);

    if (header.version != 3)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: incorrect version for chai bytecode: %d",
                  filename,
                  header.version);

    char** strings = ch_malloc(sizeof(char*) * header.nstrings);
    uint32_t* strlens = ch_malloc(4 * header.nstrings);

    int i;
    for (i = 0; i < header.nstrings; i++) {
        if (fread(&strlens[i], 4, 1, fp) < 1)
            ch_raise2(ch_eBytecodeFormatError,
                      "%s: premature end of file in strings", filename);

        strings[i] = ch_malloc(strlens[i]);

        if (fread(strings[i], 1, strlens[i], fp) < strlens[i])
            ch_raise2(ch_eBytecodeFormatError,
                      "%s: premature end of file in strings", filename);
    }

    /* skip the padding */
    if (header.padding > 3)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: invalid padding specified in header", filename);

    char pad[3];
    if (fread(pad, 1, header.padding, fp) < header.padding) {
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: premature end of file in post-string padding", filename);
    }

    /* Now read bytecode */
    uint32_t bytecode_size;
    if (fread(&bytecode_size, sizeof(bytecode_size), 1, fp) < 1)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: premature end of file in bytecode size", filename);

    uint16_t* code = ch_malloc(2 * bytecode_size + 2);

    if (fread(code, 2, bytecode_size, fp) < bytecode_size)
        ch_raise2(ch_eBytecodeFormatError,
                  "%s: premature end of file in bytecode", filename);

    code[bytecode_size] = OP_RET; // falling off the end stops.

    return ch_bytecode_new(bytecode_size + 1,
                           code,
                           header.nstrings,
                           strings,
                           strlens);
}
#endif

ch_value_t *ch_stack = NULL, *ch_backtrace = NULL;
size_t ch_stack_size = 0, ch_backtrace_size = 0;
size_t ch_stack_top = 0, ch_backtrace_top = 0;
int ch_backtrace_runbc_depth = 0;
int ch_throwing_stackdepth = 0;

void ch_backtrace_overflow(void)
{
    const char *msg = "recursion depth exceeded maximum";
    ch_throwing_stackdepth = 1;
    ch_value_t exc = ch_funcall(ch_eStackDepthError, 1, ch_string_new1(msg));
    ch_raise(exc);
}

static void ch_vmfuncall(long);
static void ch_vmsend(ch_value_t, long);

void ch_runBytecode(ch_value_t bytecode)
{
    int old_depth;

    if (CH_FALSEP(ch_isa(bytecode, ch_cBytecode)))
        ch_raise2(ch_eTypeError, "can't run non-bytecode object");

    old_depth = ch_backtrace_runbc_depth;
    ch_backtrace_runbc_depth = ch_backtrace_top;
    ch_pushSignalHandlers();

    ch_runBytecodeInternal(bytecode, 0, NULL, NULL, NULL);

    ch_popSignalHandlers();
    ch_backtrace_runbc_depth = old_depth;
}

void ch_runBytecodeInternal(ch_value_t bytecode,
                            uint32_t ip,
                            ch_value_t *locals,
                            ch_value_t *outer,
                            ch_value_t *defaults)
{
    bytecode_t* bc = (bytecode_t*)bytecode.obj->data;

    for (;;) {
        ch_check_signals();

        switch(bc->code[ip]) {
        case OP_NOP:
            ip++;
            break;

        case OP_POP:
            ch_pop();
            ip++;
            break;

        case OP_DUP:
            ch_push(ch_peek());
            ip++;
            break;

        case OP_INT2:
            ch_push(ch_fix(*((int16_t*)(&bc->code[ip + 1]))));
            ip += 2;
            break;

        case OP_INT4:
            ch_push(ch_fix(*((int32_t*)(&bc->code[ip + 1]))));
            ip += 3;
            break;

        /* no INT8 yet. */

        case OP_FLOAT:
            ch_push(ch_float(*((double*)(&bc->code[ip + 1]))));
            ip += 5;
            break;

        case OP_STRING:
            ch_push(ch_string_new2(bc->strlens[bc->code[ip + 1]],
                                   bc->strings[bc->code[ip + 1]]));
            ip += 2;
            break;

        case OP_SYM:
            ch_push(bc->symbols[bc->code[ip + 1]]);
            ip += 2;
            break;

        case OP_NIL:
            ch_push(kNil);
            ip++;
            break;

        case OP_TRUE:
            ch_push(kTrue);
            ip++;
            break;

        case OP_FALSE:
            ch_push(kFalse);
            ip++;
            break;

        case OP_LISTE:
            ch_push(ch_list_new());
            ip++;
            break;

        case OP_DICTE:
            ch_push(ch_dict_new());
            ip++;
            break;

        /* TODO: LISTN, DICT, DICTN */

        case OP_LIST:
        {
            uint16_t i, n = bc->code[ip + 1];
            ch_value_t result = ch_list_new();
            for (i = 0; i < n; i++)
                CH_SEND(result, push, ch_pop());
            ch_push(result);
            ip += 2;
            break;
        }

        case OP_FN:
            ch_push(ch_closure_new(bc->code[ip + 3], /* ndefault_values */
                                   bc->code[ip + 4], /* nouter_vars */
                                   bytecode,
                                   *((uint32_t*)&bc->code[ip + 1])));

            if (ch_backtrace_top > ch_backtrace_runbc_depth)
                ch_setSlot(ch_peek(), SYM_lexical_parent,
                           ch_backtrace[ch_backtrace_top - 1]);
            ip += 5;
            break;

        case OP_GOTO:
        {
            uint32_t label = *((uint32_t*)(&bc->code[ip + 1]));
            ip = label;
            break;
        }

        case OP_IF:
            if(CH_FALSEP(ch_pop()))
                ip += 3;
            else
                ip = *((uint32_t*)(&bc->code[ip + 1]));

            break;

        case OP_GSET: {
            ch_value_t real_module, real_name;
            ch_module_t *modt;

            ch_lookup_global_name(bc->module, bc->symbols[bc->code[ip + 1]],
                                  &real_module, &real_name);

            modt = (ch_module_t*)real_module.obj->data;
            ch_value_t ret;
            if (ch_ht_get(&modt->globals, real_name, &ret) < 0) {
                ret.type = CH_REF;
                ret.ref = ch_malloc(sizeof (ch_value_t));
                *ret.ref = kNil;
                ch_ht_set(&modt->globals, real_name, ret);
            }

            bc->code[ip] = OP_GSET | (1 << 15);
            bc->globals_cache[bc->n_globalscached] = ret.ref;
            bc->code[ip + 1] = bc->n_globalscached++;
            break;
        }

        case OP_GSET | (1 << 15):
            *bc->globals_cache[bc->code[ip + 1]] = ch_peek();
            ip += 2;
            break;

        case OP_GGET: {
            ch_value_t real_m, real_name;
            ch_module_t *modt;
            ch_lookup_global_name(bc->module, bc->symbols[bc->code[ip + 1]],
                                  &real_m, &real_name);

            modt = (ch_module_t*)real_m.obj->data;
            ch_value_t ret;
            if (ch_ht_get(&modt->globals, real_name, &ret) < 0)
                ch_raise2(ch_eNameError, "variable %s undefined",
                          bc->symbols[bc->code[ip + 1]].sym + sizeof(long));

            bc->code[ip] = OP_GGET | (1 << 15);
            bc->globals_cache[bc->n_globalscached] = ret.ref;
            bc->code[ip + 1] = bc->n_globalscached++;
            break;
        }

        case OP_GGET | (1 << 15):
            ch_push(*bc->globals_cache[bc->code[ip + 1]]);
            ip += 2;
            break;

        case OP_LSET:
            locals[bc->code[ip + 1]] = ch_peek();
            ip += 2;
            break;

        case OP_LGET:
            ch_push(locals[bc->code[ip + 1]]);
            ip += 2;
            break;

        case OP_RSET:
            *(locals[bc->code[ip + 1]].ref) = ch_peek();
            ip += 2;
            break;

        case OP_RGET:
            ch_push(*(locals[bc->code[ip + 1]].ref));
            ip += 2;
            break;

        case OP_OSET:
            *(outer[bc->code[ip + 1]].ref) = ch_peek();
            ip += 2;
            break;

        case OP_OGET:
            ch_push(*(outer[bc->code[ip + 1]].ref));
            ip += 2;
            break;

        case OP_OREF:
            ch_push(outer[bc->code[ip + 1]]);
            ip += 2;
            break;

        case OP_GET_ATTR:
        {
            ch_value_t slot_val = kNil;
            ch_getSlot(ch_pop(), bc->symbols[bc->code[ip + 1]], &slot_val);
            ch_push(slot_val);
            ip += 2;
            break;
        }

        case OP_SET_ATTR:
        {
            ch_value_t obj = ch_pop(); /* object whose slot to set */
            ch_setSlot(obj, bc->symbols[bc->code[ip + 1]], ch_peek());
            ip += 2;
            break;
        }

        case OP_POP_GET_ATTR:
        {
            ch_value_t slot_name = ch_pop();
            ch_value_t obj = ch_pop();
            if (slot_name.type != CH_SYMBOL)
                ch_raise2(ch_eTypeError, "slot names must be symbols");
            ch_value_t slot_val;
            ch_getSlot(obj, slot_name, &slot_val);
            ch_push(slot_val);
            ip++;
            break;
        }

        case OP_POP_SET_ATTR:
        {
            ch_value_t slot_name = ch_pop();
            ch_value_t obj = ch_pop();
            if (slot_name.type != CH_SYMBOL)
                ch_raise2(ch_eTypeError, "slot names must be symbols");
            ch_setSlot(obj, slot_name, ch_peek());
            ip++;
            break;
        }

        case OP_CALL:
            ch_vmfuncall(bc->code[ip + 1]);
            ip += 2;
            break;

        case OP_SEND:
        {
            method_cache_t *entry = &bc->method_cache[bc->n_methcached];
            entry->name  = bc->symbols[bc->code[ip+2]];
            entry->klass = NULL;
            entry->impl  = kNil;
            bc->code[ip + 2] = bc->n_methcached++;
            bc->code[ip] |= (1 << 15);
            break;
        }

        case OP_SEND | (1 << 15):
        {
            method_cache_t *entry = &bc->method_cache[bc->code[ip + 2]];
            ch_value_t receiver = ch_pop();
            ch_class_t *klass = (ch_class_t*)ch_classOf(receiver).obj->data;
            if (entry->klass != klass || entry->impl.type == CH_NIL) {
                entry->klass = klass;
                entry->impl = ch_lookupMethod(klass, entry->name);
            }

            if (entry->impl.type == CH_NIL || entry->impl.ref->type == CH_NIL) {
                ch_value_t method_missing_impl =
                    ch_lookupMethod(entry->klass, SYM_method_missing);
                if (method_missing_impl.type == CH_NIL ||
                    method_missing_impl.ref->type == CH_NIL) {
                    ch_default_method_missing(receiver, entry->name);
                }

                ch_push(entry->name);
                ch_push(receiver);
                if (method_missing_impl.ref->type == CH_REF)
                    ch_push(*method_missing_impl.ref->ref);
                else
                    ch_push(*method_missing_impl.ref);

                ch_vmfuncall(bc->code[ip + 1] + 2);
            } else {
                ch_push(receiver);
                if (entry->impl.ref->type == CH_REF)
                    ch_push(*(entry->impl.ref->ref));
                else
                    ch_push(*(entry->impl.ref));

                ch_vmfuncall(bc->code[ip + 1] + 1);
            }

            ip += 3;
            break;
        }

        /* TODO: CALLN, SENDN */

        case OP_CHECKARGS:
        {
            unsigned char *operands = (unsigned char*)(&bc->code[ip + 1]);
            /*
            if (operands[1] != 0 || operands[2] != 0)
                ch_raise2(ch_eUnimplementedError,
                          "variadic arguments not supported yet");
            */

            int argslo = operands[0];
            int argsopt = operands[1];
            int argshi = argslo + argsopt;
            int restp = operands[2];
            int nparams = argshi + restp;
            long nlocals = nparams + operands[3];
            locals = alloca(nlocals * sizeof(ch_value_t));
            memset(locals, '\0', nlocals * sizeof(ch_value_t));

            ch_value_t rest = kNil;
            if (restp)
                rest = ch_list_new();

            long argc = ch_pop().i;
            if (argc < argslo)
                ch_raise2(ch_eArgumentError,
                          "too few arguments: %d for %d",
                          argc,
                          argslo);

            if (!restp && argc > argshi)
                ch_raise2(ch_eArgumentError,
                          "too many arguments: %d for %d",
                          argc,
                          argshi);

            int n = bc->code[ip + 3]; /* Number of extra operands */

            int i;
            for(i = ip + 4; i < ip + 4 + n; i++) {
                uint16_t index = bc->code[i];
                locals[index].i = 1;
            }
            ip = i;

            for(i = 0; i < argc; i++) {
                if (i >= argshi) {
                    CH_SEND(rest, push, ch_pop());
                } else if (locals[i].i) {
                    locals[i].type = CH_REF;
                    locals[i].ref = ch_malloc(sizeof(ch_value_t));
                    *(locals[i].ref) = ch_pop();
                } else {
                    locals[i] = ch_pop();
                }
            }

            /*if (restp)
              CH_SEND(rest, reverse$b);*/

            if (i > argshi)
                i = argshi;

            for(; i < nlocals; i++) {
                if (locals[i].i) {
                    locals[i].type = CH_REF;
                    locals[i].ref = ch_malloc(sizeof(ch_value_t));
                    locals[i].ref->type = CH_NIL;
                    locals[i].ref->i = 0;
                }

                if (i < argshi) {
                    ch_value_t defaultval = defaults[argshi - i - 1];
                    if (locals[i].i)
                        *(locals[i].ref) = defaultval;
                    else
                        locals[i] = defaultval;
                } else if (i == argshi && restp) {
                    if (locals[i].i)
                        *(locals[i].ref) = rest;
                    else
                        locals[i] = rest;
                }
            }
            break;
        }

        case OP_REBIND:
            locals[bc->code[ip + 1]].ref = ch_malloc(sizeof(ch_value_t));
            locals[bc->code[ip + 1]].ref->type = CH_NIL;
            ip += 2;
            break;

        case OP_RET:
            return;

        case OP_NOT:
            if (CH_FALSEP(ch_pop()))
                ch_push(kTrue);
            else
                ch_push(kFalse);

            ip++;
            break;

        case OP_UWND:
            ch_unwind();
            break;

        case OP_UWNDP:
        {
            /* Use label for stack top, index for backtrace top */
            if (ch_handlers_top >= ch_handlers_size) {
                ch_handlers_size *= 2;
                ch_handlers = ch_realloc(ch_handlers,
                                         sizeof(jmp_buf) * ch_handlers_size);
            }

            size_t stack_top_old     = ch_stack_top;
            size_t backtrace_top_old = ch_backtrace_top;
            if (setjmp(ch_handlers[ch_handlers_top++]) == 0) {
                ch_runBytecodeInternal(bytecode, ip + 5, locals, outer, defaults);
                ch_handlers_top--;
                ip = *((uint32_t*)&bc->code[ip + 3]);
            } else {
                ch_stack_top = stack_top_old;
                ch_backtrace_top = backtrace_top_old;
                ip = *((uint32_t*)&bc->code[ip + 1]);
            }
            break;
        }

        default:
            ch_raise2(ch_eBytecodeFormatError,
                      "invalid opcode %d", bc->code[ip]);
        }
    }

    fatal("reached end of ch_runBytecodeInternal()");
}

/* slight modifications of ch_vfuncall() and ch_vsend(), respectively */
static void ch_vmfuncall(long argc)
{
    ch_value_t f = ch_pop();

    /* mostly a copy of ch_vfuncall() */
    if(ch_cPrimitive.obj != ch_classOf(f).obj) {
        if(ch_cClosure.obj == ch_classOf(f).obj) {
            ch_push(ch_fix(argc));
            ch_callClosure(f);
            return;
        }

        ch_push(f);
        ch_vmsend(SYM__call_, argc);
        return;
    }

    ch_value_t argv[argc];
    ch_prim_t* prim = (ch_prim_t*)f.obj->data;

    ch_backtrace_push(f);

    if (prim->arity >= 0 && argc != prim->arity)
        ch_raise2(ch_eArgumentError,
                  "wrong number of arguments: %d for %d",
                  argc,
                  prim->arity);

    int i;
    for (i = 0; i < argc; i++)
        argv[i] = ch_pop();

    switch(prim->arity) {
    case 0:
        ch_push(((ch_value_t(*)(void))prim->impl)());
        break;
    case 1:
        ch_push(((ch_value_t(*)(ch_value_t))prim->impl)(argv[0]));
        break;
    case 2:
        ch_push(((ch_value_t(*)(ch_value_t,ch_value_t))
                 prim->impl)(argv[0], argv[1]));
        break;
    case 3:
        ch_push(((ch_value_t(*)(ch_value_t,ch_value_t,ch_value_t))
                 prim->impl)(argv[0], argv[1], argv[2]));
        break;
    case 4:
        ch_push(((ch_value_t(*)(ch_value_t,ch_value_t,ch_value_t,ch_value_t))
                 prim->impl)(argv[0], argv[1], argv[2], argv[3]));
        break;
    case -1:
        ch_push(prim->impl(argc, argv));
        break;
    case -2:
        ch_push(((ch_value_t(*)(ch_value_t,long,ch_value_t*))
                 prim->impl)(argv[0], argc - 1, &argv[1]));
        break;
    }

    ch_backtrace_pop();
}

static void ch_vmsend(ch_value_t method, long argc)
{
    if(method.type != CH_SYMBOL)
        ch_raise2(ch_eTypeError, "method names must be symbols");

    ch_value_t obj = ch_peek();
    ch_class_t* class = (ch_class_t*)ch_classOf(obj).obj->data;

    ch_value_t f = ch_lookupMethod(class, method);
    if (f.type == CH_NIL || f.ref->type == CH_NIL)
        ch_raise2(ch_eNoMethodError,
                  "method %s not found",
                  method.sym + sizeof(long));

    argc++;
    if (f.ref->type == CH_REF)
        ch_push(*f.ref->ref);
    else
        ch_push(*f.ref);

    ch_vmfuncall(argc);
}
