/* Bytecode.h -- opcode specification and internal bytecode interface
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#ifndef CH_BYTECODE_H
#define CH_BYTECODE_H

#include "mem.h"
#include "obj.h"
#include "value.h"
#include <inttypes.h>
#include <string.h>
#include <stdio.h>

/*
 * BYTECODE FORMAT
 *
 * Opcodes are 2 bytes each. Any required operands follow directly
 * after the opcode.
 *
 * The VM is a stack machine. It doesn't know about much of
 * anything. It can do conditionals and loops, deal with variables,
 * create any object which has a literal, and the only real operations
 * it knows about are function calls, method calls, attribute
 * accesses. All the "real work" is done in methods defined in C.
 *
 * Labels are represented in the bytecode stream as 4-byte unsigned
 * integer values indicating an absolute position in the same
 * bytecode stream.
 *
 * CALLING CONVENTIONS
 *
 * 1. CALLER
 *
 * push arguments right to left.  Use a variant of call or send
 * specifying the number of arguments.  For message sends, the target
 * and message are assumed to be there, and are not included in the
 * argument count. A dynamic message name is pushed last, which should
 * hopefully make for less stack twiddling. Functions are considered
 * "first parameters" and are pushed last as well.
 *
 * 2. CALLEE
 *
 * The checkargs instruction will pop the number of arguments from the
 * stack, and check to make sure that it is >= the required args, and
 * if there is no rest parameter, <= required + optional args, and
 * then bind everything.
 */

/* Opcodes! */
enum {
    OP_NOP = 0,
    OP_POP,
    OP_DUP,

    /* The following push values on the stack. */
    OP_INT2, /* 2-byte */
    OP_INT4, /* etc. */
    OP_INT8,
    OP_FLOAT,  /* 8-byte */
    OP_STRING, /* two-byte index into string table */
    OP_SYM,

    OP_NIL,
    OP_TRUE,
    OP_FALSE,

    /* empty */
    OP_LISTE,
    OP_DICTE,

    /* operand is 2-byte number of elements to pop from the
     * stack. First popped is first in the list. */
    OP_LIST,
    /* Pops for the operand instead */
    OP_LISTN,

    /* number of PAIRS of elements to pop from the stack.
     * pop => value, THEN pop => key. */
    OP_DICT,
    /* operand popped from stack */
    OP_DICTN,

    /* Functions validate their own arguments */
    OP_FN, /* takes a label, N and M, then pops M values then N references */

    /* Control flow */
    OP_GOTO, /* label */
    OP_IF,   /* label - jump if popped value is true */

    /* "set" operations DON'T pop the value. */

    /* Globals - name is an operand */
    OP_GSET,
    OP_GGET,

    /* Lexical variables - index is an operand (16bit for simplicity) */
    OP_LSET,
    OP_LGET,

    /* Lexical variables through references - same deal */
    OP_RSET,
    OP_RGET,

    /* Variables referenced from outer scopes - indexed separately */
    OP_OSET,
    OP_OGET,
    OP_OREF, /* To get the ref itself. */

    /* Attributes: attribute is an operand, pop the object */
    OP_GET_ATTR,
    OP_SET_ATTR, /* Pop object, peek at value */

    /* The same, but pop the slot name first */
    OP_POP_GET_ATTR,
    OP_POP_SET_ATTR,

    /* Calls - operand is number of args. For sends, tgt & msg are
     * implied. No more than 255 arguments are allowed to any
     * procedure (including the target).  For message sends, the
     * target is pushed before the message, because the message is
     * usually popped. */
    OP_CALL,
    OP_SEND, /* takes nargs then message as an operands. */

    /* N is popped first */
    OP_CALLN,
    OP_SENDN, /* takes message as operand */

    /* checking arguments to a function - has four fixed 1-byte
     * operands (the number of required and optional parameters,
     * whether or not there's a rest parameter, and number of other
     * locals), followed by 2-byte N, followed by N 2-byte
     * variable indices that need to be heap-allocated.
     */
    OP_CHECKARGS,

    /* Allocate a new binding for a var and change its ref to point at it */
    OP_REBIND,

    /* Return from a function */
    OP_RET,

    /* Logical negation */
    OP_NOT,

    /* Unwind/Protect */
    OP_UWND,
    OP_UWNDP
};

extern void ch_Bytecode_init(void);

/* the bytecode object takes responsibility for code and strings */
extern ch_value_t ch_bytecode_new(size_t size,
                                  uint16_t* code,
                                  size_t nstrings,
                                  char** strings,
                                  uint32_t* strlens,
                                  ch_value_t module);

/* load from a file */
extern ch_value_t ch_bytecode_load(const char* filename);
extern void ch_runBytecode(ch_value_t bytecode);

#define CHC_MAGIC 0xdeafa150

typedef struct chc_header {
    uint32_t magic;      /* CHC_MAGIC */
    uint16_t version;    /* 1 right now. */
    uint16_t nstrings;
    uint16_t padding;    /* after strings */
    uint16_t my_padding; /* Just so I can see it. :) */
} chc_header_t;

extern ch_value_t *ch_stack,     *ch_backtrace;
extern size_t      ch_stack_size, ch_backtrace_size;
extern size_t      ch_stack_top,  ch_backtrace_top;
extern int         ch_throwing_stackdepth;

extern ch_value_t ch_cur_pad;
extern ch_value_t* ch_pads;
extern size_t ch_pads_size;
extern size_t ch_pads_top;

void ch_runBytecodeInternal(ch_value_t bytecode,
                            uint32_t ip,
                            ch_value_t *locals,
                            ch_value_t *outer,
                            ch_value_t *defaults);

/***
 * Push a value onto the stack.
 */
static inline void ch_push(ch_value_t val)
{
    /* Grow if needed. */
    if (ch_stack_top >= ch_stack_size) {
        ch_stack_size *= 2;
        ch_stack = ch_realloc(ch_stack, sizeof(ch_value_t) * ch_stack_size);
    }

    ch_stack[ch_stack_top++] = val;
}

/***
 * Pop a value from the stack and return it.
 */
static inline ch_value_t ch_pop(void)
{
    if (ch_stack_top == 0)
        fatal("can't pop empty stack");

    return ch_stack[--ch_stack_top];
}

/***
 * Return the top value of the stack without popping it.
 */
static inline ch_value_t ch_peek(void)
{
    if (ch_stack_top == 0)
        fatal("can't peek empty stack");

    return ch_stack[ch_stack_top - 1];
}

/***
 * Push a function onto the current backtrace.
 */
void ch_backtrace_overflow(void);

static inline void ch_backtrace_push(ch_value_t f)
{
    if (!ch_throwing_stackdepth && ch_backtrace_top >= ch_backtrace_size)
        ch_backtrace_overflow();

    ch_backtrace[ch_backtrace_top++] = f;
}

/***
 * Pop a function from the current backtrace.
 */
static inline void ch_backtrace_pop(void)
{
    if (ch_backtrace_top == 0)
        fatal("can't pop empty backtrace");

    ch_backtrace_top--;
}

#endif /* CH_BYTECODE_H */
