/* Closure.c -- implementation of chai Closure class
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "assorted.h"
#include "Bytecode.h"
#include "value.h"
#include "obj.h"
#include "sym_literals.h"
#include <inttypes.h>

typedef struct {
    ch_value_t bc;
    uint32_t ip;
    ch_value_t *env;
    ch_value_t *defs;
} clo_t;

static void clo_initData(ch_obj_t* obj)
{
    clo_t* clo = (clo_t*)obj->data;
    clo->bc = kNil;
    clo->ip = 0;
    clo->env = NULL;
}

static void clo_dup(clo_t* dest, clo_t* src)
{
    dest->env = src->env; /* We don't modify it, should be fine. */
    dest->bc = src->bc;
    dest->ip = src->ip;
}

void ch_Closure_init()
{
    ch_globalRoot(&ch_cClosure.obj);
    ch_cClosure = ch_class_create(SYM_Closure, ch_cFunction);
    unsigned int cloRepr = ch_genRepr(sizeof(clo_t),
                                      clo_initData,
                                      (ch_dupFn_t)clo_dup);

    ((ch_class_t*)ch_cClosure.obj->data)->reprID = cloRepr;
}

void ch_defineClosureMethods()
{
    /*ch_defineMethod(ch_cClosure, SYM__init_, 3, (ch_cfn_t)clo_init);*/
    /* TODO: undefine _init_ completely on Closure. */
    ch_defineMethod(ch_cClosure,
                    SYM__call_,
                    -2,
                    (ch_cfn_t)ch_vfuncall);
}

void ch_callClosure(ch_value_t f)
{
    clo_t* clo = (clo_t*)f.obj->data;
    ch_backtrace_push(f);
    ch_runBytecodeInternal(clo->bc, clo->ip, NULL, clo->env, clo->defs);
    ch_backtrace_pop();
}

ch_value_t ch_closure_new(int ndefaults, int env_size,
                          ch_value_t bc, uint32_t ip)
{
    int i;

    ch_value_t result = ch_obj_create(ch_cClosure);
    clo_t* clo = (clo_t*)result.obj->data;

    clo->env = ch_malloc(env_size * sizeof (ch_value_t));
    clo->defs = ch_malloc(ndefaults * sizeof (ch_value_t));

    /* default values are pushed after references, and in left-to-right order */
    for (i = 0; i < ndefaults; i++) {
        clo->defs[i] = ch_pop();
    }

    for (i = 0; i < env_size; i++) { /* This loop means that refs are pushed */
        clo->env[i] = ch_pop();      /* in _reverse_ order.                  */
        if (clo->env[i].type != CH_REF)
            fatal("non-ref given to closure");
    }

    clo->bc = bc;
    clo->ip = ip;
    return result;
}
