/* Dict.c -- implementation of chai Dict class
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "assorted.h"
#include "value.h"
#include "obj.h"
#include "ht.h"
#include "Exception.h"
#include "sym_literals.h"

static ch_value_t dict_init(ch_value_t self, long argc, ch_value_t* args)
{
    ch_value_t fallback = kNil;
    ch_checkArgs(argc, args, 0, 1, 0, &fallback);

    if(fallback.type != CH_NIL) {
        ch_setSlot(self, SYM_fallback, fallback);
    }

    return kNil;
}

static ch_value_t dict_size(ch_value_t self)
{
    ch_ht_t* ht = (ch_ht_t*)self.obj->data;
    return ch_fix(ht->size);
}

static ch_value_t dict_emptyp(ch_value_t self)
{
    ch_ht_t* ht = (ch_ht_t*)self.obj->data;
    return ht->size == 0 ? kTrue : kFalse;
}

static ch_value_t dict_set(ch_value_t self, ch_value_t key, ch_value_t val)
{
    ch_ht_set((ch_ht_t*)self.obj->data, key, val);
    return kNil;
}

static ch_value_t dict_get(ch_value_t self, ch_value_t key)
{
    ch_ht_t* ht = (ch_ht_t*)self.obj->data;

    ch_value_t result;
    if(ch_ht_get(ht, key, &result) < 0) {
        ch_value_t fallbackFn;
        if(ch_getSlot(self, SYM_fallback, &fallbackFn) < 0) {
            return kNil;
        }
        result = ch_funcall(fallbackFn, 1, key);
    }
    return result;
}

static ch_value_t dict_hasKeyp(ch_value_t self, ch_value_t key)
{
    ch_ht_t* ht = (ch_ht_t*)self.obj->data;
    return ch_ht_hasKeyp(ht, key) ? kTrue : kFalse;
}

static ch_value_t dict_mergeb(ch_value_t self, ch_value_t other)
{
    if (CH_FALSEP(ch_isa(other, ch_cDict)))
        ch_raise2(ch_eTypeError, "Dict#merge! expects a Dict argument");

    ch_ht_t* hta = (ch_ht_t*)self.obj->data;
    ch_ht_t* htb = (ch_ht_t*)other.obj->data;
    ch_ht_mergeb(hta, htb);
    return kNil;
}

static ch_value_t dict_merge(ch_value_t self, ch_value_t other)
{
    ch_value_t dupped = ch_obj_dup(self);
    dict_mergeb(dupped, other);
    return dupped;
}

static ch_value_t dict_keys(ch_value_t self)
{
    ch_ht_t *ht = (ch_ht_t*)self.obj->data;
    return ch_ht_keys(ht);
}

static ch_value_t dict_delete_b(ch_value_t self, ch_value_t k)
{
    ch_ht_t *ht = (ch_ht_t*)self.obj->data;
    return ch_ht_delete(ht, k) < 0 ? kFalse : kTrue;
}

static ch_value_t dict_without(ch_value_t self, ch_value_t k)
{
    ch_value_t result = ch_obj_dup(self);
    dict_delete_b(result, k);
    return result;
}

void ch_defineDictMethods()
{
    ch_defineMethod(ch_cDict, SYM__init_, -2, (ch_cfn_t)dict_init);
    ch_defineMethod(ch_cDict, SYM_size, 0, (ch_cfn_t)dict_size);
    ch_defineMethod(ch_cDict, SYM_empty$q, 0, (ch_cfn_t)dict_emptyp);
    ch_defineMethod(ch_cDict, SYM_set, 2, (ch_cfn_t)dict_set);
    ch_defineMethod(ch_cDict, SYM__call_$e, 2, (ch_cfn_t)dict_set);
    ch_defineMethod(ch_cDict, SYM_get, 1, (ch_cfn_t)dict_get);
    ch_defineMethod(ch_cDict, SYM__call_, 1, (ch_cfn_t)dict_get);
    ch_defineMethod(ch_cDict, SYM_has_key$q, 1, (ch_cfn_t)dict_hasKeyp);
    ch_defineMethod(ch_cDict, SYM_merge$b, 1, (ch_cfn_t)dict_mergeb);
    ch_defineMethod(ch_cDict, SYM_merge, 1, (ch_cfn_t)dict_merge);
    ch_defineMethod(ch_cDict, SYM_keys, 0, (ch_cfn_t)dict_keys);
    ch_defineMethod(ch_cDict, SYM_delete_key$b, 1, (ch_cfn_t)dict_delete_b);
    ch_defineMethod(ch_cDict, SYM_without_key, 1, (ch_cfn_t)dict_without);
}
