/* Exception.c -- chai error handling/Exception class
   September 30, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "Exception.h"
#include "mem.h"
#include "obj.h"
#include "state.h"
#include "Bytecode.h"
#include "Module.h"
#include "sym_literals.h"
#include <stdlib.h>
#include <setjmp.h>
#include <stdarg.h>

ch_value_t ch_eException;
ch_value_t ch_eTypeError;
ch_value_t ch_eNameError;
ch_value_t ch_eArgumentError;
ch_value_t ch_eNoMethodError;
ch_value_t ch_eUnimplementedError;
ch_value_t ch_eBytecodeFormatError;
ch_value_t ch_eFileModeError;
ch_value_t ch_eOutOfBoundsError;
ch_value_t ch_eArityError;
ch_value_t ch_eParseError;
ch_value_t ch_eStackDepthError;
ch_value_t ch_eFileOpenError;
ch_value_t ch_eDivideByZeroError;

/* Exception handler stack */
size_t ch_handlers_size = 0;
size_t ch_handlers_top = 0;
jmp_buf *ch_handlers = NULL;

void ch_handlers_init()
{
    ch_handlers_size = 256;
    ch_handlers_top = 0;
    ch_handlers = ch_malloc(sizeof(jmp_buf) * ch_handlers_size);
}

ch_value_t ch_protect(ch_value_t (*f)(ch_value_t), ch_value_t arg, int *status)
{
    if (ch_handlers_top >= ch_handlers_size) {
        ch_handlers_size *= 2;
        ch_handlers = ch_realloc(ch_handlers,
                                 sizeof(jmp_buf) * ch_handlers_size);
    }

    ch_value_t result;
    size_t stack_top = ch_stack_top;
    size_t bt_top = ch_backtrace_top;

    if(setjmp(ch_handlers[ch_handlers_top++]) == 0) {
        result = f(arg);
        ch_handlers_top--;
        *status = CH_PROTECT_NORMAL;
        return result;
    }

    /* f() threw an exception. handler is already popped */
    ch_stack_top = stack_top;
    ch_backtrace_top = bt_top;
    *status = CH_PROTECT_UNWOUND;
    return kNil;
}

void ch_unwind()
{
    if (ch_handlers_top == 0)
        fatal("no exception handlers what?");

    longjmp(ch_handlers[--ch_handlers_top], 1);
}

void ch_raise(ch_value_t exc)
{
    ch_set_global(ch_m_builtins, SYM_$_exc, exc);
    ch_setSlot(exc, SYM_backtrace,
               ch_list_new2(ch_backtrace_top, ch_backtrace));

    ch_throwing_stackdepth = 0; /* in case we are */

    ch_unwind();
}

void ch_raise2(ch_value_t type, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZ];

    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);
    va_end(args);

    ch_value_t exc = ch_funcall(type, 1, ch_string_new1(buf));
    ch_raise(exc);
}

static ch_value_t exc_init(ch_value_t self, long argc, ch_value_t *argv)
{
    ch_value_t message = kNil;
    ch_checkArgs(argc, argv, 0, 1, 0, &message);

    if (message.type == CH_NIL)
        message = ch_string_new();

    ch_setSlot(self, SYM_message, message);

    return kNil;
}

static ch_value_t exc_message(ch_value_t self)
{
    ch_value_t message;
    ch_getSlot(self, SYM_message, &message);
    return message;
}

static ch_value_t exc_backtrace(ch_value_t self)
{
    ch_value_t bt;
    ch_getSlot(self, SYM_backtrace, &bt);
    return bt;
}

static ch_value_t prettify_fn(ch_value_t f)
{
    ch_value_t tmp, result;
    ch_value_t method_of;

    method_of = kNil;
    result = ch_string_new();
    ch_getSlot(f, SYM_method_of, &method_of);

    tmp = method_of.type == CH_NIL ? f : method_of;

    if (!ch_getSlot(tmp, SYM_containing_module, &tmp))
        if (tmp.type == CH_SYMBOL) {
            ch_value_t curmod_name;
            ch_module_t *curmod;

            curmod = (ch_module_t*)ch_current_module.obj->data;
            curmod_name = curmod->module_name;

            if (tmp.sym != curmod_name.sym) {
                CH_SEND(result, concat$b,
                        ch_string_new1(tmp.sym + sizeof (long)));
                CH_SEND(result, concat$b, ch_string_new1(":"));
            }
        }

    if (method_of.type != CH_NIL) {
        CH_SEND(result, concat$b, method_of);
        CH_SEND(result, concat$b, ch_string_new1("#"));
    }

    if (ch_getSlot(f, SYM_name, &tmp) < 0)
        CH_SEND(result, concat$b, ch_string_new1("<anon>"));
    else
        CH_SEND(result, concat$b, tmp);

    CH_SEND(result, concat$b, ch_string_new1("()"));

    if (method_of.type == CH_NIL && !ch_getSlot(f, SYM_lexical_parent, &tmp)) {
        CH_SEND(result, concat$b, ch_string_new1(" (from "));
        CH_SEND(result, concat$b, prettify_fn(tmp));
        CH_SEND(result, concat$b, ch_string_new1(")"));
    }

    return result;
}

static void exc_pretty_error_append(ch_value_t s,
                                    ch_value_t fn,
                                    ch_value_t instr,
                                    ch_value_t newline)
{
    CH_SEND(s, concat$b, instr);
    CH_SEND(s, concat$b, prettify_fn(fn));
    CH_SEND(s, concat$b, newline);
}

static ch_value_t exc_pretty_error(ch_value_t self)
{
    ch_value_t result = ch_string_new();
    ch_value_t tmp, bt, newline, instr;
    ch_list_t *btl;
    int i;

    newline = ch_string_new1("\n");
    instr = ch_string_new1("  in ");

    if (!ch_getSlot(ch_classOf(self), SYM_containing_module, &tmp)) {
        if (tmp.type == CH_SYMBOL && tmp.sym != ch_current_module_name().sym) {
            CH_SEND(result, concat$b, CH_SEND2(tmp, to_s));
            CH_SEND(result, concat$b, ch_string_new1(":"));
        }
    }

    CH_SEND(result, concat$b, CH_SEND2(ch_classOf(self), to_s));
    CH_SEND(result, concat$b, ch_string_new1(": "));

    ch_getSlot(self, SYM_message, &tmp);
    CH_SEND(result, concat$b, tmp);
    CH_SEND(result, concat$b, newline);

    ch_getSlot(self, SYM_backtrace, &bt);
    if (CH_FALSEP(ch_isa(bt, ch_cList)))
        return result;

    btl = (ch_list_t*)bt.obj->data;
    if (btl->size > 14) {
        for (i = btl->size - 1; i >= (int)btl->size - 7; i--)
            exc_pretty_error_append(result, btl->elems[i], instr, newline);

        CH_SEND(result, concat$b, ch_string_new1("  ["));
        CH_SEND(result, concat$b, ch_fix(btl->size - 14));
        CH_SEND(result, concat$b, ch_string_new1(" levels omitted...]\n"));

        for (i = 6; i >= 0; i--)
            exc_pretty_error_append(result, btl->elems[i], instr, newline);
    } else {
        for (i = btl->size - 1; i >= 0; i--)
            exc_pretty_error_append(result, btl->elems[i], instr, newline);
    }

    return result;
}

void ch_Exceptions_init()
{
    ch_eException = ch_class_create(SYM_Exception, ch_cObject);
    ch_eTypeError = ch_class_create(SYM_TypeError, ch_eException);
    ch_eNameError = ch_class_create(SYM_NameError, ch_eException);
    ch_eArgumentError = ch_class_create(SYM_ArgumentError, ch_eException);
    ch_eNoMethodError = ch_class_create(SYM_NoMethodError, ch_eException);
    ch_eUnimplementedError = ch_class_create(SYM_UnimplementedError, ch_eException);
    ch_eBytecodeFormatError = ch_class_create(SYM_BytecodeFormatError, ch_eException);
    ch_eFileModeError = ch_class_create(SYM_FileModeError, ch_eException);
    ch_eOutOfBoundsError = ch_class_create(SYM_OutOfBoundsError, ch_eException);
    ch_eArityError = ch_class_create(SYM_ArityError, ch_eException);
    ch_eParseError = ch_class_create(SYM_ParseError, ch_eException);
    ch_eStackDepthError = ch_class_create(SYM_StackDepthError, ch_eException);
    ch_eDivideByZeroError = ch_class_create(SYM_DivideByZeroError, ch_eException);

    /* Methods */
    ch_defineMethod(ch_eException, SYM__init_, -2, (ch_cfn_t)exc_init);
    ch_defineMethod(ch_eException, SYM_message, 0, (ch_cfn_t)exc_message);
    ch_defineMethod(ch_eException, SYM_backtrace, 0, (ch_cfn_t)exc_backtrace);
    ch_defineMethod(ch_eException, SYM_pretty_error,
                    0, (ch_cfn_t)exc_pretty_error);
    ch_defineMethod(ch_eException, SYM_to_s, 0, (ch_cfn_t)exc_message);
}
