/* File.c -- implementation of chai File class
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "assorted.h"
#include "mem.h"
#include "value.h"
#include "obj.h"
#include "Exception.h"
#include "ch_signal.h"
#include "sym_literals.h"
#include <stdio.h>
#include <errno.h>
#include <inttypes.h>
#include <string.h>

/* Underlying type info for hidden data */
typedef struct {
    FILE* fp;
    int open;
    int writing;
    int reading;
    char* filename;
} ch_file_t;

static void file_initData(ch_obj_t* obj)
{
    ch_file_t** file = (ch_file_t**)obj->data;
    *file = NULL;
}

static void file_dup(ch_file_t** dest, ch_file_t** src)
{
    *dest = *src;
}

/* Opening files, pretty much the most important thing ever. */
static ch_value_t file_init(ch_value_t self, long argc, ch_value_t* args)
{
    ch_value_t filename;
    ch_value_t mode = ch_string_new1("r");

    ch_checkArgs(argc, args, 1, 1, 0, &filename, &mode);

    if (CH_FALSEP(ch_isa(filename, ch_cString)))
        ch_raise2(ch_eTypeError,
                  "method File#_init_ expects a String as the filename");

    if (CH_FALSEP(ch_isa(mode, ch_cString)))
        ch_raise2(ch_eTypeError,
                  "method File#_init_ expects a String as the file mode");

    ch_file_t **file = (ch_file_t**)self.obj->data;

    *file = ch_malloc(sizeof(ch_file_t));

    char* fnameString = CH_CSTR(filename);
    char* modeString = CH_CSTR(mode);
    (*file)->fp = fopen(fnameString, modeString);
    if (!(*file)->fp)
        ch_raise2(ch_eFileOpenError, "%s: %s", fnameString, strerror(errno));

    (*file)->open = 1;
    char* plus = strstr(modeString, "+");
    int rw = plus != NULL && plus - modeString <= 2;

    (*file)->writing = modeString[0] == 'w' || modeString[0] == 'a' || rw;
    (*file)->reading = modeString[0] == 'r' || rw;
    char* tmpFilename = ch_strdup(fnameString);
    (*file)->filename = tmpFilename;

    return kNil;
}

static ch_value_t file_print(ch_value_t self, long argc, ch_value_t* args)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;

    if (!(*file)->open)
        ch_raise2(ch_eFileModeError, "attempt to write to closed file");

    if (!(*file)->writing)
        ch_raise2(ch_eFileModeError,
                  "attempt to write to a file not opened for writing");

    int i;
    for (i = 0; i < argc; i++) {
        ch_value_t str = CH_SEND(args[i], to_s);
        char *cstr = CH_CSTR(str);
        fwrite(cstr, 1, ((ch_string_t*)str.obj->data)->size, (*file)->fp);
    }

    return kNil;
}

/* exactly the same as print, but if the last character output was not
 * a newline, this appends one. */
static ch_value_t file_puts(ch_value_t self, long argc, ch_value_t* args)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;

    if (!(*file)->open)
        ch_raise2(ch_eFileModeError, "attempt to write to closed file");

    if (!(*file)->writing)
        ch_raise2(ch_eFileModeError,
                  "attempt to write to a file not opened for writing");

    int i;
    char lastChar = '\0';
    for(i = 0; i < argc; i++) {
        ch_value_t str = CH_SEND(args[i], to_s);
        char* cstr = CH_CSTR(str);
        long len = ((ch_string_t*)str.obj->data)->size;
        lastChar = cstr[len - 1];
        fwrite(cstr, 1, len, (*file)->fp);
    }
    if(lastChar != '\n') fputs("\n", (*file)->fp);

    return kNil;
}

/* Read a line of input. */
static ch_value_t file_gets_noCheck(ch_file_t** file)
{
    ch_value_t result;
    char buf[1024];

    buf[sizeof(buf) - 1] = '\0';
    buf[sizeof(buf) - 2] = '\n';

    CH_STATIC_ASSERT(EINTR != 0);
    errno = 0;
    while (!fgets(buf, sizeof (buf), (*file)->fp)) {
        if (errno != EINTR)
            return kNil;

        ch_check_signals();
        errno = 0;
    }

    result = ch_string_new1(buf);

    while (buf[sizeof(buf) - 2] != '\n') {
        buf[sizeof(buf) - 2] = '\n';
        errno = 0;
        while (!fgets(buf, sizeof(buf), (*file)->fp)) {
            if (errno != EINTR)
                fatal("weird fgets() error");

            ch_check_signals();
            errno = 0;
        }

        ch_value_t str = ch_string_new1(buf);
        CH_SEND(result, concat$b, str);
    }

    return result;
}

static ch_value_t file_gets(ch_value_t self)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;

    if (!(*file)->open)
        ch_raise2(ch_eFileModeError, "attempt to read from closed file");

    if (!(*file)->reading)
        ch_raise2(ch_eFileModeError,
                  "attempt to read from a file not opened for reading");

    return file_gets_noCheck(file);
}

/* Read the entire file */
static ch_value_t file_read(ch_value_t self)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;

    if (!(*file)->open)
        ch_raise2(ch_eFileModeError, "attempt to read from closed file");

    if (!(*file)->reading)
        ch_raise2(ch_eFileModeError,
                  "attempt to read from a file not opened for reading");

    ch_value_t result = ch_string_new();
    ch_value_t line = file_gets_noCheck(file);
    while (line.type != CH_NIL) {
        CH_SEND(result, concat$b, line);
        line = file_gets_noCheck(file);
    }
    return result;
}

/* Flush the buffers */
static ch_value_t file_flush(ch_value_t self)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;

    if (!(*file)->open)
        ch_raise2(ch_eFileModeError, "attempt to flush closed file");

    fflush((*file)->fp);
    return kNil;
}

/* I wonder? */
static ch_value_t file_close(ch_value_t self)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;

    if ((*file)->open) {
        fclose((*file)->fp);
        (*file)->open = 0;
    }

    return kNil;
}

/* This returns a string that looks like #<File filename> or
 * #<File[closed] filename> */
static ch_value_t file_inspect(ch_value_t self)
{
    ch_file_t** file = (ch_file_t**)self.obj->data;
    ch_value_t result = ch_string_new1("#<");

    ch_value_t className;
    if(ch_getSlot(ch_classOf(self), SYM_name, &className) < 0)
        className = ch_string_new1("[unnamed]");

    CH_SEND(result, concat$b, className);

    ch_value_t tmp;

    if (!(*file)->open) {
        tmp = ch_string_new1("[closed]");
        CH_SEND(result, concat$b, tmp);
    }

    tmp = ch_string_new1(" ");
    CH_SEND(result, concat$b, tmp);

    if ((*file)->filename) {
        tmp = ch_string_new1((*file)->filename);
    } else {
        uintptr_t ptrInt = (uintptr_t)self.obj;
        char s[64];
        snprintf(s, sizeof(s), "0x%x", (unsigned int)ptrInt);
        tmp = ch_string_new1(s);
    }
    CH_SEND(result, concat$b, tmp);

    tmp = ch_string_new1(">");
    CH_SEND(result, concat$b, tmp);

    return result;
}

/* Set it all up */
void ch_File_init()
{
    ch_globalRoot(&ch_cFile.obj);
    ch_cFile = ch_class_create(SYM_File, ch_cObject);
    ch_eFileOpenError = ch_class_create(SYM_FileOpenError, ch_eException);

    unsigned int fileRepr = ch_genRepr(sizeof(ch_file_t*),
                                       file_initData,
                                       (ch_dupFn_t)file_dup);
    ((ch_class_t*)ch_cFile.obj->data)->reprID = fileRepr;

    /* Define methods */
    ch_defineMethod(ch_cFile, SYM__init_, -2, (ch_cfn_t)file_init);
    ch_defineMethod(ch_cFile, SYM_print, -2, (ch_cfn_t)file_print);
    ch_defineMethod(ch_cFile, SYM_puts, -2, (ch_cfn_t)file_puts);
    ch_defineMethod(ch_cFile, SYM_read, 0, (ch_cfn_t)file_read);
    ch_defineMethod(ch_cFile, SYM_gets, 0, (ch_cfn_t)file_gets);
    ch_defineMethod(ch_cFile, SYM_flush, 0, (ch_cfn_t)file_flush);
    ch_defineMethod(ch_cFile, SYM_close, 0, (ch_cfn_t)file_close);
    ch_defineMethod(ch_cFile, SYM_inspect_raw, 0, (ch_cfn_t)file_inspect);

    /* Stdin, Stdout, Stderr */
    ch_globalRoot(&ch_stdin.obj);
    ch_globalRoot(&ch_stdout.obj);
    ch_globalRoot(&ch_stderr.obj);
    ch_stdin = ch_file_new(stdin, 1, 0, NULL);
    ch_stdout = ch_file_new(stdout, 0, 1, NULL);
    ch_stderr = ch_file_new(stderr, 0, 1, NULL);
}

/* Create a file object given an existing open FILE pointer. Closing the
 * file becomes the responsibility of the file object. */
ch_value_t ch_file_new(FILE* fp, int reading, int writing, const char* name)
{
    ch_file_t* file = ch_malloc(sizeof(ch_file_t));

    file->fp = fp;
    file->open = 1;
    file->writing = writing;
    file->reading = reading;
    file->filename = name ? ch_strdup(name) : NULL;

    ch_value_t result = ch_obj_create(ch_cFile);
    *((ch_file_t**)result.obj->data) = file;

    return result;
}
