/* Float.c -- implementation of chai Float class
   September 30, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "value.h"
#include "obj.h"
#include "assorted.h"
#include "Exception.h"
#include "sym_literals.h"
#include <string.h>
#include <stdio.h>
#include <math.h>

#define CHECK_TYPE(var, meth)                                           \
    do {                                                                \
        int _0type = (var).type;                                        \
        if(_0type != CH_INT && _0type != CH_FLOAT)                      \
            ch_raise2(                                                  \
                ch_eTypeError,                                          \
                "method " meth " expected an int or float argument"     \
                );                                                      \
    } while(0)

#define CHECK_TYPE_DIV(var, meth)                                       \
    do {                                                                \
        int _0type = (var).type;                                        \
        if (_0type == CH_INT) {                                         \
            if ((var).i == 0)                                           \
                ch_raise2(ch_eDivideByZeroError, "");                   \
        } else if (_0type == CH_FLOAT) {                                \
            if ((var).f == 0.0)                                         \
                ch_raise2(ch_eDivideByZeroError, "");                   \
        } else                                                          \
            ch_raise2(                                                  \
                ch_eTypeError,                                          \
                "method " meth " expected an int or float argument"     \
                );                                                      \
    } while(0)

/* a and b should be variables! */
#define FLOAT_OP(a,b, op)                       \
    (b.type == CH_INT ? ch_float(a.f op b.i) : ch_float(a.f op b.f))

static ch_value_t float_add(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#+");
    return FLOAT_OP(self, b, +);
}

static ch_value_t float_sub(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#-");
    return FLOAT_OP(self, b, -);
}

static ch_value_t float_mul(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#*");
    return FLOAT_OP(self, b, *);
}

static ch_value_t float_div(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE_DIV(b, "Float#/");
    return FLOAT_OP(self, b, /);
}

static ch_value_t float_idiv(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE_DIV(b, "Float#//");
    return ch_fix((int)(b.type == CH_INT
                        ? self.f / b.i
                        : self.f / b.f));
}

static ch_value_t float_mod(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE_DIV(b, "Float#%");
    float result = b.type == CH_INT
        ? self.f - (int)(self.f / b.i) * b.i
        : self.f - (int)(self.f / b.f) * b.f;

    if(result < 0) { return ch_float(b.type == CH_INT
                                     ? b.i + result
                                     : b.f + result); }
    return ch_float(result);
}

static ch_value_t float_pow(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#**");
    return ch_float(b.type == CH_FLOAT
                    ? pow(self.f, b.f)
                    : pow(self.f, (float)b.i));
}

static ch_value_t float_eq(ch_value_t self, ch_value_t b)
{
    if(b.type != CH_FLOAT && b.type != CH_INT) { return kFalse; }
    return (b.type == CH_FLOAT ? self.f == b.f : self.f == b.i)
        ? kTrue : kFalse;
}

static ch_value_t float_lt(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#<");
    return (b.type == CH_FLOAT ? self.f < b.f : self.f < b.i)
        ? kTrue : kFalse;
}

static ch_value_t float_le(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#<=");
    return (b.type == CH_FLOAT ? self.f <= b.f : self.f <= b.i)
        ? kTrue : kFalse;
}

static ch_value_t float_gt(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#>");
    return (b.type == CH_FLOAT ? self.f > b.f : self.f > b.i)
        ? kTrue : kFalse;
}

static ch_value_t float_ge(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Float#>=");
    return (b.type == CH_FLOAT ? self.f >= b.f : self.f >= b.i)
        ? kTrue : kFalse;
}

static ch_value_t float_toFloat(ch_value_t self)
{
    return self;
}

static ch_value_t float_toInt(ch_value_t self)
{
    return ch_fix((long)self.f);
}

static ch_value_t float_inspect(ch_value_t self)
{
    char str[32];
    snprintf(str, sizeof(str), "%f", self.f);

    size_t last = strlen(str) - 1;
    while(last > 0 && str[last] == '0' && str[last - 1] != '.')
        str[last--] = '\0';

    return ch_string_new1(str);
}

void ch_defineFloatMethods()
{
    ch_defineMethod(ch_cFloat, SYM_$p, 1, (ch_cfn_t)float_add);    //   +
    ch_defineMethod(ch_cFloat, SYM_$m, 1, (ch_cfn_t)float_sub);    //   -
    ch_defineMethod(ch_cFloat, SYM_$t, 1, (ch_cfn_t)float_mul);    //   *
    ch_defineMethod(ch_cFloat, SYM_$d, 1, (ch_cfn_t)float_div);    //   /
    ch_defineMethod(ch_cFloat, SYM_$d$d, 1, (ch_cfn_t)float_idiv); //  //
    ch_defineMethod(ch_cFloat, SYM_$r, 1, (ch_cfn_t)float_mod);    //   %
    ch_defineMethod(ch_cFloat, SYM_$t$t, 1, (ch_cfn_t)float_pow);  //  **
    ch_defineMethod(ch_cFloat, SYM_$e$e, 1, (ch_cfn_t)float_eq);   //  ==
    ch_defineMethod(ch_cFloat, SYM_$l, 1, (ch_cfn_t)float_lt);     //   <
    ch_defineMethod(ch_cFloat, SYM_$l$e, 1, (ch_cfn_t)float_le);   //  <=
    ch_defineMethod(ch_cFloat, SYM_$g, 1, (ch_cfn_t)float_gt);     //   >
    ch_defineMethod(ch_cFloat, SYM_$g$e, 1, (ch_cfn_t)float_ge);   //  >=
    ch_defineMethod(ch_cFloat, SYM_to_i, 0, (ch_cfn_t)float_toInt);
    ch_defineMethod(ch_cFloat, SYM_to_f, 0, (ch_cfn_t)float_toFloat);
    ch_defineMethod(ch_cFloat, SYM_inspect_raw, 0, (ch_cfn_t)float_inspect);
}
