/* Integer.c -- implementation of chai Integer class
   September 30, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "value.h"
#include "obj.h"
#include "assorted.h"
#include "Exception.h"
#include "sym_literals.h"
#include <stdio.h>
#include <math.h>
#include <limits.h>

#define CHECK_TYPE(var, meth)                                           \
    do {                                                                \
        int _0type = (var).type;                                        \
        if(_0type != CH_INT && _0type != CH_FLOAT)                      \
            ch_raise2(                                                  \
                ch_eTypeError,                                          \
                "method " meth " expected an int or float argument");   \
    } while(0)

#define CHECK_TYPE_DIV(var, meth)                                       \
    do {                                                                \
        int _0type = (var).type;                                        \
        if (_0type == CH_INT) {                                         \
            if ((var).i == 0)                                           \
                ch_raise2(ch_eDivideByZeroError, "");                   \
        } else if (_0type == CH_FLOAT) {                                \
            if ((var).f == 0.0)                                         \
                ch_raise2(ch_eDivideByZeroError, "");                   \
        } else                                                          \
            ch_raise2(                                                  \
                ch_eTypeError,                                          \
                "method " meth " expected an int or float argument"     \
                );                                                      \
    } while(0)

/* NB. Use a variable for each of a and b. They will both be multiple
 * evaluated. */
#define INT_OP(a,b,op)                                                  \
    (b.type == CH_INT ? ch_fix(a.i op b.i) : ch_float(a.i op b.f))

static ch_value_t int_add(ch_value_t self, ch_value_t b)
{
    if (b.type == CH_FLOAT)
        return ch_float(self.i + b.f);
    else if (b.type == CH_INT) {
        if (b.i < 0 && self.i < LONG_MIN - b.i)
            ch_raise2(ch_eException, "integer underflow");
        else if (b.i > 0 && self.i > LONG_MAX - b.i)
            ch_raise2(ch_eException, "integer overflow");

        return ch_fix(self.i + b.i);
    }

    ch_raise2(ch_eTypeError, "expected a number");
}

static ch_value_t int_succ(ch_value_t self)
{
    return int_add(self, ch_fix(1));
}

static ch_value_t int_sub(ch_value_t self, ch_value_t b)
{
    if (b.type == CH_FLOAT)
        return ch_float(self.i - b.f);
    else if (b.type == CH_INT) {
        if (b.i > 0 && self.i < LONG_MIN + b.i)
            ch_raise2(ch_eException, "integer underflow");
        else if (b.i < 0 && self.i > LONG_MAX + b.i)
            ch_raise2(ch_eException, "integer overflow");

        return ch_fix(self.i - b.i);
    }

    ch_raise2(ch_eTypeError, "expected a number");
}

static ch_value_t int_mul(ch_value_t self, ch_value_t b)
{
    if (b.type == CH_FLOAT)
        return ch_float(self.i * b.f);
    else if (b.type == CH_INT) {
        //if (sizeof (long) != 4)
        //    fatal("multiplication with overflow check on non-32bit");

        /* 32-bit machine yay */
        long long result = (long long)self.i * b.i;
        if (result < LONG_MIN)
            ch_raise2(ch_eException, "integer underflow");
        else if (result > LONG_MAX)
            ch_raise2(ch_eException, "integer overflow");

        return ch_fix(result);
    }

    CHECK_TYPE(b, "Fixnum#*");
    return INT_OP(self, b, *);
}

static ch_value_t int_div(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE_DIV(b, "Fixnum#/");
    float result =
        (float)self.i / (b.type == CH_INT ? (float)b.i : b.f);
    if(result == floor(result)) { return ch_fix((long)result); }
    return ch_float(result);
}

static ch_value_t int_idiv(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE_DIV(b, "Fixnum#//");
    return ch_fix(b.type == CH_INT
                  ? self.i / b.i
                  : (int)(self.i / b.f));
}

static ch_value_t int_mod(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE_DIV(b, "Fixnum#%");
    float result = b.type == CH_INT
        ? (float)(self.i - self.i / b.i * b.i)
        : self.i - (int)(self.i / b.f) * b.f;

    if(result < 0) {
        result += b.type == CH_INT ? (float)b.i : b.f;
    }

    if(result == floor(result)) { return ch_fix(result); }
    return ch_float(result);
}

static ch_value_t int_pow(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Fixnum#**");
    return (b.type == CH_INT
            ? ch_fix((int)pow((float)self.i, (float)b.i))
            : ch_float(pow((float)self.i, b.f)));
}

static ch_value_t int_toInt(ch_value_t self)
{
    return self;
}

static ch_value_t int_toFloat(ch_value_t self)
{
    return ch_float((float)self.i);
}

static ch_value_t int_eq(ch_value_t self, ch_value_t b)
{
    if(b.type != CH_INT && b.type != CH_FLOAT) { return kFalse; }
    return (b.type == CH_INT ? self.i == b.i : self.i == b.f)
        ? kTrue : kFalse;
}

static ch_value_t int_lt(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Fixnum#<");
    return (b.type == CH_INT ? self.i < b.i : self.i < b.f)
        ? kTrue : kFalse;
}

static ch_value_t int_le(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Fixnum#<=");
    return (b.type == CH_INT ? self.i <= b.i : self.i <= b.f)
        ? kTrue : kFalse;
}

static ch_value_t int_gt(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Fixnum#>");
    return (b.type == CH_INT ? self.i > b.i : self.i > b.f)
        ? kTrue : kFalse;
}

static ch_value_t int_ge(ch_value_t self, ch_value_t b)
{
    CHECK_TYPE(b, "Fixnum#>=");
    return (b.type == CH_INT ? self.i >= b.i : self.i >= b.f)
        ? kTrue : kFalse;
}

static ch_value_t int_inspect(ch_value_t self)
{
    char str[32]; /* Should be plenty. */

    snprintf(str, sizeof(str), "%ld", self.i);
    return ch_string_new1(str);
}

#define BITWISE(methodname, op)                                         \
    static ch_value_t int_##methodname(ch_value_t self, ch_value_t b)   \
    {                                                                   \
        if(b.type != CH_INT) {                                          \
            ch_raise2(                                                  \
                ch_eTypeError,                                          \
                "method Fixnum#" #op " expects an "                     \
                "integer argument");                                    \
        }                                                               \
        return ch_fix(self.i op b.i);                                   \
    }
BITWISE(bitand, &);
BITWISE(bitor, |);
BITWISE(bitxor, ^);
BITWISE(lshift, <<);
BITWISE(rshift, >>);

static ch_value_t int_bitnot(ch_value_t self)
{
    return ch_fix(~self.i);
}

static ch_value_t int_dotdot(ch_value_t a, ch_value_t b)
{
    return ch_funcall(ch_cIntRange, 2, a, b);
}

void ch_defineIntegerMethods()
{
    ch_defineMethod(ch_cFixnum, SYM_$p, 1, (ch_cfn_t)int_add);    //   +
    ch_defineMethod(ch_cFixnum, SYM_$m, 1, (ch_cfn_t)int_sub);    //   -
    ch_defineMethod(ch_cFixnum, SYM_$t, 1, (ch_cfn_t)int_mul);    //   *
    ch_defineMethod(ch_cFixnum, SYM_$d, 1, (ch_cfn_t)int_div);    //   /
    ch_defineMethod(ch_cFixnum, SYM_$d$d, 1, (ch_cfn_t)int_idiv); //  //
    ch_defineMethod(ch_cFixnum, SYM_$r, 1, (ch_cfn_t)int_mod);    //   %
    ch_defineMethod(ch_cFixnum, SYM_$t$t, 1, (ch_cfn_t)int_pow);  //  **
    ch_defineMethod(ch_cFixnum, SYM_$e$e, 1, (ch_cfn_t)int_eq);   //  ==
    ch_defineMethod(ch_cFixnum, SYM_$l, 1, (ch_cfn_t)int_lt);     //   <
    ch_defineMethod(ch_cFixnum, SYM_$l$e, 1, (ch_cfn_t)int_le);   //  <=
    ch_defineMethod(ch_cFixnum, SYM_$g, 1, (ch_cfn_t)int_gt);     //   >
    ch_defineMethod(ch_cFixnum, SYM_$g$e, 1, (ch_cfn_t)int_ge);   //  >=
    ch_defineMethod(ch_cFixnum, SYM_$a, 1, (ch_cfn_t)int_bitand); //   &
    ch_defineMethod(ch_cFixnum, SYM_$o, 1, (ch_cfn_t)int_bitor);  //   |
    ch_defineMethod(ch_cFixnum, SYM_$x, 1, (ch_cfn_t)int_bitxor); //   ^
    ch_defineMethod(ch_cFixnum, SYM_$l$l, 1, (ch_cfn_t)int_lshift); //   <<
    ch_defineMethod(ch_cFixnum, SYM_$g$g, 1, (ch_cfn_t)int_rshift); //   >>
    ch_defineMethod(ch_cFixnum, SYM_$n, 0, (ch_cfn_t)int_bitnot);   //    ~
    ch_defineMethod(ch_cFixnum, SYM_succ, 0, (ch_cfn_t)int_succ);
    ch_defineMethod(ch_cFixnum, SYM_to_i, 0, (ch_cfn_t)int_toInt);
    ch_defineMethod(ch_cFixnum, SYM_to_f, 0, (ch_cfn_t)int_toFloat);
    ch_defineMethod(ch_cFixnum, SYM_inspect_raw, 0, (ch_cfn_t)int_inspect);
    ch_defineMethod(ch_cFixnum, SYM_$D$D, 1, (ch_cfn_t)int_dotdot);
}
