/* List.c -- implementation of chai List class
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "assorted.h"
#include "mem.h"
#include "obj.h"
#include "misc.h"
#include "Exception.h"
#include "sym_literals.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Static resizing functions. */
static void list_grow(ch_list_t *ary)
{
    ary->cap *= 2;
    ary->elems = ch_realloc(ary->elems, ary->cap * sizeof(ch_value_t));
}

static void list_shrink(ch_list_t *ary)
{
    ary->cap /= 2;

    if(ary->cap < CH_DEFAULT_LIST_CAP)
        ary->cap = CH_DEFAULT_LIST_CAP;

    ary->elems = ch_realloc(ary->elems, ary->cap * sizeof(ch_value_t));
}

/* Is it empty? Cuh-razy! */
static ch_value_t list_emptyp(ch_value_t self)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    return list->size == 0 ? kTrue : kFalse;
}

static ch_value_t list_size(ch_value_t self)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    return ch_fix(list->size);
}

/* Resizing */
static ch_value_t list_resizeb(ch_value_t self, ch_value_t newSize)
{
    ch_list_t *list;

    if (newSize.type != CH_INT || newSize.i < 0)
        ch_raise2(ch_eTypeError,
                  "method List.resize! expects a non-negative integer");

    list = (ch_list_t*)self.obj->data;
    if (list->size == newSize.i)
        return kFalse;

    if (newSize.i < list->size) {
        while (newSize.i * 4 < list->cap &&
               list->cap != CH_DEFAULT_LIST_CAP)
            list_shrink(list);
    } else {
        while (newSize.i > list->cap)
            list_grow(list);
    }

    list->size = newSize.i;
    return kTrue;
}

/* Random access. */
static ch_value_t list_set(ch_value_t self, ch_value_t index, ch_value_t val)
{
    if (index.type != CH_INT)
        ch_raise2(ch_eTypeError, "method List#set expects an integer index");

    ch_list_t* list = (ch_list_t*)self.obj->data;

    int i = index.i >= 0 ? index.i : list->size + index.i;
    if (i < 0 || i >= list->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", i);

    list->elems[i] = val;
    return kNil;
}

static ch_value_t list_get(ch_value_t self, ch_value_t index)
{
    if (index.type != CH_INT)
        ch_raise2(ch_eTypeError, "method List#get expects an integer index");

    ch_list_t* list = (ch_list_t*)self.obj->data;

    int i = index.i >= 0 ? index.i : list->size + index.i;
    if (i < 0 || i >= list->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", i);

    return list->elems[i];
}

/* Pretend it's a stack. :D */
static ch_value_t list_push(ch_value_t self, ch_value_t val)
{
    long size = list_size(self).i;
    list_resizeb(self, ch_fix(size + 1));
    return list_set(self, ch_fix(size), val);
}

static ch_value_t list_pop(ch_value_t self)
{
    long size = list_size(self).i;

    if (size == 0)
        return kNil;

    ch_value_t newSize = ch_fix(size - 1);
    ch_value_t result = list_get(self, newSize);
    list_resizeb(self, newSize);

    return result;
}

static ch_value_t list_unshift(ch_value_t self, ch_value_t val)
{
    ch_list_t *selflist = (ch_list_t*)self.obj->data;
    long size = selflist->size;
    list_resizeb(self, ch_fix(size + 1));
    memmove(&selflist->elems[1], selflist->elems, size * sizeof (ch_value_t));
    selflist->elems[0] = val;
    return kNil;
}

static ch_value_t list_join(ch_value_t self, long argc, ch_value_t *args)
{
    ch_value_t sep = ch_string_new1("");
    ch_checkArgs(argc, args, 0, 1, 0, &sep);

    ch_value_t result = ch_string_new1("");

    int i;
    ch_list_t* list = (ch_list_t*)self.obj->data;
    for (i = 0; i < list->size; i++) {
        if (i)
            CH_SEND(result, concat$b, sep);

        CH_SEND(result, concat$b, list->elems[i]);
    }

    return result;
}

static ch_value_t list_concatb(ch_value_t self, ch_value_t b)
{
    ch_list_t *list_a, *list_b;
    long old_size;

    list_a = (ch_list_t*)self.obj->data;

    if (CH_FALSEP(ch_isa(b, ch_cList)))
        ch_raise2(ch_eTypeError,
                  "method List.concat! expects a List argument");

    list_b = (ch_list_t*)b.obj->data;

    old_size = list_a->size;

    list_resizeb(self, ch_fix(list_a->size + list_b->size));

    memcpy(&list_a->elems[old_size],
           list_b->elems,
           list_b->size * sizeof(ch_value_t));

    return self;
}

static ch_value_t list_concat(ch_value_t self, ch_value_t b)
{
    ch_value_t result = ch_obj_dup(self);
    CH_SEND(result, concat$b, b);
    return result;
}

static ch_value_t list_toString(ch_value_t self)
{
    return CH_SEND(self, join);
}

static ch_value_t list_inspect(ch_value_t self)
{
    ch_value_t result = ch_string_new1("[");
    ch_value_t comma = ch_string_new1(", ");

    ch_list_t* list = (ch_list_t*)self.obj->data;
    int i;
    for (i = 0; i < list->size; i++) {
        ch_value_t inspected = CH_SEND(list->elems[i], inspect);

        if (i)
            CH_SEND(result, concat$b, comma);

        CH_SEND(result, concat$b, inspected);
    }

    ch_value_t rightParen = ch_string_new1("]");
    CH_SEND(result, concat$b, rightParen);

    return result;
}

static ch_value_t list_shift(ch_value_t self)
{
    ch_list_t *list;
    ch_value_t result;

    list = (ch_list_t*)self.obj->data;

    if (list->size == 0)
        return kNil;

    result = list->elems[0];

    memmove(list->elems,
            &list->elems[1],
            (list->size - 1) * sizeof(ch_value_t));

    list->elems[list->size - 1] = kNil;
    list_resizeb(self, ch_fix(list->size - 1));

    return result;
}

static ch_value_t list_each(ch_value_t self, ch_value_t f)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;

    int i;
    for(i = 0; i < list->size; i++) {
        ch_funcall(f, 1, list->elems[i]);
    }

    return kNil;
}

static ch_value_t list_map(ch_value_t self, ch_value_t f)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    ch_value_t result = ch_list_new();
    ch_list_t* resultList = (ch_list_t*)result.obj->data;

    /* Set its size properly in the first place. */
    list_resizeb(result, ch_fix(list->size));

    int i;
    for(i = 0; i < list->size; i++) {
        ch_value_t val = ch_funcall(f, 1, list->elems[i]);
        resultList->elems[i] = val;
    }

    return result;
}

static ch_value_t list_select(ch_value_t self, ch_value_t f)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    ch_value_t result = ch_list_new();

    int i;
    for(i = 0; i < list->size; i++) {
        ch_value_t pred = ch_funcall(f, 1, list->elems[i]);
        if(CH_TRUEP(pred)) {
            list_push(result, list->elems[i]);
        }
    }

    return result;
}

static ch_value_t list_reject(ch_value_t self, ch_value_t f)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    ch_value_t result = ch_list_new();

    int i;
    for(i = 0; i < list->size; i++) {
        ch_value_t pred = ch_funcall(f, 1, list->elems[i]);
        if(CH_FALSEP(pred)) {
            list_push(result, list->elems[i]);
        }
    }

    return result;
}

static ch_value_t list_reduce(ch_value_t self, ch_value_t id, ch_value_t f)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    ch_value_t result = id;

    int i;
    for(i = 0; i < list->size; i++) {
        ch_value_t newResult = ch_funcall(f, 2, result, list->elems[i]);
        result = newResult;
    }

    return result;
}

static ch_value_t list_reverseb(ch_value_t self)
{
    ch_list_t* list = (ch_list_t*)self.obj->data;
    int i;
    for(i = 0; i < list->size / 2; i++) {
        ch_value_t tmp = list->elems[i];
        list->elems[i] = list->elems[list->size - 1 - i];
        list->elems[list->size - 1 - i] = tmp;
    }
    return self;
}

static ch_value_t list_reverse(ch_value_t self)
{
    ch_value_t result = ch_obj_dup(self);
    list_reverseb(result);
    return result;
}

static ch_value_t list_toList(ch_value_t self)
{
    return self;
}

static void list_sort_b_help(ch_value_t self, int start, int end)
{
    if(start >= end) { return; }

    int pivIndex = (start + end) / 2;
    ch_value_t piv = list_get(self, ch_fix(pivIndex));
    list_set(self, ch_fix(pivIndex), list_get(self, ch_fix(start)));
    list_set(self, ch_fix(start), piv);

    int begindex = start;
    int endex = end;

    while(begindex < endex)
    {
        while(begindex < endex &&
              CH_TRUEP(CH_SEND(list_get(self, ch_fix(begindex)),
                               $l$e, // <=
                               piv)))
            begindex++;

        while(begindex < endex &&
              CH_TRUEP(CH_SEND(list_get(self, ch_fix(endex)),
                               $g$e, // >=
                               piv)))
            endex--;

        if(begindex < endex) {
            ch_value_t tmp = list_get(self, ch_fix(begindex));
            list_set(self, ch_fix(begindex), list_get(self, ch_fix(endex)));
            list_set(self, ch_fix(endex), tmp);
        }
    }

    if(CH_TRUEP(CH_SEND(list_get(self, ch_fix(begindex)),
                        $g$e, // >=
                        piv)))
        begindex--;


    ch_value_t tmp = list_get(self, ch_fix(start));
    list_set(self, ch_fix(start), list_get(self, ch_fix(begindex)));
    list_set(self, ch_fix(begindex), tmp);

    list_sort_b_help(self, start, begindex - 1);
    list_sort_b_help(self, begindex + 1, end);
}

static ch_value_t list_sort_b(ch_value_t self)
{
    list_sort_b_help(self, 0, list_size(self).i - 1);
    return self;
}

static ch_value_t list_sort(ch_value_t self)
{
    ch_value_t dupped = ch_obj_dup(self);
    list_sort_b(dupped);
    return dupped;
}

/* Initialization */
static ch_value_t return_nil(ch_value_t ignore)
{
    return kNil;
}

static ch_value_t list_init(ch_value_t self, long argc, ch_value_t *args)
{
    static ch_value_t retnil = { CH_NIL };
    ch_value_t n, f, i;

    if (retnil.type == CH_NIL)
        retnil = ch_prim_new(1, (ch_cfn_t)return_nil);

    n = ch_fix(0);
    f = retnil;
    ch_checkArgs(argc, args, 0, 2, 0, &n, &f);

    if (n.type != CH_INT)
        ch_raise2(ch_eTypeError,
                  "List._init_ expects an integer as the first argument");

    for (i = ch_fix(0); i.i < n.i; i.i++)
        list_push(self, ch_vfuncall(f, 1, &i));

    return kNil;
}

static ch_value_t list_slice(ch_value_t self, ch_value_t start, ch_value_t end)
{
    ch_list_t *list = (ch_list_t*)self.obj->data;
    long istart, iend;

    if (start.type != CH_INT || end.type != CH_INT)
        ch_raise2(ch_eTypeError, "expected integer indices");

    istart = start.i + (start.i < 0 ? list->size : 0);
    iend = end.i + (end.i < 0 ? list->size : 0);

    if (istart > iend)
        return ch_list_new();

    if (istart < 0 || istart >= list->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", start.i);
    else if (iend < 0 || iend >= list->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", end.i);

    return ch_list_new2(iend - istart + 1, &list->elems[istart]);
}

static ch_value_t list_delete_b(ch_value_t self, ch_value_t x)
{
    ch_list_t *l = (ch_list_t*)self.obj->data;
    int i;

    for (i = 0; i < l->size; i++)
        if (CH_TRUEP(CH_SEND(x, $e$e, l->elems[i])))
            break;

    if (i == l->size)
        return kFalse;

    for (; i < l->size - 1; i++)
        l->elems[i] = l->elems[i + 1];

    l->size--;
    return kTrue;
}

static ch_value_t list_without(ch_value_t self, ch_value_t x)
{
    ch_value_t result = ch_obj_dup(self);
    list_delete_b(result, x);
    return result;
}

/* Define the methods */
void ch_defineListMethods()
{
    ch_defineMethod(ch_cList, SYM__init_, -2, (ch_cfn_t)list_init);
    ch_defineMethod(ch_cList, SYM_empty$q, 0, (ch_cfn_t)list_emptyp);
    ch_defineMethod(ch_cList, SYM_size, 0, (ch_cfn_t)list_size);
    ch_defineMethod(ch_cList, SYM_resize$b, 1, (ch_cfn_t)list_resizeb);
    ch_defineMethod(ch_cList, SYM_set, 2, (ch_cfn_t)list_set);
    ch_defineMethod(ch_cList, SYM__call_$e, 2, (ch_cfn_t)list_set); // call=
    ch_defineMethod(ch_cList, SYM_get, 1, (ch_cfn_t)list_get);
    ch_defineMethod(ch_cList, SYM__call_, 1, (ch_cfn_t)list_get);
    ch_defineMethod(ch_cList, SYM_slice, 2, (ch_cfn_t)list_slice);
    ch_defineMethod(ch_cList, SYM_push, 1, (ch_cfn_t)list_push);
    ch_defineMethod(ch_cList, SYM_pop, 0, (ch_cfn_t)list_pop);
    ch_defineMethod(ch_cList, SYM_shift, 0, (ch_cfn_t)list_shift);
    ch_defineMethod(ch_cList, SYM_unshift, 1, (ch_cfn_t)list_unshift);
    ch_defineMethod(ch_cList, SYM_join, -2, (ch_cfn_t)list_join);
    ch_defineMethod(ch_cList, SYM_concat$b, 1, (ch_cfn_t)list_concatb);
    ch_defineMethod(ch_cList, SYM_concat, 1, (ch_cfn_t)list_concat);
    ch_defineMethod(ch_cList, SYM_to_s, 0, (ch_cfn_t)list_toString);
    ch_defineMethod(ch_cList, SYM_inspect_raw, 0, (ch_cfn_t)list_inspect);
    ch_defineMethod(ch_cList, SYM_to_l, 0, (ch_cfn_t)list_toList);
    ch_defineMethod(ch_cList, SYM_each, 1, (ch_cfn_t)list_each);
    ch_defineMethod(ch_cList, SYM_map, 1, (ch_cfn_t)list_map);
    ch_defineMethod(ch_cList, SYM_select, 1, (ch_cfn_t)list_select);
    ch_defineMethod(ch_cList, SYM_reject, 1, (ch_cfn_t)list_reject);
    ch_defineMethod(ch_cList,
                    SYM_reverse$b,
                    0,
                    (ch_cfn_t)list_reverseb);
    ch_defineMethod(ch_cList, SYM_reverse, 0, (ch_cfn_t)list_reverse);
    ch_defineMethod(ch_cList, SYM_reduce, 2, (ch_cfn_t)list_reduce);
    ch_defineMethod(ch_cList, SYM_sort$b, 0, (ch_cfn_t)list_sort_b);
    ch_defineMethod(ch_cList, SYM_sort, 0, (ch_cfn_t)list_sort);
    ch_defineMethod(ch_cList, SYM_delete$b, 1, (ch_cfn_t)list_delete_b);
    ch_defineMethod(ch_cList, SYM_without, 1, (ch_cfn_t)list_without);
}
