/* Module.c -- implementation of chai Module class
   October 3, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "Module.h"
#include "Exception.h"
#include "compile.h"
#include "sym.h"
#include "sym_literals.h"

ch_value_t ch_cModule;
ch_ht_t *ch_all_modules;
ch_value_t ch_current_module;

static void init_module(ch_obj_t *obj)
{
    ch_module_t *m = (ch_module_t*)obj->data;
    ch_ht_init(&m->globals, NULL, NULL);
    ch_ht_init(&m->macros, NULL, NULL);
    ch_ht_init(&m->imports, NULL, NULL);
    ch_ht_init(&m->imported_names, NULL, NULL);
    ch_ht_init(&m->prec_table, NULL, NULL);
    ch_ht_init(&m->uprec_table, NULL, NULL);
    ch_ht_init(&m->assoc_table, ch_hashValue, NULL);
    ch_ht_init(&m->op_idents, NULL, NULL);
    m->module_name = kNil;
}

static void dup_module(ch_module_t *dst, ch_module_t *src)
{
    ch_raise2(ch_eTypeError, "Can't dup a Module.");
}

static ch_value_t module_init_method(ch_value_t self)
{
    ch_raise2(ch_eTypeError, "Don't create new Modules. Go die.");
}

void ch_Module_init(void)
{
    ch_class_t *klass;

    ch_all_modules = ch_ht_create(NULL, NULL);
    ch_current_module = kNil;

    ch_cModule = ch_class_create(SYM_Module, ch_cObject);
    klass = (ch_class_t*)ch_cModule.obj->data;
    klass->reprID = ch_genRepr(sizeof (ch_module_t),
                               init_module,
                               (ch_dupFn_t)dup_module);

    ch_defineMethod(ch_cModule, SYM__init_, 0, (ch_cfn_t)module_init_method);
}

void ch_lookup_global_name(ch_value_t m, ch_value_t name,
                           ch_value_t *rm, ch_value_t *rname)
{
    ch_module_t *mod;
    char *strname;
    ch_value_t tmp;
    int i;

    strname = (char*)name.sym + sizeof (long);
    *rm = m;

    for (i = 0; strname[i] && strname[i] != ':'; i++);
    while (strname[i]) {   /* found a colon, reach through module */
        ch_value_t in_module = ch_internSizedString(i, strname);
        mod = (ch_module_t*)rm->obj->data;

        if (ch_ht_get(&mod->imports, in_module, rm) < 0)
            ch_raise2(ch_eNameError, "no module named %.*s in module %s\n",
                      i, strname, mod->module_name.sym + sizeof (long));

        /* look for another */
        strname = &strname[i + 1];
        for (i = 0; strname[i] && strname[i] != ':'; i++);
    }

    if (rm->obj != m.obj)
        *rname = ch_internCString(strname);
    else
        *rname = name;

    mod = (ch_module_t*)rm->obj->data;
    while (!ch_ht_get(&mod->imported_names, *rname, &tmp)) {
        *rm = tmp;
        mod = (ch_module_t*)rm->obj->data;
    }
}

void ch_lookup_macro_name(ch_value_t m, ch_value_t name,
                          ch_value_t *rm, ch_value_t *rname)
{
    if (name.sym == SYM_quote.sym || name.sym == SYM_fn.sym
        || name.sym == SYM_let.sym || name.sym == SYM_slot.sym
        || name.sym == SYM_setq.sym || name.sym == SYM_if.sym
        || name.sym == SYM_while.sym || name.sym == SYM_unwind.sym
        || name.sym == SYM_unwind_protect.sym || name.sym == SYM_not.sym
        || name.sym == SYM_set_slot.sym || name.sym == SYM__progn.sym
        || name.sym == SYM_list.sym || name.sym == SYM_varslot.sym
        || name.sym == SYM_set_varslot.sym || name.sym == SYM_call.sym) {
        /* It's a *really* builtin special form. */
        *rm = ch_m_builtins;
        *rname = name;
    } else {
        ch_lookup_global_name(m, name, rm, rname);
    }
}

ch_value_t ch_load_module(const char *name)
{
    ch_value_t m, name_sym, old_cur_module;
    ch_module_t *m_inside;

    name_sym = ch_internCString(name);
    if (!ch_ht_get(ch_all_modules, name_sym, &m))
        return m;

    m = ch_obj_create(ch_cModule);
    m_inside = (ch_module_t*)m.obj->data;
    m_inside->module_name = name_sym;
    ch_module_import_all_from(m, ch_m_builtins);
    ch_module_import_all_ops_from(m, ch_m_builtins);

    /* Implicit imports */
    ch_module_import(m, ch_m_builtins, SYM__builtins);
    ch_module_import(m, ch_m_root, SYM__r);
    ch_module_import(ch_m_root, m, name_sym);

    old_cur_module = ch_current_module;
    ch_current_module = m;

    ch_ht_set(ch_all_modules, name_sym, m);
    ch_load(name);

    ch_current_module = old_cur_module;
    return m;
}

void ch_module_import(ch_value_t mod, ch_value_t imp, ch_value_t name)
{
    ch_module_t *cur = (ch_module_t*)mod.obj->data;
    ch_ht_set(&cur->imports, name, imp);
}


ch_value_t ch_get_global(ch_value_t mod, ch_value_t name)
{
    ch_module_t *m;
    ch_value_t ref;

    ch_lookup_global_name(mod, name, &mod, &name);
    m = (ch_module_t*)mod.obj->data;

    if (ch_ht_get(&m->globals, name, &ref) < 0)
        ch_raise2(ch_eNameError, "variable %s undefined",
                  name.sym + sizeof (long));

    return *ref.ref;
}

ch_value_t ch_bind_global_(ch_value_t mod, ch_value_t name, ch_value_t *place)
{
    ch_module_t *m;
    ch_value_t ref;

    ch_lookup_global_name(mod, name, &mod, &name);
    m = (ch_module_t*)mod.obj->data;

    ref.type = CH_REF;
    ref.ref = place;
    ch_ht_set(&m->globals, name, ref);

    return ref;
}

void ch_set_global(ch_value_t mod, ch_value_t name, ch_value_t val)
{
    ch_module_t *m;
    ch_value_t ref;

    ch_lookup_global_name(mod, name, &mod, &name);
    m = (ch_module_t*)mod.obj->data;

    if (ch_ht_get(&m->globals, name, &ref) < 0)
        ref = ch_bind_global_(mod, name, ch_malloc(sizeof (ch_value_t)));

    *ref.ref = val;
}

void ch_module_import_name_from(ch_value_t m, ch_value_t name, ch_value_t from)
{
    ch_module_t *mm;

    mm = (ch_module_t*)m.obj->data;
    ch_ht_set(&mm->imported_names, name, from);
}

void ch_module_import_all_from(ch_value_t m, ch_value_t from)
{
    ch_value_t keys_val;
    ch_list_t *keys;
    ch_module_t *mfrom;
    int i;

    mfrom = (ch_module_t*)from.obj->data;

    /* Global variables */
    keys_val = ch_ht_keys(&mfrom->globals);
    keys = (ch_list_t*)keys_val.obj->data;

    for (i = 0; i < keys->size; i++)
        ch_module_import_name_from(m, keys->elems[i], from);

    /* Macros */
    keys_val = ch_ht_keys(&mfrom->macros);
    keys = (ch_list_t*)keys_val.obj->data;

    for (i = 0; i < keys->size; i++)
        ch_module_import_name_from(m, keys->elems[i], from);
}

void ch_module_import_op_from(ch_value_t m, ch_value_t op, ch_value_t from)
{
    ch_module_t *mm, *mfrom;
    ch_value_t prec, assoc, existing;
    char op_macro_name[1024];

    mm = (ch_module_t*)m.obj->data;
    mfrom = (ch_module_t*)from.obj->data;

    if (ch_ht_get(&mfrom->prec_table, op, &prec) < 0)
        return;

    ch_ht_get(&mfrom->assoc_table, prec, &assoc);

    if (!ch_ht_get(&mm->prec_table, op, &existing) && existing.i != prec.i)
        ch_raise2(ch_eParseError,
                  "conflicting precedences for operator %s: %d (existing)"
                  " and %d (imported)",
                  op.sym + sizeof (long),
                  existing.i,
                  prec.i);

    if (!ch_ht_get(&mm->assoc_table, prec, &existing)
        && existing.sym != assoc.sym)

        ch_raise2(ch_eParseError,
                  "conflicting associativities for operator %s: %s (existing)"
                  " and %d (imported)",
                  op.sym + sizeof (long),
                  existing.sym + sizeof (long),
                  assoc.sym + sizeof (long));

    ch_ht_set(&mm->assoc_table, prec, assoc);
    ch_ht_set(&mm->prec_table, op, prec);

    if (!ch_ht_get(&mfrom->op_idents, op, &existing))
        ch_ht_set(&mm->op_idents, op, kTrue);

    snprintf(op_macro_name, sizeof (op_macro_name), "op_%s",
             op.sym + sizeof (long));
    ch_module_import_name_from(m, ch_internCString(op_macro_name), from);
}

void ch_module_import_uop_from(ch_value_t m, ch_value_t op, ch_value_t from)
{
    ch_module_t *mm, *mfrom;
    ch_value_t prec, assoc, existing;
    char op_macro_name[1024];

    mm = (ch_module_t*)m.obj->data;
    mfrom = (ch_module_t*)from.obj->data;

    if (ch_ht_get(&mfrom->uprec_table, op, &prec) < 0)
        return;

    ch_ht_get(&mfrom->assoc_table, prec, &assoc);

    if (!ch_ht_get(&mm->uprec_table, op, &existing) && existing.i != prec.i)
        ch_raise2(ch_eParseError,
                  "conflicting precedences for operator %s: %d (existing)"
                  " and %d (imported)",
                  op.sym + sizeof (long),
                  existing.i,
                  prec.i);

    if (!ch_ht_get(&mm->assoc_table, prec, &existing)
        && existing.sym != assoc.sym)

        ch_raise2(ch_eParseError,
                  "conflicting associativities for operator %s: %s (existing)"
                  " and %s (imported)",
                  op.sym + sizeof (long),
                  existing.sym + sizeof (long),
                  assoc.sym + sizeof (long));

    ch_ht_set(&mm->assoc_table, prec, assoc);
    ch_ht_set(&mm->uprec_table, op, prec);

    if (!ch_ht_get(&mfrom->op_idents, op, &existing))
        ch_ht_set(&mm->op_idents, op, kTrue);

    snprintf(op_macro_name, sizeof (op_macro_name), "uop_%s",
             op.sym + sizeof (long));
    ch_module_import_name_from(m, ch_internCString(op_macro_name), from);
}

void ch_module_import_all_ops_from(ch_value_t m, ch_value_t from)
{
    ch_value_t keys_val;
    ch_list_t *keys;
    ch_module_t *mfrom;
    int i;

    mfrom = (ch_module_t*)from.obj->data;

    /* Binary ops */
    keys_val = ch_ht_keys(&mfrom->prec_table);
    keys = (ch_list_t*)keys_val.obj->data;

    for (i = 0; i < keys->size; i++)
        ch_module_import_op_from(m, keys->elems[i], from);

    /* Unary ops */
    keys_val = ch_ht_keys(&mfrom->uprec_table);
    keys = (ch_list_t*)keys_val.obj->data;

    for (i = 0; i < keys->size; i++)
        ch_module_import_uop_from(m, keys->elems[i], from);
}

ch_value_t ch_module_precedence_of(ch_value_t mod,
                                   ch_value_t op,
                                   ch_value_t unary)
{
    if (CH_FALSEP(ch_isa(mod, ch_cModule)))
        ch_raise2(ch_eTypeError, "expected a module");

    if (op.type != CH_SYMBOL)
        ch_raise2(ch_eTypeError, "expected a symbol operator name");

    ch_module_t *m = (ch_module_t*)mod.obj->data;
    ch_value_t result;

    if (ch_ht_get(CH_FALSEP(unary) ? &m->prec_table : &m->uprec_table,
                  op, &result) < 0)
        ch_raise2(ch_eNameError, "no operator %s in module %s\n",
                  op.sym + sizeof (long),
                  m->module_name.sym + sizeof (long));

    return result;
}

ch_value_t ch_get_module(ch_value_t name)
{
    if (name.type != CH_SYMBOL)
        ch_raise2(ch_eTypeError, "ch_get_module() expects a symbol");

    ch_value_t result;
    ch_ht_get(ch_all_modules, name, &result);
    return result;
}

void ch_inject_module(ch_value_t mod)
{
    if (CH_FALSEP(ch_isa(mod, ch_cModule)))
        ch_raise2(ch_eTypeError, "ch_inject_module() expects a module");

    ch_module_t *m = (ch_module_t*)mod.obj->data;
    ch_ht_set(ch_all_modules, m->module_name, mod);
}
