/* Object.c -- methods of the root class Object
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "obj.h"
#include "value.h"
#include "assorted.h"
#include "assorted.h"
#include "Exception.h"
#include "Module.h"
#include "sym_literals.h"
#include <stdio.h>

/* static */
ch_value_t object_eq(ch_value_t self, ch_value_t b)
{
    return ch_eq(self, b) ? kTrue : kFalse;
}

/* static */
ch_value_t object_hash(ch_value_t self)
{
    return ch_fix(ch_hashCString((char*)&self, sizeof(self)));
}

/* static */
ch_value_t object_nilp(ch_value_t self)
{
    return kFalse;
}

static ch_value_t object_isap(ch_value_t self, ch_value_t klass)
{
    return ch_isa(self, klass);
}

static ch_value_t object_rng(ch_value_t self, ch_value_t other)
{
    return ch_funcall(ch_cRange, 2, self, other);
}

static ch_value_t object_inspect(ch_value_t self)
{
    ch_value_t result = ch_string_new1("#<");
    ch_value_t tmp;

    if (!ch_getSlot(ch_classOf(self), SYM_containing_module, &tmp))
        if (tmp.type == CH_SYMBOL && tmp.sym != ch_current_module_name().sym) {
            CH_SEND(result, concat$b, tmp);
            CH_SEND(result, concat$b, ch_string_new1(":"));
        }

    if (ch_getSlot(ch_classOf(self), SYM_name, &tmp) < 0)
        tmp = ch_string_new1("[unnamed]");

    CH_SEND(result, concat$b, CH_SEND2(tmp, to_s));
    CH_SEND(result, concat$b, ch_string_new1(" "));

    char hex[64];
    snprintf(hex, sizeof (hex), "0x%lx", (unsigned long)self.obj);
    CH_SEND(result, concat$b, ch_string_new1(hex));

    int i;
    for(i = 0; i < self.obj->slots.cap; i++) {
        ch_hashBucket_t* bucket = self.obj->slots.buckets[i];
        while(bucket) {
            CH_SEND(result, concat$b, ch_string_new1(" "));
            CH_SEND(result, concat$b, ch_string_new1(bucket->key.sym + sizeof(long)));
            CH_SEND(result, concat$b, ch_string_new1("="));
            CH_SEND(result, concat$b, CH_SEND2(bucket->value, inspect));
            bucket = bucket->next;
        }
    }

    CH_SEND(result, concat$b, ch_string_new1(">"));

    return result;
}

static ch_value_t object_init(ch_value_t self)
{
    return kNil;
}

static ch_value_t object_toString(ch_value_t self)
{
    return CH_SEND(self, inspect);
}

static ch_value_t object_class(ch_value_t self)
{
    return ch_classOf(self);
}

static ch_value_t object_apply(ch_value_t self, ch_value_t arglist)
{
    ch_value_t asList = CH_SEND(arglist, to_l);
    if (CH_FALSEP(ch_isa(asList, ch_cList)))
        ch_raise2(ch_eTypeError,
                  "method Object#apply expects an argument whose to_l method "
                  "returns a list");

    ch_list_t* l = (ch_list_t*)asList.obj->data;
    return ch_vfuncall(self, l->size, l->elems);
}

static ch_value_t object_send(ch_value_t self, long argc, ch_value_t* args)
{
    if (argc < 1)
        ch_raise2(ch_eArgumentError, "expected at least 1 argument");

    return ch_vsend(self, args[0], argc - 1, &args[1]);
}

static ch_value_t object_methods(ch_value_t self)
{
    ch_value_t klass = ch_classOf(self);
    ch_ht_t set;

    ch_ht_init(&set, NULL, NULL);
    while (klass.obj) {
        ch_class_t *k = (ch_class_t*)klass.obj->data;
        ch_ht_mergeb(&set, &k->methods);
        klass.obj = k->superclass;
    }

    return ch_ht_keys(&set);
}

static ch_value_t object_slots(ch_value_t self)
{
    if (self.type != CH_OBJECT)
        return ch_list_new();

    return ch_ht_keys(&((ch_obj_t*)self.obj)->slots);
}

void ch_defineObjectMethods()
{
    ch_defineMethod(ch_cObject, SYM_$e$e, 1, (ch_cfn_t)object_eq); // ==
    ch_defineMethod(ch_cObject, SYM_eq$q, 1, (ch_cfn_t)object_eq);
    ch_defineMethod(ch_cObject, SYM_$D$D, 1, (ch_cfn_t)object_rng); // ..
    ch_defineMethod(ch_cObject, SYM_hash, 0, (ch_cfn_t)object_hash);
    ch_defineMethod(ch_cObject, SYM_nil$q, 0, (ch_cfn_t)object_nilp);
    ch_defineMethod(ch_cObject, SYM_dup, 0, (ch_cfn_t)ch_obj_dup);
    ch_defineMethod(ch_cObject, SYM_hash, 0, (ch_cfn_t)object_hash);
    ch_defineMethod(ch_cObject, SYM_dup, 0, (ch_cfn_t)ch_obj_dup);
    ch_defineMethod(ch_cObject,
                    SYM_inspect_raw,
                    0,
                    (ch_cfn_t)object_inspect);
    ch_defineMethod(ch_cObject,
                    SYM_to_s,
                    0,
                    (ch_cfn_t)object_toString);
    ch_defineMethod(ch_cObject, SYM__init_, 0, (ch_cfn_t)object_init);
    ch_defineMethod(ch_cObject, SYM_class, 0, (ch_cfn_t)object_class);
    ch_defineMethod(ch_cObject, SYM_apply, 1, (ch_cfn_t)object_apply);
    ch_defineMethod(ch_cObject, SYM_send, -2, (ch_cfn_t)object_send);
    ch_defineMethod(ch_cObject, SYM_is_a$q, 1, (ch_cfn_t)object_isap);
    ch_defineMethod(ch_cObject, SYM_methods, 0, (ch_cfn_t)object_methods);
    ch_defineMethod(ch_cObject, SYM_slots, 0, (ch_cfn_t)object_slots);
    ch_defineMethod(ch_cObject, SYM_responds_to$q, 1, (ch_cfn_t)ch_respondsTop);
}
