/* Range.c -- implementation of chai Regex class
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "assorted.h"
#include "obj.h"
#include "sym_literals.h"
#include <pcre.h>
#include <stdio.h>

ch_value_t ch_cRegex, ch_cRegexMatchData, ch_eRegexCompilationError;

typedef struct {
    pcre *re;
    const char *source;
} ch_re_t;

static void re_initData(ch_obj_t *obj)
{
    ch_re_t *re = (ch_re_t*)obj->data;
    re->re = NULL;
    re->source = NULL;
}

static void re_dup(void *dest, void *src)
{
    ch_re_t *dest_re = (ch_re_t*)dest;
    ch_re_t *src_re = (ch_re_t*)src;

    *dest_re = *src_re;
    if (dest_re->re)
        pcre_refcount(dest_re->re, 1);
}

static void re_destroy(ch_obj_t *obj)
{
    ch_re_t *re = (ch_re_t*)obj->data;
    if (re->re && pcre_refcount(re->re, -1) == 0)
        pcre_free(re->re);

    re->re = NULL;
    re->source = NULL;
}

static ch_value_t re_init(ch_value_t self, ch_value_t source)
{
    if (CH_FALSEP(ch_isa(source, ch_cString)))
        ch_raise2(ch_eTypeError, "expected a string");

    ch_re_t *re = (ch_re_t*)self.obj->data;
    re->source = CH_CSTR(source);

    int erroffset;
    const char *error;
    re->re = pcre_compile(re->source,
                          PCRE_MULTILINE | PCRE_NEWLINE_ANYCRLF | PCRE_BSR_ANYCRLF,
                          &error, &erroffset, NULL);

    if (!re->re)
        ch_raise2(ch_eRegexCompilationError, "%s", error);

    pcre_refcount(re->re, 1);

    return kNil;
}

static ch_value_t re_source(ch_value_t self)
{
    ch_re_t *re = (ch_re_t*)self.obj->data;
    return ch_string_new1(re->source);
}

static ch_value_t re_match(ch_value_t self, long argc, ch_value_t *argv)
{
    ch_value_t str = kNil, start = ch_fix(0);
    ch_checkArgs(argc, argv, 1, 1, 0, &str, &start);

    if (CH_FALSEP(ch_isa(str, ch_cString)))
        ch_raise2(ch_eTypeError, "expected a string");

    ch_re_t *re = (ch_re_t*)self.obj->data;
    ch_string_t *subject_str = (ch_string_t*)str.obj->data;
    ch_string_t subj = *subject_str;

    if (start.type != CH_INT)
        ch_raise2(ch_eTypeError, "expected an integer index");
    else if (start.i < 0 || start.i > subj.size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", start.i);
    else if (start.i == subj.size)
        subj.size = 0;

    if (!re->re)
        ch_raise2(ch_eTypeError, "freak Regex object with no compiled regex");

    int md[33];
    int rc = pcre_exec(re->re, NULL, subj.text, subj.size,
                       start.i, 0, md, 33);

    if (rc < 0)
        return kNil;

    ch_value_t offsets = ch_list_new();
    int i;
    for (i = 0; i < rc * 2; i += 2) {
        CH_SEND(offsets, push, ch_fix(md[i]));
        CH_SEND(offsets, push, ch_fix(md[i+1]));
    }
    return ch_funcall(ch_cRegexMatchData, 2, str, offsets);
}

static ch_value_t md_init(ch_value_t self, ch_value_t s, ch_value_t offs)
{
    ch_setSlot(self, SYM_text, s);
    ch_setSlot(self, SYM_offsets, offs);
    return self;
}

void ch_Regex_init()
{
    ch_cRegex = ch_class_create(SYM_Regex, ch_cObject);
    ch_cRegexMatchData = ch_class_create(SYM_RegexMatchData, ch_cObject);
    ch_eRegexCompilationError = ch_class_create(SYM_RegexCompilationError,
                                                ch_eException);

    ch_class_t *klass = (ch_class_t*)ch_cRegex.obj->data;
    klass->reprID = ch_genRepr(sizeof (ch_re_t), re_initData, re_dup);
    ch_reprDtor(klass->reprID, re_destroy);

    /* Methods */
    ch_defineMethod(ch_cRegex, SYM__init_, 1, (ch_cfn_t)re_init);
    ch_defineMethod(ch_cRegex, SYM_source, 0, (ch_cfn_t)re_source);
    ch_defineMethod(ch_cRegex, SYM_match, -2, (ch_cfn_t)re_match);

    ch_defineMethod(ch_cRegexMatchData, SYM__init_, 2, (ch_cfn_t)md_init);
}
