/* String.c -- implementation of chai String class
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "assorted.h"
#include "mem.h"
#include "value.h"
#include "obj.h"
#include "sym.h"
#include "Exception.h"
#include "sym_literals.h"
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

/* Static resizing functions - adapted from the List code */
static void str_grow(ch_string_t* str)
{
    str->cap *= 2;
    str->text = ch_realloc(str->text, str->cap);
}

static void str_shrink(ch_string_t *str)
{
    str->cap /= 2;

    if (str->cap < CH_DEFAULT_STRING_CAP)
        str->cap = CH_DEFAULT_STRING_CAP;

    str->text = ch_realloc(str->text, str->cap);
}

/* Resizing method - adapted from List code */
static ch_value_t str_resizeb(ch_value_t self, ch_value_t newSize)
{
    if (newSize.type != CH_INT || newSize.i < 0)
        ch_raise2(ch_eTypeError,
                  "method String#resize! expects a non-negative integer");

    ch_string_t* str = (ch_string_t*)self.obj->data;
    if(str->size == newSize.i) { return kFalse; }

    if(newSize.i < str->size) {
        while(newSize.i * 4 < str->cap && str->cap != CH_DEFAULT_STRING_CAP) {
            str_shrink(str);
        }
    } else {
        /* +1 to make room for the terminating NUL that we're keeping
         * for the hell of it. */
        while (newSize.i + 1 >= str->cap)
            str_grow(str);

        memset(&str->text[str->size], ' ', newSize.i - str->size);
        str->text[newSize.i] = '\0';
    }

    str->size = newSize.i;
    return kTrue;
}

static ch_value_t str_emptyp(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    return str->size == 0 ? kTrue : kFalse;
}

static ch_value_t str_size(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    return ch_fix(str->size);
}

static ch_value_t str_toString(ch_value_t self)
{
    return self;
}

static ch_value_t str_toInt(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    return ch_fix(atoi(str->text));
}

static ch_value_t str_toFloat(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    return ch_float(atof(str->text));
}

static ch_value_t str_concatb(ch_value_t self, ch_value_t b)
{
    ch_string_t *str_a, *str_b;
    ch_value_t b_as_str;
    long old_size;

    str_a = (ch_string_t*)self.obj->data;

    b_as_str = CH_SEND(b, to_s);
    if (CH_FALSEP(ch_isa(b_as_str, ch_cString)))
        ch_raise2(ch_eTypeError,
                  "method String#concat! expects an argument whose "
                  "to_s method returns an actual String");

    str_b = (ch_string_t*)b_as_str.obj->data;

    old_size = str_a->size;
    str_resizeb(self, ch_fix(str_a->size + str_b->size));
    memcpy(&str_a->text[old_size], str_b->text, str_b->size);

    return self;
}

static ch_value_t str_concat_char_codeb(ch_value_t self, ch_value_t c)
{
    ch_string_t *selfstr = (ch_string_t*)self.obj->data;
    str_resizeb(self, ch_fix(selfstr->size + 1));
    selfstr->text[selfstr->size - 1] = (char)c.i;
    return self;
}

static ch_value_t str_concat(ch_value_t self, ch_value_t b)
{
    ch_value_t clone;

    clone = ch_obj_dup(self);
    str_concatb(clone, b);

    return clone;
}

static ch_value_t str_repeat(ch_value_t self, ch_value_t b)
{
    int i;
    ch_value_t result;

    if (b.type != CH_INT)
        ch_raise2(ch_eTypeError, "String#repeat expects an Integer");

    if (b.i < 0)
        ch_raise2(ch_eTypeError,
                  "String#repeat expects a non-negative Integer");

    if (b.i == 0)
        return ch_string_new();

    result = ch_obj_dup(self);
    for (i = 1; i < b.i; i++)
        str_concatb(result, self);

    return result;
}

static ch_value_t str_reverseb(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    int i;
    for(i = 0; i < str->size / 2; i++) {
        int endI = str->size - 1 - i;
        char tmp = str->text[i];
        str->text[i] = str->text[endI];
        str->text[endI] = tmp;
    }
    return self;
}

static ch_value_t str_reverse(ch_value_t self)
{
    ch_value_t dupped = ch_obj_dup(self);
    str_reverseb(dupped);
    return dupped;
}

static ch_value_t str_chompb(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    if(str->size != 0 && str->text[str->size - 1] == '\n') {
        str->text[str->size - 1] = '\0';
        str_resizeb(self, ch_fix(str->size - 1));
    }
    return self;
}

static ch_value_t str_chomp(ch_value_t self)
{
    ch_value_t dupped = ch_obj_dup(self);
    str_chompb(dupped);
    return dupped;
}

static ch_value_t str_chopb(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    if(str->size != 0) {
        str->text[str->size - 1] = '\0';
        str_resizeb(self, ch_fix(str->size - 1));
    }
    return self;
}

static ch_value_t str_chop(ch_value_t self)
{
    ch_value_t dupped = ch_obj_dup(self);
    str_chopb(dupped);
    return dupped;
}

static ch_value_t str_upcaseb(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    int i;
    for(i = 0; i < str->size; i++) {
        str->text[i] = toupper(str->text[i]);
    }
    return self;
}

static ch_value_t str_upcase(ch_value_t self)
{
    ch_value_t dupped = ch_obj_dup(self);
    str_upcaseb(dupped);
    return dupped;
}

static ch_value_t str_downcaseb(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    int i;
    for(i = 0; i < str->size; i++) {
        str->text[i] = tolower(str->text[i]);
    }
    return self;
}

static ch_value_t str_downcase(ch_value_t self)
{
    ch_value_t dupped = ch_obj_dup(self);
    str_downcaseb(dupped);
    return dupped;
}

static ch_value_t str_intern(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    return ch_internCString(str->text);
}

static ch_value_t str_inspect(ch_value_t self)
{
    ch_value_t result = ch_string_new1("\"");

    ch_string_t *selfstr = (ch_string_t*)self.obj->data;
    int i;

    for (i = 0; i < selfstr->size; i++) {
        char c = selfstr->text[i];
        char hex[2];
        if (c == '\r') {
            str_concat_char_codeb(result, ch_fix('\\'));
            str_concat_char_codeb(result, ch_fix('r'));
        } else if (c == '\n') {
            str_concat_char_codeb(result, ch_fix('\\'));
            str_concat_char_codeb(result, ch_fix('n'));
        } else if (c == '\t') {
            str_concat_char_codeb(result, ch_fix('\\'));
            str_concat_char_codeb(result, ch_fix('t'));
        } else if (c < 32 || c > 126) {
            str_concat_char_codeb(result, ch_fix('\\'));
            str_concat_char_codeb(result, ch_fix('x'));
            hex[1] = '0' + (c & 0xf);
            c >>= 4;
            hex[0] = '0' + (c & 0xf);

            if (hex[0] > '9')
                hex[0] += 'a' - '9' - 1;
            if (hex[1] > '9')
                hex[1] += 'a' - '9' - 1;

            str_concat_char_codeb(result, ch_fix(hex[0]));
            str_concat_char_codeb(result, ch_fix(hex[1]));
        } else if (c == '"' || c == '\\') {
            str_concat_char_codeb(result, ch_fix('\\'));
            str_concat_char_codeb(result, ch_fix(c));
        } else
            str_concat_char_codeb(result, ch_fix(c));
    }

    str_concat_char_codeb(result, ch_fix('"'));

    return result;
}

static ch_value_t str_eq(ch_value_t self, ch_value_t b)
{
    if(CH_FALSEP(ch_isa(b, ch_cString))) { return kFalse; }

    ch_string_t* strA = (ch_string_t*)self.obj->data;
    ch_string_t* strB = (ch_string_t*)b.obj->data;
    return strcmp(strA->text, strB->text) ? kFalse : kTrue;
}

#define COMPARE_METHOD(name, op)                                \
    static ch_value_t str##name(ch_value_t self, ch_value_t b)  \
    {                                                           \
        if (CH_FALSEP(ch_isa(b, ch_cString)))                   \
            ch_raise2(ch_eTypeError,                            \
                "method String#" #name " expected a string");   \
                                                                \
        return (strcmp(CH_CSTR(self), CH_CSTR(b)) op 0)         \
            ? kTrue : kFalse;                                   \
    }

COMPARE_METHOD(__lt__, <);
COMPARE_METHOD(__le__, <=);
COMPARE_METHOD(__gt__, >);
COMPARE_METHOD(__ge__, >=);

static ch_value_t str_hash(ch_value_t self)
{
    ch_string_t* str = (ch_string_t*)self.obj->data;
    return ch_fix(ch_hashCString(str->text, -1));
}

static ch_value_t str_nth(ch_value_t self, ch_value_t i)
{
    ch_string_t *str = (ch_string_t*)self.obj->data;

    if (i.type != CH_INT)
        ch_raise2(ch_eTypeError, "expected an integer index");

    if (i.i < 0)
        i.i += str->size;

    if (i.i < 0 || i.i >= str->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", i.i);

    return ch_string_new2(1, &str->text[i.i]);
}

static ch_value_t str_slice(ch_value_t self, ch_value_t start, ch_value_t end)
{
    ch_string_t *str = (ch_string_t*)self.obj->data;
    long istart, iend;

    if (start.type != CH_INT || end.type != CH_INT)
        ch_raise2(ch_eTypeError, "expected integer indices");

    istart = start.i + (start.i < 0 ? str->size : 0);
    iend = end.i + (end.i < 0 ? str->size : 0);

    if (istart > iend)
        return ch_string_new();

    if (istart < 0 || istart >= str->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", start.i);
    else if (iend < 0 || iend >= str->size)
        ch_raise2(ch_eOutOfBoundsError, "index %d out of bounds", end.i);

    return ch_string_new2(iend - istart + 1, &str->text[istart]);
}

static ch_value_t str_become_b(ch_value_t self, ch_value_t other)
{
    if (CH_FALSEP(ch_isa(other, ch_cString)))
        ch_raise2(ch_eTypeError, "String#become! expects a string");

    ch_string_t *me = (ch_string_t*)self.obj->data;
    ch_string_t *it = (ch_string_t*)other.obj->data;

    me->size = it->size;
    me->text = ch_strdup(it->text);

    return kNil;
}

void ch_defineStringMethods()
{
    ch_defineMethod(ch_cString, SYM_empty$q, 0, (ch_cfn_t)str_emptyp);
    ch_defineMethod(ch_cString, SYM_size, 0, (ch_cfn_t)str_size);
    ch_defineMethod(ch_cString,
                    SYM_to_s,
                    0,
                    (ch_cfn_t)str_toString);
    ch_defineMethod(ch_cString, SYM_to_i, 0, (ch_cfn_t)str_toInt);
    ch_defineMethod(ch_cString, SYM_to_f, 0, (ch_cfn_t)str_toFloat);
    ch_defineMethod(ch_cString, SYM_inspect_raw, 0, (ch_cfn_t)str_inspect);
    ch_defineMethod(ch_cString, SYM_resize$b, 1, (ch_cfn_t)str_resizeb);
    ch_defineMethod(ch_cString, SYM_concat$b, 1, (ch_cfn_t)str_concatb);
    ch_defineMethod(ch_cString, SYM_concat, 1, (ch_cfn_t)str_concat);
    ch_defineMethod(ch_cString, SYM_repeat, 1, (ch_cfn_t)str_repeat);
    ch_defineMethod(ch_cString, SYM_$e$e, 1, (ch_cfn_t)str_eq);   // ==
    ch_defineMethod(ch_cString, SYM_$l, 1, (ch_cfn_t)str__lt__);  // <
    ch_defineMethod(ch_cString, SYM_$l$e, 1, (ch_cfn_t)str__le__);// <=
    ch_defineMethod(ch_cString, SYM_$g, 1, (ch_cfn_t)str__gt__);  // >
    ch_defineMethod(ch_cString, SYM_$g$e, 1, (ch_cfn_t)str__ge__);// >=
    ch_defineMethod(ch_cString, SYM_hash, 0, (ch_cfn_t)str_hash);
    ch_defineMethod(ch_cString, SYM_intern, 0, (ch_cfn_t)str_intern);
    ch_defineMethod(ch_cString,
                    SYM_reverse$b,
                    0,
                    (ch_cfn_t)str_reverseb);
    ch_defineMethod(ch_cString, SYM_reverse, 0, (ch_cfn_t)str_reverse);
    ch_defineMethod(ch_cString, SYM_chomp$b, 0, (ch_cfn_t)str_chompb);
    ch_defineMethod(ch_cString, SYM_chomp, 0, (ch_cfn_t)str_chomp);
    ch_defineMethod(ch_cString, SYM_chop$b, 0, (ch_cfn_t)str_chopb);
    ch_defineMethod(ch_cString, SYM_chop, 0, (ch_cfn_t)str_chop);
    ch_defineMethod(ch_cString, SYM_upcase$b, 0, (ch_cfn_t)str_upcaseb);
    ch_defineMethod(ch_cString, SYM_upcase, 0, (ch_cfn_t)str_upcase);
    ch_defineMethod(ch_cString,
                    SYM_downcase$b,
                    0,
                    (ch_cfn_t)str_downcaseb);
    ch_defineMethod(ch_cString,
                    SYM_downcase,
                    0,
                    (ch_cfn_t)str_downcase);

    ch_defineMethod(ch_cString, SYM__call_, 1, (ch_cfn_t)str_nth);
    ch_defineMethod(ch_cString, SYM_slice, 2, (ch_cfn_t)str_slice);
    ch_defineMethod(ch_cString, SYM_substr, 2, (ch_cfn_t)str_slice);
    ch_defineMethod(ch_cString, SYM_become$b, 1, (ch_cfn_t)str_become_b);
}
