/* ch_signal.c -- chai signal handling
   September 30, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "misc.h"
#include "value.h"
#include "obj.h"
#include "ch_signal.h"
#include "Exception.h"
#include <string.h>
#include <ctype.h>

volatile sig_atomic_t ch_got_signal = 0;
static const int ch_catchable[] = {
    SIGALRM, SIGCHLD, SIGCONT, SIGHUP,  SIGINT, SIGIO,
    SIGTTIN, SIGTTOU, SIGUSR1, SIGUSR2, SIGWINCH, SIGTERM,
    SIGFPE
};

static const char ignore_by_default[] = "011000000010";

#define NCATCHABLE (sizeof (ch_catchable) / sizeof (ch_catchable[0]))

static int ch_sigdepth = 0;

static struct {
    ch_value_t handler[NCATCHABLE];
    struct sigaction old[NCATCHABLE];
} ch_siginfo;

/* expects substr to be all lower-case */
static int startswith(const char *name, const char *substr)
{
    while (*substr)
        if (tolower(*name++) != *substr++)
            return 0;

    return 1;
}

int ch_signal_from_name(const char *name)
{
    /* skip any SIG prefix */
    if (startswith(name, "sig"))
        name = &name[3];

    /* compare longer ones first in case of substring matches */
    if (startswith(name, "winch"))
        return SIGWINCH;
    else if (startswith(name, "alrm"))
        return SIGALRM;
    else if (startswith(name, "chld"))
        return SIGCHLD;
    else if (startswith(name, "cont"))
        return SIGCONT;
    else if (startswith(name, "ttin"))
        return SIGTTIN;
    else if (startswith(name, "ttou"))
        return SIGTTOU;
    else if (startswith(name, "usr1"))
        return SIGUSR1;
    else if (startswith(name, "usr2"))
        return SIGUSR2;
    else if (startswith(name, "term"))
        return SIGTERM;
    else if (startswith(name, "hup"))
        return SIGHUP;
    else if (startswith(name, "int"))
        return SIGINT;
    else if (startswith(name, "fpe"))
        return SIGFPE;
    else if (startswith(name, "io"))
        return SIGIO;

    ch_raise2(ch_eNameError, "unknown signal name %s\n", name);
}

static int signal_index(int signal)
{
    switch (signal) {
    case SIGALRM:  return  0;
    case SIGCHLD:  return  1;
    case SIGCONT:  return  2;
    case SIGHUP:   return  3;
    case SIGINT:   return  4;
    case SIGIO:    return  5;
    case SIGTTIN:  return  6;
    case SIGTTOU:  return  7;
    case SIGUSR1:  return  8;
    case SIGUSR2:  return  9;
    case SIGWINCH: return 10;
    case SIGTERM:  return 11;
    case SIGFPE:   return 12;
    }

    ch_raise2(ch_eArgumentError, "invalid signal %d", signal);
}

static const char *signal_name(int signal)
{
    switch (signal) {
    case SIGALRM:  return "SIGALRM";
    case SIGCHLD:  return "SIGCHLD";
    case SIGCONT:  return "SIGCONT";
    case SIGHUP:   return "SIGHUP";
    case SIGINT:   return "SIGINT";
    case SIGIO:    return "SIGIO";
    case SIGTTIN:  return "SIGTTIN";
    case SIGTTOU:  return "SIGTTOU";
    case SIGUSR1:  return "SIGUSR1";
    case SIGUSR2:  return "SIGUSR2";
    case SIGWINCH: return "SIGWINCH";
    case SIGTERM:  return "SIGTERM";
    case SIGFPE:   return "SIGFPE";
    }

    ch_raise2(ch_eArgumentError, "invalid signal %d", signal);
}

void ch_handle_signal(int sig)
{
    ch_value_t handler;
    int index;

    index = signal_index(sig);
    handler = ch_siginfo.handler[index];
    if (handler.type == CH_NIL) {
        if (ignore_by_default[index] == '1')
            return;

        signal(sig, SIG_DFL);
        raise(sig);
    }

    ch_funcall(handler, 1, ch_fix(sig));
}

static void ch_check_can_set_sig(int sig)
{
    const char *msg = "not allowed to set handler for %s";

    if (sig == SIGFPE) /* for now */
        ch_raise2(ch_eArgumentError, msg, signal_name(sig));
}

ch_value_t ch_signal_set_handler(int sig, ch_value_t handler)
{
    ch_value_t old;
    int index;

    ch_check_can_set_sig(sig);

    index = signal_index(sig);
    old = ch_siginfo.handler[index];
    ch_siginfo.handler[index] = handler;

    return old;
}

static void god_handler(int signal, siginfo_t *ignore_, void *ignore2_)
{
    ch_got_signal = signal;
}

void ch_pushSignalHandlers(void)
{
    if (ch_sigdepth++ == 0) {
        struct sigaction god, old;
        int i;

        /* Zero chai handlers */
        memset(ch_siginfo.handler, '\0', sizeof (ch_siginfo.handler));

        /* Set sigaction */
        god.sa_sigaction = god_handler;
        god.sa_flags = SA_SIGINFO;

        /* Ignore all signals I care about handling while in handler */
        sigemptyset(&god.sa_mask);
        for (i = 0; i < NCATCHABLE; i++)
            sigaddset(&god.sa_mask, ch_catchable[i]);

        for (i = 0; i < NCATCHABLE; i++) {
            sigaction(ch_catchable[i], &god, &old);
            ch_siginfo.old[i] = old;
        }
    }
}

void ch_popSignalHandlers(void)
{
    if (--ch_sigdepth == 0)
    {
        int i;

        /* Reset sigaction to old values */
        for (i = 0; i < NCATCHABLE; i++)
            sigaction(ch_catchable[i], &ch_siginfo.old[i], NULL);
    }
}
