/* compile.c -- chai lexer, parser, macroexpander, compiler
   October 1, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "compile.h"
#include "value.h"
#include "mem.h"
#include "obj.h"
#include "state.h"
#include "sym.h"
#include "Exception.h"
#include "Module.h"
#include "Bytecode.h"
#include "sym_literals.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <dlfcn.h>

typedef struct token {
    ch_value_t type;
    ch_value_t value;
} token_t;

typedef struct lexer {
    void *stream;
    int (*fgetc)(void*);
    int (*ungetc)(int, void*);
    int eofp;

    int ungot_count;
    int ungot_cap;
    token_t *ungot;
    ch_ht_t *op_idents;
} lexer_t;

typedef struct ch_parser {
    lexer_t *lex;
    ch_ht_t *prec_table, *uprec_table, *assoc_table, *matching;
} parser_t;

typedef struct compiler {
    parser_t *p;
} compiler_t;

static ch_ht_t *opcodes;
static ch_ht_t *operand_align;

static ch_ht_t *call_impls;
static ch_ht_t *setf_impls;

static char *op_chars = "-+.<>/*%!?&|^=:~@$";
static char lowest_op_char = '\0';
static char highest_op_char = '\0';

static ch_ht_t *parse_matching = NULL;

static ch_ht_t *ch_my_bindings;
static ch_ht_t *ch_tagstrs;

static int in_op_chars(int c)
{
    int i;

    if (lowest_op_char == '\0') {
        for (i = 0; i < strlen(op_chars); i++) {
            if (op_chars[i] < lowest_op_char)
                lowest_op_char = op_chars[i];
            if (op_chars[i] > highest_op_char)
                highest_op_char = op_chars[i];
        }
    }

    if (c < lowest_op_char || c > highest_op_char)
        return 0;

    for (i = 0; i < strlen(op_chars); i++)
        if (op_chars[i] == c)
            return 1;

    return 0;
}

static int l_getc(lexer_t *lex)
{
    int c;

    if (lex->eofp)
        return EOF;

    c = lex->fgetc(lex->stream);
    if (c == EOF)
        lex->eofp = 1;

    return c;
}

static int l_ungetc(lexer_t *lex, int c)
{
    if (c == EOF)
        return EOF;

    lex->eofp = 0;
    return lex->ungetc(c, lex->stream);
}

static lexer_t *l_new(void *stream, int (*fgetc)(void*),
                      int (*ungetc)(int, void*))
{
    lexer_t *l;
    int c, c2;

    l = ch_malloc(sizeof(lexer_t));
    l->stream = stream;
    l->fgetc = fgetc;
    l->eofp = 0;
    l->ungetc = ungetc;
    l->ungot_count = 0;
    l->ungot_cap = 8;
    l->ungot = ch_malloc(l->ungot_cap * sizeof (token_t));

    ch_module_t *cur_m = (ch_module_t*)ch_current_module.obj->data;
    l->op_idents = &cur_m->op_idents;

    c = l_getc(l);
    if (c == '#') {
        c2 = l_getc(l);
        if (c2 == '!')
            while ((c = l_getc(l)) != '\n');
        else {
            if (c2 != EOF)
                l_ungetc(l, c2);

            l_ungetc(l, '#');
        }
    } else if (c != EOF)
        l_ungetc(l, c);

    return l;
}

static void l_slurp_lines(lexer_t *lex)
{
    int comment = 0;
    int c, c2;

    while ((c = l_getc(lex)) != EOF && (comment || c == ';' || isspace(c))) {
        if (comment) {
            if (c == '\n')
                comment = 0;
        } else if (c == ';') {
            c2 = l_getc(lex);
            if (c2 == ';')
                comment = 1;
            else {
                if (c2 != EOF)
                    l_ungetc(lex, c2);
                break;
            }
        }
    }

    if (c != EOF)
        l_ungetc(lex, c);
}

static token_t l_read_tagstring(lexer_t *lex, char *tag)
{
    char buf[8192 * 2];
    int len = 0;
    int escape = 0;
    int c;

    memset(buf, '\0', sizeof (buf));
    len = snprintf(buf, sizeof (buf), "%s\"", tag);
    while ((c = l_getc(lex)) != EOF) {
        if (len >= sizeof (buf))
            ch_raise2(ch_eParseError, "tagged string too long");

        buf[len++] = c;

        if (escape)
            escape = 0;
        else if (c == '\\')
            escape = 1;
        else if (c == '"')
            break;
    }

    if (c == EOF)
        ch_raise2(ch_eParseError, "premature end of stream in tagged string");

    token_t result = { SYM_tagstring, ch_string_new1(buf) };
    return result;
}

static token_t l_read_ident(lexer_t *lex)
{
    char buf[1024];
    int len, c;
    token_t result;
    ch_value_t dummy;

    len = 0;
    while (len < sizeof (buf) - 2 && (c = l_getc(lex)) != EOF) {
        /* beginning or after a colon, only alphanumerics and underscores */
        if ((len == 0 || buf[len - 1] == ':') && !isalpha(c) && c != '_')
            break;

        /* Normally alphanumerics, underscores, and colons */
        if (!isalnum(c) && c != '_' && c != ':')
            break;

        buf[len++] = c;
    }

    if (len == sizeof (buf) - 2)
        ch_raise2(ch_eParseError, "identifier too long");

    /* ! and ? allowed at the end */
    if (c == '!' || c == '?')
        buf[len++] = c;
    else if (c == '"') {
        buf[len] = '\0';
        return l_read_tagstring(lex, buf);
    } else if (c != EOF)
        l_ungetc(lex, c);

    result.value = ch_internSizedString(len, buf);
    if (ch_ht_get(lex->op_idents, result.value, &dummy))
        result.type = SYM_ident;
    else {
        result.type = SYM_op;
        l_slurp_lines(lex);
    }

    return result;
}

static token_t l_read_op(lexer_t *lex)
{
    char buf[1024];
    int len, c;
    token_t result;

    len = 0;
    while (len < sizeof (buf) - 2 && in_op_chars(c = l_getc(lex)))
        buf[len++] = c;

    if (len == sizeof (buf) - 2)
        ch_raise2(ch_eParseError, "operator too long");

    if (c != EOF)
        l_ungetc(lex, c);

    result.type = SYM_op;
    result.value = ch_internSizedString(len, buf);

    l_slurp_lines(lex);

    return result;
}

static token_t l_read_float(lexer_t *lex, const char *whole_part)
{
    char buf[1024] = "0.";
    int len, c;
    token_t result;

    len = 2;
    while (len < sizeof (buf) - 1 && (c = l_getc(lex)) != EOF
           && c >= '0' && c <= '9')
        buf[len++] = c;

    if (c != EOF)
        l_ungetc(lex, c);

    if (len == sizeof (buf) - 1)
        ch_raise2(ch_eParseError, "literal float too long");

    buf[len] = '\0';

    if (len == 2) {
        /* It's not really a float! Aaaah! */
        l_ungetc(lex, '.');
        result.type = SYM_int;
        result.value = ch_fix(atoi(whole_part));
        return result;
    }

    result.type = SYM_float;
    result.value = ch_float(atoi(whole_part) + atof(buf));
    return result;
}

static token_t l_read_hex(lexer_t *lex)
{
    char buf[1024] = "0x";
    int len, c;
    token_t result;

    len = 2;
    while (len < sizeof (buf) - 1 && (c = l_getc(lex)) != EOF
           && ((c >= '0' && c <= '9') ||
               (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F')))
        buf[len++] = c;

    if (len == sizeof (buf) - 1)
        ch_raise2(ch_eParseError, "hex literal too long");

    buf[len] = '\0';

    l_ungetc(lex, c);

    result.type = SYM_int;
    result.value = ch_fix(strtol(buf, NULL, 0));
    return result;
}

static token_t l_read_octal(lexer_t *lex)
{
    char buf[1024] = "0";
    int len, c;
    token_t result;

    len = 1;
    while (len < sizeof (buf) - 1 && (c = l_getc(lex)) != EOF
           && (c >= '0' && c <= '7'))
        buf[len++] = c;

    if (len == sizeof (buf) - 1)
        ch_raise2(ch_eParseError, "octal literal too long");

    buf[len] = '\0';

    l_ungetc(lex, c);

    result.type = SYM_int;
    result.value = ch_fix(strtol(buf, NULL, 0));
    return result;
}

static token_t l_read_num(lexer_t *lex)
{
    char buf[1024];
    int len, c;
    token_t result;

    c = l_getc(lex);
    if (c == '0') {
        c = l_getc(lex);
        if (c == 'x' || c == 'X')
            return l_read_hex(lex);
        else if (c >= '0' && c <= '7') {
            l_ungetc(lex, c);
            return l_read_octal(lex);
        } else if (c == '.')
            return l_read_float(lex, "0");

        l_ungetc(lex, c);
        result.type = SYM_int;
        result.value = ch_fix(0);
        return result;
    }

    len = 0;
    buf[len++] = c;
    while (len < sizeof (buf) - 1 && (c = l_getc(lex)) != EOF
           && c >= '0' && c <= '9')
        buf[len++] = c;

    if (len == sizeof (buf) - 1)
        ch_raise2(ch_eParseError, "number literal too long");

    buf[len] = '\0';

    if (c == '.')
        return l_read_float(lex, buf);

    l_ungetc(lex, c);
    result.type = SYM_int;
    result.value = ch_fix(atoi(buf));
    return result;
}

static token_t l_read_string(lexer_t *lex, char end)
{
    const char *ueos = "unexpected end of stream in %c-string";
    const char *inseq = "invalid escape sequence: \\%c%s";
    char buf[8192];
    int len, c, done, escape;
    token_t result;

    len = 0;
    done = 0;
    escape = 0;
    while (len < sizeof (buf) - 1 && !done) {
        c = l_getc(lex);
        if (c == EOF)
            ch_raise2(ch_eParseError, ueos, end);

        if (escape) {
            escape = 0;
            if (c == 'n')
                buf[len++] = '\n';
            else if (c == 't')
                buf[len++] = '\t';
            else if (c == 'r')
                buf[len++] = '\r';
            else if (c == 'a')
                buf[len++] = '\a';
            else if (c == end || c == '\\')
                buf[len++] = c;
            else if (c != 'x')
                ch_raise2(ch_eParseError, inseq, c, "");
            else {
                int hex[2];
                char hexstr[3] = "\0\0";

                hex[0] = l_getc(lex);
                if (hex[0] == EOF)
                    ch_raise2(ch_eParseError, ueos, end);

                hex[1] = l_getc(lex);
                if (hex[1] == EOF)
                    ch_raise2(ch_eParseError, ueos, end);

                hexstr[0] = (char)hex[0];
                if (!isxdigit(hexstr[0]))
                    ch_raise2(ch_eParseError, inseq, 'x', hexstr);

                hexstr[1] = (char)hex[1];
                if (!isxdigit(hexstr[1]))
                    ch_raise2(ch_eParseError, inseq, 'x', hexstr);

                buf[len++] = (char)strtol(hexstr, NULL, 16);
            }
        } else if (c == end)
            done = 1;
        else if (c == '\\')
            escape = 1;
        else
            buf[len++] = c;
    }

    if (len == sizeof (buf) - 1)
        ch_raise2(ch_eParseError, "string literal too long");

    buf[len] = '\0';

    result.type = SYM_str;
    result.value = ch_string_new2(len, buf);
    return result;
}

static void l_unget(lexer_t *lex, token_t tok)
{
    if (lex->ungot_count >= lex->ungot_cap) {
        lex->ungot_cap *= 2;
        lex->ungot = ch_realloc(lex->ungot, lex->ungot_cap * sizeof (token_t));
    }

    lex->ungot[lex->ungot_count++] = tok;
}

static token_t l_next(lexer_t *lex)
{
# define RET_TYP(typ)  do { result.type = SYM_##typ; return result; } while(0)
    int c, c2;
    token_t result;

    if (lex->ungot_count)
        return lex->ungot[--lex->ungot_count];

    result.type = SYM_eof;
    result.value = kNil;

    while ((c = l_getc(lex)) != EOF && (c == ' ' || c == '\t'));

    if (c == EOF)
        return result;

    switch (c) {
    case ';':
        c2 = l_getc(lex);
        if (c2 != ';') {
            if (c2 != EOF)
                l_ungetc(lex, c2);

            RET_TYP(newline);
        }

        l_ungetc(lex, ';');
        l_ungetc(lex, ';');

    case '\n':
        l_slurp_lines(lex);

        c2 = l_getc(lex);
        if (c2 == ')')
            RET_TYP(rparen);
        else if (c2 == ']')
            RET_TYP(rbracket);
        else if (c2 == '}')
            RET_TYP(rbrace);
        else if (c2 == '{')
            RET_TYP(lbrace);
        else if (c2 == ':') {
            c = l_getc(lex);
            if (c == ':')
                RET_TYP(double_colon);

            if (c != EOF)
                l_ungetc(lex, c);
            if (c2 != EOF)
                l_ungetc(lex, c2);
        } else if (c2 != EOF)
            l_ungetc(lex, c2);

        RET_TYP(newline);

    case ',':
        l_slurp_lines(lex);
        RET_TYP(comma);

    case '$':
        c2 = l_getc(lex);
        if (in_op_chars(c2)) {
            l_ungetc(lex, c2);
            l_ungetc(lex, '$');
            return l_read_op(lex);
        }
        l_slurp_lines(lex);
        RET_TYP(comma);

    case '"':
        return l_read_string(lex, '"');

    /*
    case '\'': {
        l_unget(lex, l_read_string(lex, '\''));
        token_t tag;
        tag.type = SYM_ident;
        tag.value = SYM_r;
        return tag;
    }
    */

    case '{':
        RET_TYP(lbrace);
    case '}':
        RET_TYP(rbrace);
    case '[':
        RET_TYP(lbracket);
    case ']':
        RET_TYP(rbracket);
    case '(':
        RET_TYP(lparen);
    case ')':
        RET_TYP(rparen);

    case ':':
        c2 = l_getc(lex);
        if (c2 == ':')
            RET_TYP(double_colon);
        else if (c2 != EOF)
            l_ungetc(lex, c2);

        ch_raise2(ch_eParseError, "invalid character sequence: :%c", c2);

    case '#':
        c2 = l_getc(lex);
        if (c2 == '"') {
            token_t str_tok = l_read_string(lex, '"');
            ch_string_t *str = (ch_string_t*)str_tok.value.obj->data;
            result.type = SYM_sym;
            result.value = ch_internSizedString(str->size, str->text);
            return result;
        } else if (c2 == '|') {
            token_t str_tok = l_read_string(lex, '|');
            ch_string_t *str = (ch_string_t*)str_tok.value.obj->data;
            result.type = SYM_ident;
            result.value = ch_internSizedString(str->size, str->text);
            return result;
        } else if ((c2 >= 'a' && c2 <= 'z') || (c2 >= 'A' && c2 <= 'Z')
                   || c2 == '_') {
            token_t ident_tok;
            l_ungetc(lex, c2);
            ident_tok = l_read_ident(lex);
            result.type = SYM_sym;
            result.value = ident_tok.value;
            return result;
        }
        ch_raise2(ch_eParseError, "invalid character sequence: #%c", c2);
    }

    l_ungetc(lex, c);

    if (isalpha(c) || c == '_')
        return l_read_ident(lex);
    else if (isdigit(c))
        return l_read_num(lex);
    else if (in_op_chars(c))
        return l_read_op(lex);

    ch_raise2(ch_eParseError, "invalid start of token: %c", c);

    return result; /* Makes the compiler shut up */
#undef RET_TYP
}

static token_t l_peek(lexer_t *lex)
{
    token_t result;

    result = l_next(lex);
    l_unget(lex, result);
    return result;
}

static ch_value_t p_stmts(parser_t *p);

static parser_t *p_new(void *stream, int (*fgetc)(void*),
                       int (*ungetc)(int, void*))
{
    parser_t *p = ch_malloc(sizeof(parser_t));
    ch_module_t *cur_m = (ch_module_t*)ch_current_module.obj->data;

    p->lex = l_new(stream, fgetc, ungetc);
    p->prec_table = &cur_m->prec_table;
    p->uprec_table = &cur_m->uprec_table;
    p->assoc_table = &cur_m->assoc_table;

    if (parse_matching == NULL) {
        parse_matching = ch_ht_create(NULL, NULL);
        ch_ht_set(parse_matching, SYM_lparen, SYM_rparen);
        ch_ht_set(parse_matching, SYM_rparen, SYM_lparen);
        ch_ht_set(parse_matching, SYM_lbracket, SYM_rbracket);
        ch_ht_set(parse_matching, SYM_rbracket, SYM_lbracket);
        ch_ht_set(parse_matching, SYM_lbrace, SYM_rbrace);
        ch_ht_set(parse_matching, SYM_rbrace, SYM_lbrace);
    }
    p->matching = parse_matching;

    return p;
}

static token_t p_next_token(parser_t *p)
{
    return l_next(p->lex);
}

static token_t p_peek_token(parser_t *p)
{
    return l_peek(p->lex);
}

static void p_unget_token(parser_t *p, token_t t)
{
    l_unget(p->lex, t);
}

static int p_next_is_term(parser_t *p)
{
    token_t tok = p_peek_token(p);
    return tok.type.sym == SYM_semicolon.sym
        || tok.type.sym == SYM_newline.sym;
}

static int p_peek_is(parser_t *p, ch_value_t kind)
{
    token_t tok = p_peek_token(p);
    return tok.type.sym == kind.sym;
}

static token_t p_expect(parser_t *p, ch_value_t kind)
{
    token_t t = p_next_token(p);

    if (t.type.sym != kind.sym) {
        ch_value_t s;

        s = CH_SEND(t.value.type == CH_NIL ? t.type : t.value, to_s);
        ch_raise2(ch_eParseError, "expected %s near %s",
                  kind.sym + sizeof(long), CH_CSTR(s));
    }

    return t;
}

static ch_value_t p_expr(parser_t*);

static ch_value_t p_paramlist_inner(parser_t *p, int bitch)
{
    ch_value_t parentype, matching, params;

    params = ch_list_new();

    parentype = p_next_token(p).type;
    if (ch_ht_get(p->matching, parentype, &matching) < 0)
        ch_raise2(ch_eParseError, "paramlist called with no entry in matching");

    if (p_peek_is(p, matching)
        || (!bitch && (p_peek_is(p, SYM_newline) || p_peek_is(p, SYM_semicolon))))
        return params;

    CH_SEND(params, push, p_expr(p));

    while (p_peek_is(p, SYM_comma)) {
        p_next_token(p);
        CH_SEND(params, push, p_expr(p));
    }

    if (bitch && !p_peek_is(p, matching)) {
        token_t tok = p_peek_token(p);
        ch_value_t s;

        s = CH_SEND(tok.value.type == CH_NIL ? tok.type : tok.value, to_s);
        ch_raise2(ch_eParseError, "expected %s near %s",
                  matching.sym + sizeof (long), CH_CSTR(s));
    }

    return params;
}

static ch_value_t p_paramlist(parser_t *p)
{
    return p_paramlist_inner(p, 1);
}

static ch_value_t p_paramlist_dont_bitch(parser_t *p)
{
    return p_paramlist_inner(p, 0);
}

static ch_value_t ch_expand_tagstr(const char *name, const char *text)
{
    ch_value_t tag = ch_internCString(name);
    ch_value_t str = ch_string_new1(text);

    ch_value_t f;
    if (ch_ht_get(ch_tagstrs, tag, &f) < 0)
        ch_raise2(ch_eParseError, "unknown tagged string type: %s", name);

    return ch_vfuncall(f, 1, &str);
}

static ch_value_t p_special_form(parser_t *p, ch_value_t name)
{
    token_t rbrace_tok = { SYM_rbrace, kNil };
    ch_value_t args = ch_list_new1(1, name);
    ch_list_t *lst;

    do {
        if (p_peek_is(p, SYM_newline) || p_peek_is(p, SYM_semicolon)
            || p_peek_is(p, SYM_rbracket)
            || p_peek_is(p, SYM_rparen)
            || p_peek_is(p, SYM_rbrace))
            return args;

        l_unget(p->lex, rbrace_tok);
        CH_SEND(args, concat$b, p_paramlist_dont_bitch(p));

        if (p_peek_is(p, SYM_newline) || p_peek_is(p, SYM_semicolon)
            || p_peek_is(p, SYM_rbracket)
            || p_peek_is(p, SYM_rparen)
            || p_peek_is(p, SYM_rbrace))
            return args;

        if (p_peek_is(p, SYM_lbrace)) {
            p_next_token(p);
            CH_SEND(args, push, p_stmts(p));
            p_expect(p, SYM_rbrace);
        } else if (!p_peek_is(p, SYM_eof)) {
            token_t tok = l_peek(p->lex);
            ch_raise2(ch_eParseError, "unexpected token %s in special form",
                      tok.type.sym + sizeof (long));
        }
    } while (p_peek_is(p, SYM_double_colon) && p_next_token(p).type.sym);

    /* No [#progn, [#progn, foo, bar]] nonsense from the parser when
     * the user explicitly does progn: { foo; bar } */
    lst = (ch_list_t*)args.obj->data;
    if (name.sym == SYM_progn.sym && lst->size == 2) {
        ch_value_t first = lst->elems[1];
        lst = (ch_list_t*)first.obj->data;
        if (CH_TRUEP(ch_isa(first, ch_cList)) && lst->size > 0)
            if (lst->elems[0].type == CH_SYMBOL)
                if (lst->elems[0].sym == SYM_progn.sym)
                    return first;
    }

    return args;
}

static ch_value_t p_sexpr(parser_t *p)
{
    token_t tok;
    const char *typ;

    tok = p_next_token(p);
    typ = tok.type.sym;

    if (typ == SYM_eof.sym)
        ch_raise2(ch_eParseError, "unexpected end of stream");

    if (typ != SYM_ident.sym && typ != SYM_int.sym && typ != SYM_float.sym
        && typ != SYM_str.sym && typ != SYM_sym.sym && typ != SYM_tagstring.sym) {

        if (typ == SYM_lparen.sym) {
            ch_value_t result = p_stmts(p);
            p_expect(p, SYM_rparen);
            return result;
        }

        if (typ == SYM_lbracket.sym) {
            ch_value_t list;

            p_unget_token(p, tok);
            list = p_paramlist(p);
            p_expect(p, SYM_rbracket);
            CH_SEND(list, unshift, SYM_list);
            return list;
        }

        ch_value_t s;
        s = CH_SEND(tok.value.type == CH_NIL ? tok.type : tok.value, to_s);
        ch_raise2(ch_eParseError, "unexpected token %s", CH_CSTR(s));
    }

    if (typ == SYM_ident.sym) {
        char *text = ch_strdup(tok.value.sym + sizeof (long));
        int len = strlen(text);
        if (text[len - 1] == ':') {
            text[len - 1] = '\0';
            return p_special_form(p, ch_internCString(text));
        }

        if (p_peek_is(p, SYM_lbracket)) {
            ch_value_t args = p_paramlist(p);
            p_expect(p, SYM_rbracket);
            CH_SEND(args, unshift, tok.value);
            return args;
        }
    }

    if (typ == SYM_tagstring.sym) {
        ch_string_t *str = (ch_string_t*)tok.value.obj->data;
        char *c = str->text;

        while (*c != '"') c++;
        *c++ = '\0';
        str->text[--str->size] = '\0';

        return ch_expand_tagstr(str->text, c);
    }

    if (typ == SYM_sym.sym)
        return ch_list_new1(2, SYM_quote, tok.value);

    return tok.value;
}

static void p_check_unary(parser_t *p, ch_value_t ops,
                          ch_value_t precs, ch_value_t stack,
                          ch_list_t *oplist, ch_list_t *preclist)
{
    while (p_peek_is(p, SYM_op)) {
        token_t op = p_next_token(p);

        ch_value_t prec;
        if (ch_ht_get(p->uprec_table, op.value, &prec) < 0
            || prec.type != CH_INT)
            ch_raise2(ch_eParseError, "%s is not a unary operator",
                      op.value.sym + sizeof (long));

        /* Is this associativity check necessary? */
        ch_value_t assoc;
        if (ch_ht_get(p->assoc_table, prec, &assoc) < 0
            || assoc.sym != SYM_unary.sym)
            ch_raise2(ch_eParseError, "%s is unary but its precedence is not",
                      op.value.sym + sizeof (long));

        if (oplist->size != 0 && prec.i < preclist->elems[preclist->size - 1].i
            && oplist->elems[oplist->size - 1].sym != SYM___call_op__.sym)
            ch_raise2(ch_eParseError, "%s has a lower precedence than %s",
                      op.value.sym + sizeof (long),
                      oplist->elems[oplist->size -1].sym + sizeof (long));

        CH_SEND(stack, push, kNil);
        CH_SEND(ops, push, op.value);
        CH_SEND(precs, push, prec);
    }
}

static ch_value_t p_expr(parser_t *p)
{
    ch_value_t ops, precs, stack;
    ch_list_t *oplist, *preclist, *stacklist;

    ops = ch_list_new();
    precs = ch_list_new();
    stack = ch_list_new();
    oplist = (ch_list_t*)ops.obj->data;
    preclist = (ch_list_t*)precs.obj->data;
    stacklist = (ch_list_t*)stack.obj->data;

    p_check_unary(p, ops, precs, stack, oplist, preclist);

    CH_SEND(stack, push, p_sexpr(p));

    while (p_peek_is(p, SYM_op) || p_peek_is(p, SYM_lparen)) {
        token_t op_tok = p_next_token(p);
        ch_value_t op = op_tok.type.sym == SYM_lparen.sym
            ? SYM___call_op__
            : op_tok.value;

        ch_value_t prec;
        if (ch_ht_get(p->prec_table, op, &prec) < 0 || prec.type != CH_INT)
            ch_raise2(ch_eParseError, "%s is not a binary operator",
                      CH_CSTR(CH_SEND(op, to_s)));

        ch_value_t assoc;
        if (ch_ht_get(p->assoc_table, prec, &assoc) < 0)
            ch_raise2(ch_eParseError, "precedence %d has no associativity",
                      prec.i);

        while (stacklist->size > 1) {
            ch_value_t last_prec = preclist->elems[preclist->size - 1];
            ch_value_t last_assoc;

            /* Not checking the return because it worked before */
            ch_ht_get(p->assoc_table, last_prec, &last_assoc);

            if (last_prec.i <= prec.i
                && !(last_assoc.sym == SYM_left.sym && last_prec.i == prec.i))
                break;

            ch_value_t right, left;
            right = CH_SEND(stack, pop);
            left = CH_SEND(stack, pop);

            if (oplist->elems[oplist->size - 1].sym == SYM___call_op__.sym) {
                CH_SEND(ops, pop);

                ch_value_t call = ch_list_new1(2, SYM_call, left);
                CH_SEND(call, concat$b, right);
                CH_SEND(stack, push, call);
            } else if (left.type == CH_NIL) {
                /* unary operator */
                ch_value_t op_sym, op_name;
                char opstr[512];

                op_sym = CH_SEND(ops, pop);
                snprintf(opstr, sizeof (opstr), "uop_%s",
                         op_sym.sym + sizeof (long));

                op_name = ch_internCString(opstr);
                CH_SEND(stack, push, ch_list_new1(2, op_name, right));
             } else {
                ch_value_t op_sym, op_name;
                char opstr[512];

                op_sym = CH_SEND(ops, pop);
                snprintf(opstr, sizeof (opstr), "op_%s",
                         op_sym.sym + sizeof (long));

                op_name = ch_internCString(opstr);
                CH_SEND(stack, push, ch_list_new1(3, op_name, left, right));
            }

            CH_SEND(precs, pop);
        }

        CH_SEND(ops, push, op);
        CH_SEND(precs, push, prec);

        if (op.sym == SYM___call_op__.sym) {
            p_unget_token(p, op_tok);
            ch_value_t arg_exprs = p_paramlist(p);
            p_expect(p, SYM_rparen);
            CH_SEND(stack, push, arg_exprs);
        } else {
            p_check_unary(p, ops, precs, stack, oplist, preclist);
            CH_SEND(stack, push, p_sexpr(p));
        }
    }

    while (stacklist->size > 1) {
        ch_value_t right, left;
        right = CH_SEND(stack, pop);
        left = CH_SEND(stack, pop);

        if (oplist->elems[oplist->size - 1].sym == SYM___call_op__.sym) {
            CH_SEND(ops, pop);

            ch_value_t call = ch_list_new1(2, SYM_call, left);
            CH_SEND(call, concat$b, right);
            CH_SEND(stack, push, call);
        } else if (left.type == CH_NIL){
            /* unary op */
            ch_value_t op_sym, op_name;
            char opstr[512];

            op_sym = CH_SEND(ops, pop);
            snprintf(opstr, sizeof (opstr), "uop_%s",
                     op_sym.sym + sizeof (long));

            op_name = ch_internCString(opstr);
            CH_SEND(stack, push, ch_list_new1(2, op_name, right));
        } else {
            ch_value_t op_sym, op_name;
            char opstr[512];

            op_sym = CH_SEND(ops, pop);
            snprintf(opstr, sizeof (opstr), "op_%s",
                     op_sym.sym + sizeof (long));

            op_name = ch_internCString(opstr);
            CH_SEND(stack, push, ch_list_new1(3, op_name, left, right));
        }
    }

    return stacklist->elems[0];
}

static void p_terminators(parser_t *p)
{
    if (!p_peek_is(p, SYM_newline) && !p_peek_is(p, SYM_semicolon))
        ch_raise2(ch_eParseError, "expected newline or semicolon");

    while (p_next_is_term(p))
        p_next_token(p);
}

static ch_value_t p_exprs(parser_t *p)
{
    ch_value_t result = ch_list_new1(1, p_expr(p));

    while (p_next_is_term(p)) {
        p_terminators(p);
        if (p_peek_is(p, SYM_eof) || p_peek_is(p, SYM_rbrace))
            break;

        CH_SEND(result, push, p_expr(p));
    }

    return result;
}

static ch_value_t p_stmts(parser_t *p)
{
    ch_value_t result;
    ch_list_t *reslist;

    if (p_next_is_term(p))
        p_terminators(p);

    if (p_peek_is(p, SYM_rbrace) || p_peek_is(p, SYM_eof))
        return SYM_nil;

    result = p_exprs(p);

    if (p_next_is_term(p))
        p_terminators(p);

    reslist = (ch_list_t*)result.obj->data;
    if (reslist->size == 1)
        return reslist->elems[0];

    CH_SEND(result, unshift, SYM_progn);
    return result;
}

ch_value_t ch_defcalledp(ch_value_t func)
{
    ch_list_t *func_list;
    ch_value_t dummy;

    func_list = (ch_list_t*)func.obj->data;
    return (CH_TRUEP(ch_isa(func, ch_cList)) && func_list->size > 0
            && func_list->elems[0].type == CH_SYMBOL
            && !ch_ht_get(call_impls, func_list->elems[0], &dummy))
        ? kTrue
        : kFalse;
}

ch_value_t ch_defsetfedp(ch_value_t lval)
{
    ch_list_t *lval_list;
    ch_value_t dummy;

    lval_list = (ch_list_t*)lval.obj->data;
    return (CH_TRUEP(ch_isa(lval, ch_cList)) && lval_list->size > 0
            && lval_list->elems[0].type == CH_SYMBOL
            && !ch_ht_get(setf_impls, lval_list->elems[0], &dummy))
        ? kTrue
        : kFalse;
}

ch_value_t ch_macroexpand1(ch_value_t expr_val)
{
    ch_list_t *expr;
    ch_value_t form, expander;

    if (CH_FALSEP(ch_isa(expr_val, ch_cList)))
        return expr_val;

    expr = (ch_list_t*)expr_val.obj->data;
    form = expr->elems[0];
    if (expr->size == 0 || form.type != CH_SYMBOL
        || form.sym == SYM_quote.sym || form.sym == SYM_fn.sym
        || form.sym == SYM_let.sym || form.sym == SYM_slot.sym
        || form.sym == SYM_setq.sym || form.sym == SYM_if.sym
        || form.sym == SYM_while.sym || form.sym == SYM_unwind.sym
        || form.sym == SYM_unwind_protect.sym || form.sym == SYM_not.sym
        || form.sym == SYM_set_slot.sym || form.sym == SYM__progn.sym
        || form.sym == SYM_list.sym || form.sym == SYM_varslot.sym
        || form.sym == SYM_set_varslot.sym)
        return expr_val;

    if (form.sym == SYM_call.sym) {
        ch_value_t f;

        f = expr->elems[1];
        if (CH_FALSEP(ch_defcalledp(f)))
            return expr_val;
        else {
            ch_value_t name, call_defn;
            ch_list_t *f;

            f = (ch_list_t*)expr->elems[1].obj->data;
            if (f->size == 0)
                ch_raise2(ch_eParseError, "zero-size AST list as f in call[]");

            name = f->elems[0];
            if (name.type != CH_SYMBOL)
                ch_raise2(ch_eParseError, "improperly named form as f in call[]");

            if (ch_ht_get(call_impls, name, &call_defn) < 0)
                fatal("can't get call_defn when expr is defcalled?(), "
                      "should never happen");

            return ch_funcall(call_defn, 2,
                              expr->elems[1],
                              ch_list_new2(expr->size - 2,
                                           &expr->elems[2]));
        }
    }

    ch_value_t mod = ch_current_module, macro_name = form;
    ch_lookup_global_name(mod, macro_name, &mod, &macro_name);
    ch_module_t *m = (ch_module_t*)mod.obj->data;
    if(ch_ht_get(&m->macros, macro_name, &expander) < 0)
        ch_raise2(ch_eParseError, "no macro %s", form.sym + sizeof (long));

    return ch_vfuncall(expander, expr->size - 1, &expr->elems[1]);
}

static ch_value_t macroexpand_list(ch_value_t expr)
{
    ch_value_t form, result;
    ch_list_t *list;
    int i;

    list = (ch_list_t*)expr.obj->data;
    if (list->size < 1)
        ch_raise2(ch_eParseError, "empty list in ast");

    form = list->elems[0];

    if (form.sym == SYM_quote.sym) {
        return expr;
    } else if (form.sym == SYM_send.sym) {
        if (list->size < 3)
            ch_raise2(ch_eParseError, "too few arguments to send[]");

        result = ch_list_new1(
            3,
            SYM_send,
            ch_macroexpand(list->elems[1]),
            list->elems[2]);

        for (i = 3; i < list->size; i++)
            CH_SEND(result, push, ch_macroexpand(list->elems[i]));

        return result;
    } else if (form.sym == SYM_fn.sym || form.sym == SYM_let.sym) {
        int hit_optional = 0;

        if ((form.sym == SYM_fn.sym && list->size < 2)
            || (form.sym == SYM_let.sym && list->size < 3))
            ch_raise2(ch_eParseError, "too few arguments to %s[]",
                      form.sym + sizeof (long));

        result = ch_list_new1(1, form);

        for (i = 1; i < list->size - 1; i++) {
            ch_value_t param_expr_val;
            ch_list_t *param_expr;

            if (list->elems[i].type == CH_SYMBOL) {
                if (hit_optional && form.sym == SYM_fn.sym)
                    goto fn_fail;

                CH_SEND(result, push, list->elems[i]);
                continue;
            }

            hit_optional = 1;

            if (CH_FALSEP(ch_isa(list->elems[i], ch_cList)))
                goto fn_fail;

            param_expr_val = ch_obj_dup(list->elems[i]);
            param_expr = (ch_list_t*)param_expr_val.obj->data;

            if (param_expr->size != 3
                || param_expr->elems[0].type != CH_SYMBOL
                || (param_expr->elems[0].sym != SYM_op_$e.sym)
                || param_expr->elems[1].type != CH_SYMBOL) {
                if (i != list->size - 2
                    || param_expr->size != 2
                    || param_expr->elems[0].type != CH_SYMBOL
                    || param_expr->elems[0].sym != SYM_uop_$t.sym
                    || param_expr->elems[1].type != CH_SYMBOL)
                    goto fn_fail;

                /* Last argument is splat, that's fine */
                CH_SEND(result, push, param_expr_val);
                break;
            }

            param_expr->elems[2] = ch_macroexpand(param_expr->elems[2]);
            CH_SEND(result, push, param_expr_val);
            continue;

        fn_fail:
            ch_raise2(ch_eParseError,
                      "%s[] expects only identifiers and assignments, maybe splat",
                      form.sym + sizeof (long));
        }

        CH_SEND(result, push, ch_macroexpand(list->elems[list->size - 1]));
        return result;
    } else if (form.sym == SYM_slot.sym) {
        if (list->size != 3)
            ch_raise2(ch_eParseError, "slot[] expects 2 arguments");

        return ch_list_new1(3, SYM_slot,
                            ch_macroexpand(list->elems[1]), list->elems[2]);
    } else if (form.sym == SYM_setq.sym) {
        if (list->size != 3)
            ch_raise2(ch_eParseError, "setq[] expects 2 arguments");

        return ch_list_new1(3, SYM_setq, list->elems[1],
                            ch_macroexpand(list->elems[2]));
    } else if (form.sym == SYM_call.sym
               && CH_TRUEP(ch_defcalledp(list->elems[1]))) {
        return ch_macroexpand(ch_macroexpand1(expr));
    } else if (form.sym == SYM_call.sym
               || form.sym == SYM_if.sym
               || form.sym == SYM_while.sym
               || form.sym == SYM_unwind.sym
               || form.sym == SYM_unwind_protect.sym
               || form.sym == SYM__progn.sym
               || form.sym == SYM_not.sym
               || form.sym == SYM_set_slot.sym
               || form.sym == SYM_list.sym
               || form.sym == SYM_varslot.sym
               || form.sym == SYM_set_varslot.sym) {
        result = ch_list_new1(1, form);
        for (i = 1; i < list->size; i++)
            CH_SEND(result, push, ch_macroexpand(list->elems[i]));
        return result;
    } else {
        /* Not a builtin. Really macroexpand. :) */
        return ch_macroexpand(ch_macroexpand1(expr));
    }
}

ch_value_t ch_macroexpand(ch_value_t expression)
{
    if (CH_FALSEP(ch_isa(expression, ch_cList)))
        return expression;

    return macroexpand_list(expression);
}

static int in_scope_p(ch_value_t name, ch_value_t env)
{
    /* name: Symbol, env: List[List[Symbol|List[Symbol]]] */
    int i, j;
    ch_list_t *env_list, *scope_list, *var_list;

    env_list = (ch_list_t*)env.obj->data;
    for (i = 0; i < env_list->size; i++) {
        scope_list = (ch_list_t*)env_list->elems[i].obj->data;
        for (j = 0; j < scope_list->size; j++) {
            ch_value_t var;

            var = scope_list->elems[j];
            if (var.type == CH_SYMBOL && var.sym == name.sym)
                return 1;
            else if (CH_TRUEP(ch_isa(var, ch_cList))) {
                var_list = (ch_list_t*)var.obj->data;
                if (var_list->size < 2)
                    continue;

                var = var_list->elems[1];
                if (var.type == CH_SYMBOL && var.sym == name.sym)
                    return 1;
            }
        }
    }

    return 0;
}

static int find_var(ch_value_t name, int scope, ch_value_t env)
{
    ch_list_t *env_list, *scope_list, *var_list;
    int i;

    env_list = (ch_list_t*)env.obj->data;
    if (scope < 0)
        scope += env_list->size;

    if (scope < 0)
        ch_raise2(ch_eParseError, "scope index out of range");

    scope_list = (ch_list_t*)env_list->elems[scope].obj->data;
    for (i = 0; i < scope_list->size; i++) {
        ch_value_t var;

        var = scope_list->elems[i];
        if (var.type == CH_SYMBOL && var.sym == name.sym)
            return i;

        if (CH_TRUEP(ch_isa(var, ch_cList))) {
            var_list = (ch_list_t*)var.obj->data;
            if (var_list->size < 2)
                continue;

            var = var_list->elems[1];
            if (var.type == CH_SYMBOL && var.sym == name.sym)
                return i;
        }
    }

    return -1;
}

static void mark_outer(ch_value_t name, ch_value_t env)
{
    ch_list_t *env_list, *scope_list;
    int i, j;

    env_list = (ch_list_t*)env.obj->data;
    for (i = env_list->size - 1; i >= 0; i--) {
        ch_value_t scope;

        scope = env_list->elems[i];
        scope_list = (ch_list_t*)scope.obj->data;
        j = find_var(name, i, env);

        if (j != -1) {
            if (scope_list->elems[j].type == CH_SYMBOL)
                scope_list->elems[j] = ch_list_new1(2, SYM_captured, name);
            break;
        }

        CH_SEND(scope, push, ch_list_new1(2, SYM_outer, name));
    }
}

static ch_value_t flatten_scope1(ch_value_t, ch_value_t, ch_value_t);

static ch_value_t flatten_fn(ch_list_t *expr, ch_value_t env,
                             ch_value_t renames)
{
    ch_value_t params, shadow_dict, body, local, outer, result;
    ch_value_t restp, defaults;
    ch_list_t *env_list, *scope_list;
    int i;

    params = ch_list_new();
    shadow_dict = ch_obj_create(ch_cObject);
    restp = kFalse;
    defaults = ch_list_new();
    for (i = 1; i < expr->size - 1; i++) {
        if (expr->elems[i].type == CH_SYMBOL) {
            CH_SEND(params, push, expr->elems[i]);
            ch_setSlot(shadow_dict, expr->elems[i], kNil);
        } else {
            ch_list_t *l = (ch_list_t*)expr->elems[i].obj->data;
            CH_SEND(params, push, l->elems[1]);
            ch_setSlot(shadow_dict, l->elems[1], kNil);

            if (l->elems[0].sym == SYM_uop_$t.sym)
                restp = kTrue;
            else
                CH_SEND(defaults, push, flatten_scope1(l->elems[2],
                                                       env, renames));
        }
    }

    CH_SEND(renames, unshift, shadow_dict);
    CH_SEND(env, push, ch_obj_dup(params));

    /* Flattening body first so that info for locals/outers is accurate */
    body = flatten_scope1(expr->elems[expr->size - 1], env, renames);

    env_list = (ch_list_t*)env.obj->data;
    scope_list = (ch_list_t*)env_list->elems[env_list->size - 1].obj->data;

    /* Vars that are local */
    local = ch_list_new();
    outer = ch_list_new();
    for (i = 0; i < scope_list->size; i++) {
        ch_value_t var;
        ch_list_t *var_list;

        var = scope_list->elems[i];
        if (var.type == CH_SYMBOL) {
            CH_SEND(local, push, var);
            continue;
        }

        var_list = (ch_list_t*)var.obj->data;
        if (var_list->size == 2 && var_list->elems[0].type == CH_SYMBOL
            && var_list->elems[0].sym == SYM_outer.sym)
            CH_SEND(outer, push, var);
        else
            CH_SEND(local, push, var);
    }

    /* Params end up duped in local, oh well. */
    result = ch_list_new1(7, SYM_fn, params, defaults, restp, local, outer, body);

    CH_SEND(env, pop);
    CH_SEND(renames, shift);

    return result;
}

static ch_value_t flatten_global_scoped_let(ch_list_t *expr,
                                            ch_value_t env,
                                            ch_value_t renames)
{
    ch_value_t call, fn;
    int i;

    fn = ch_list_new1(1, SYM_fn);
    call = ch_list_new1(2, SYM_call, fn);
    for (i = 1; i < expr->size - 1; i++) {
        ch_value_t var;
        ch_list_t *var_list;

        var = expr->elems[i];
        if (var.type == CH_SYMBOL)
            var = ch_list_new1(3, SYM_op_$e, var, SYM_nil);

        var_list = (ch_list_t*)var.obj->data;
        if (CH_FALSEP(ch_isa(var, ch_cList)) || var_list->size != 3)
            ch_raise2(ch_eParseError, "weird binding in let[] at topmost scope");

        CH_SEND(fn, push, var_list->elems[1]);
        CH_SEND(call, push, var_list->elems[2]);
    }

    CH_SEND(fn, push, expr->elems[i]); /* Body of let is body of fn */

    return flatten_scope1(call, env, renames);
}

static ch_value_t flatten_let(ch_list_t *expr, ch_value_t env,
                              ch_value_t renames)
{
    ch_value_t result, my_renames, innermost_scope;
    ch_list_t *env_list;
    int i;

    my_renames = ch_obj_create(ch_cObject);

    env_list = (ch_list_t*)env.obj->data;
    if (env_list->size == 0)
        return flatten_global_scoped_let(expr, env, renames);

    innermost_scope = env_list->elems[env_list->size - 1];

    result = ch_list_new1(1, SYM_let);

    for (i = 1; i < expr->size - 1; i++) {
        ch_value_t clause, var, val, new_name;

        clause = expr->elems[i];
        if (clause.type == CH_SYMBOL) {
            var = clause;
            val = SYM_nil;
        } else {
            /* It's an assignment operator or macroexpand failed */
            ch_list_t *clause_list = (ch_list_t*)clause.obj->data;
            var = clause_list->elems[1];
            val = clause_list->elems[2];
        }

        if (find_var(var, -1, env) == -1) {
            new_name = var;
        } else {
            new_name = ch_gensym("collide");
            ch_setSlot(my_renames, var, new_name);
        }

        CH_SEND(innermost_scope, push, new_name);
        CH_SEND(result, push, ch_list_new1(
                    3,
                    SYM_op_$e, new_name, flatten_scope1(val, env, renames)));
    }

    CH_SEND(renames, unshift, my_renames);
    CH_SEND(result, push,
            flatten_scope1(expr->elems[expr->size - 1], env, renames));
    CH_SEND(renames, shift);

    return result;
}

static ch_value_t flatten_scope1(ch_value_t expr, ch_value_t env,
                                 ch_value_t renames)
{
    ch_list_t *expr_list;
    ch_value_t form, result;
    int i;

    if (expr.type == CH_SYMBOL) {
        ch_value_t mapping;
        ch_list_t *renames_list;

        mapping = kNil;
        renames_list = (ch_list_t*)renames.obj->data;
        for (i = 0; i < renames_list->size; i++) {
            if (ch_getSlot(renames_list->elems[i], expr, &mapping) == 0)
                break;
        }

        if (mapping.type != CH_NIL)
            expr = mapping;

        if (in_scope_p(expr, env) && find_var(expr, -1, env) == -1)
            mark_outer(expr, env);

        return expr;
    }

    if (CH_FALSEP(ch_isa(expr, ch_cList)))
        return expr;

    expr_list = (ch_list_t*)expr.obj->data;
    if (expr_list->size == 0)
        return expr; /* Screw that! */

    form = expr_list->elems[0];
    if (form.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "form not named by a symbol");

    if (form.sym == SYM_quote.sym)
        return expr;

    if (form.sym == SYM_fn.sym)
        return flatten_fn(expr_list, env, renames);

    if (form.sym == SYM_let.sym)
        return flatten_let(expr_list, env, renames);

    if (form.sym == SYM_send.sym) {
        result = ch_list_new1(1, SYM_send);
        CH_SEND(result, push, flatten_scope1(expr_list->elems[1], env, renames));
        CH_SEND(result, push, expr_list->elems[2]);
        for (i = 3; i < expr_list->size; i++)
            CH_SEND(result, push, flatten_scope1(expr_list->elems[i], env, renames));

        return result;
    }

    result = ch_list_new1(1, form);
    for (i = 1; i < expr_list->size; i++)
        CH_SEND(result, push,
                flatten_scope1(expr_list->elems[i], env, renames));

    return result;
}

static ch_value_t flatten_scope(ch_value_t expr)
{
    return flatten_scope1(expr, ch_list_new(), ch_list_new());
    return expr;
}

static void c_compile_in_context(ch_value_t expr,
                                 ch_value_t locals,
                                 ch_value_t outers,
                                 ch_value_t asm_);

#define EMIT(asm_, op, ...)                                             \
    do {                                                                \
        ch_value_t ch__operands__[] = { SYM_##op, ## __VA_ARGS__ };     \
        CH_SEND(asm_, push, ch_list_new2(                               \
                    sizeof (ch__operands__) / sizeof (ch_value_t),      \
                    ch__operands__));                                   \
        /*printf("DEBUG: %s with %d operands\n",                        \
                 SYM_##op.sym + sizeof (long),                          \
                 sizeof (ch__operands__) / sizeof (ch_value_t) - 1);*/  \
    } while (0)

static void c_compile_int(ch_value_t x, ch_value_t asm_)
{
    if (-(1 << 15) <= x.i && x.i < (1 << 15))
        EMIT(asm_, int2, x);
    else
        EMIT(asm_, int4, x);
}

static void c_compile_identifier(ch_value_t ident,
                                 ch_value_t locals,
                                 ch_value_t outers,
                                 ch_value_t asm_)
{
    int i;
    ch_list_t *list, *list2;

    if (ident.sym == SYM_true.sym) {
        EMIT(asm_, true);
    } else if (ident.sym == SYM_false.sym) {
        EMIT(asm_, false);
    } else if (ident.sym == SYM_nil.sym) {
        EMIT(asm_, nil);
    } else if (locals.type != CH_NIL) {
        list = (ch_list_t*)locals.obj->data;
        for (i = 0; i < list->size; i++) {
            if (list->elems[i].type == CH_SYMBOL) {
                if (list->elems[i].sym == ident.sym) {
                    EMIT(asm_, lget, ch_fix(i));
                    return;
                }
            } else if (CH_FALSEP(ch_isa(list->elems[i], ch_cList))) {
                ch_raise2(ch_eParseError, "non-symbol, non-list in locals");
            } else {
                /* [#captured, var] */
                list2 = (ch_list_t*)list->elems[i].obj->data;
                if (list2->size != 2
                    || list2->elems[0].type != CH_SYMBOL
                    || list2->elems[0].sym != SYM_captured.sym)
                    ch_raise2(ch_eParseError, "weird list in locals");

                if (list2->elems[1].type == CH_SYMBOL
                    && list2->elems[1].sym == ident.sym) {
                    EMIT(asm_, rget, ch_fix(i));
                    return;
                }
            }
        }

        /* Not local */
        if (outers.type != CH_NIL) {
            list = (ch_list_t*)outers.obj->data;
            for (i = 0; i < list->size; i++) {
                ch_list_t *outer_var;

                outer_var = (ch_list_t*)list->elems[i].obj->data;
                if (outer_var->elems[1].type == CH_SYMBOL
                    && outer_var->elems[1].sym == ident.sym) {
                    EMIT(asm_, oget, ch_fix(i));
                    return;
                }
            }
        }

        /* Not in lexical scope, ergo global */
        EMIT(asm_, gget, ident);
    } else {
        EMIT(asm_, gget, ident);
    }
}

static void c_compile_send(ch_list_t *expr,
                           ch_value_t locals,
                           ch_value_t outers,
                           ch_value_t asm_)
{
    int i;
    for (i = expr->size - 1; i > 2; i--)
        c_compile_in_context(expr->elems[i], locals, outers, asm_);

    c_compile_in_context(expr->elems[1], locals, outers, asm_);
    EMIT(asm_, send, ch_fix(expr->size - 3), expr->elems[2]);
}

static void c_compile_call(ch_list_t *expr,
                           ch_value_t locals,
                           ch_value_t outers,
                           ch_value_t asm_)
{
    int i;
    for (i = expr->size - 1; i > 0; i--)
        c_compile_in_context(expr->elems[i], locals, outers, asm_);

    EMIT(asm_, call, ch_fix(expr->size - 2));
}

static void c_compile_setq(ch_list_t *expr,
                           ch_value_t locals,
                           ch_value_t outers,
                           ch_value_t asm_)
{
    ch_value_t lval, rval;
    ch_list_t *env, *var;
    int i, flag;

    if (expr->size != 3)
        ch_raise2(ch_eParseError, "setq[] expects 2 arguments");

    lval = expr->elems[1];
    rval = expr->elems[2];

    if (lval.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "lvalue in setq[] not a symbol");

    c_compile_in_context(rval, locals, outers, asm_);

    if (locals.type != CH_NIL) {
        env = (ch_list_t*)locals.obj->data;
        flag = 0;
        for (i = 0; i < env->size; i++) {
            ch_value_t var_val;

            var_val = env->elems[i];
            if (var_val.type == CH_SYMBOL) {
                if (var_val.sym == lval.sym)
                    break;
            } else {
                var = (ch_list_t*)var_val.obj->data;
                var_val = var->elems[1];
                if (var_val.sym == lval.sym) {
                    flag = 1;
                    break;
                }
            }
        }

        if (i < env->size) {
            /* found it in locals */
            if (flag)
                /* It's captured */
                EMIT(asm_, rset, ch_fix(i));
            else
                EMIT(asm_, lset, ch_fix(i));

            return;
        }

        env = (ch_list_t*)outers.obj->data;
        for (i = 0; i < env->size; i++) {
            ch_value_t var_name;

            var = (ch_list_t*)env->elems[i].obj->data;
            var_name = var->elems[1];
            if (var_name.sym == lval.sym)
                break;
        }

        if (i < env->size) {
            /* Found it in outers */
            EMIT(asm_, oset, ch_fix(i));
            return;
        }
    }

    /* When all else fails, it's a global */
    EMIT(asm_, gset, lval);
}

static void c_compile_set_slot(ch_list_t *expr,
                               ch_value_t locals,
                               ch_value_t outers,
                               ch_value_t asm_)
{
    ch_value_t lval_val, rval, obj, slot;
    ch_list_t *lval;

    if (expr->size != 3)
        ch_raise2(ch_eParseError, "set_slot[] expects two arguments");

    lval_val = expr->elems[1];
    rval = expr->elems[2];

    if (CH_FALSEP(ch_isa(lval_val, ch_cList)))
        goto bad_lval;

    lval = (ch_list_t*)lval_val.obj->data;
    if (lval->size < 3 || lval->elems[0].type != CH_SYMBOL
        || lval->elems[0].sym != SYM_slot.sym)
        goto bad_lval;

    obj = lval->elems[1];
    slot = lval->elems[2];

    if (slot.type != CH_SYMBOL)
        ch_raise2(ch_eParseError,
                  "slot name in slot[] in set_slot[] lvalue not a symbol");

    c_compile_in_context(rval, locals, outers, asm_);
    c_compile_in_context(obj, locals, outers, asm_);
    EMIT(asm_, set_attr, slot);
    return;

bad_lval:
    ch_raise2(ch_eParseError, "set_slot[] expects a slot[] as lvalue");
}

static void c_compile_varslot(ch_list_t *expr,
                              ch_value_t locals,
                              ch_value_t outers,
                              ch_value_t asm_)
{
    if (expr->size != 3)
        ch_raise2(ch_eParseError, "varslot[] expects 2 arguments");

    c_compile_in_context(expr->elems[1], locals, outers, asm_);
    c_compile_in_context(expr->elems[2], locals, outers, asm_);
    EMIT(asm_, pop_get_attr);
}

static void c_compile_set_varslot(ch_list_t *expr,
                                  ch_value_t locals,
                                  ch_value_t outers,
                                  ch_value_t asm_)
{
    if (expr->size != 4)
        ch_raise2(ch_eParseError, "set_varslot[] expects 3 arguments");

    c_compile_in_context(expr->elems[3], locals, outers, asm_);
    c_compile_in_context(expr->elems[1], locals, outers, asm_);
    c_compile_in_context(expr->elems[2], locals, outers, asm_);
    EMIT(asm_, pop_set_attr);
}

static void c_compile_if(ch_list_t *expr,
                         ch_value_t locals,
                         ch_value_t outers,
                         ch_value_t asm_)
{
    ch_value_t then_label, end_label;

    if (expr->size < 3)
        ch_raise2(ch_eParseError, "if[] expects at least 2 arguments");

    then_label = ch_gensym("then");
    end_label = ch_gensym("endif");

    c_compile_in_context(expr->elems[1], locals, outers, asm_);
    EMIT(asm_, if, then_label);

    if (expr->size == 3)
        EMIT(asm_, nil);
    else if (expr->size == 4)
        c_compile_in_context(expr->elems[3], locals, outers, asm_);
    else {
        ch_value_t else_clause;

        if (expr->elems[3].type == CH_SYMBOL
            && expr->elems[3].sym == SYM_else.sym) {
            else_clause = expr->elems[4];
        } else {
            else_clause = ch_list_new2(expr->size - 3, &expr->elems[3]);
            CH_SEND(else_clause, unshift, SYM_if);
        }

        c_compile_in_context(else_clause, locals, outers, asm_);
    }

    EMIT(asm_, goto, end_label);
    EMIT(asm_, label, then_label);

    c_compile_in_context(expr->elems[2], locals, outers, asm_);

    EMIT(asm_, label, end_label);
}

static void c_compile_fn(ch_list_t *expr,
                         ch_value_t locals,
                         ch_value_t outers,
                         ch_value_t asm_)
{
    ch_list_t *fnargs_list, *fnlocals_list, *fnouters_list, *cap_inds, *defaults;
    ch_value_t fn_label, end_label;
    ch_value_t fnargs, fnlocals, fnouters, fnbody, checkargs, cap_inds_val;
    int i, restp;

    fn_label = ch_gensym("fn");
    end_label = ch_gensym("endfn");

    fnargs = expr->elems[1];
    fnargs_list = (ch_list_t*)fnargs.obj->data;
    defaults = (ch_list_t*)expr->elems[2].obj->data;
    restp = expr->elems[3].i ? 1 : 0;
    fnlocals = expr->elems[4];
    fnlocals_list = (ch_list_t*)fnlocals.obj->data;
    fnouters = expr->elems[5];
    fnouters_list = (ch_list_t*)fnouters.obj->data;
    fnbody = expr->elems[6];

    cap_inds_val = ch_list_new();
    cap_inds = (ch_list_t*)cap_inds_val.obj->data;
    for (i = 0; i < fnlocals_list->size; i++)
        if (fnlocals_list->elems[i].type != CH_SYMBOL)
            CH_SEND(cap_inds_val, push, ch_fix(i));

    EMIT(asm_, goto, end_label);
    EMIT(asm_, label, fn_label);

    checkargs = ch_list_new1(
        6,
        SYM_checkargs,
        ch_fix(fnargs_list->size - defaults->size - restp),
        ch_fix(defaults->size),
        ch_fix(restp),
        ch_fix(fnlocals_list->size - fnargs_list->size),
        ch_fix(cap_inds->size));

    CH_SEND(checkargs, concat$b, cap_inds_val);

    // EMIT
    CH_SEND(asm_, push, checkargs);

    c_compile_in_context(fnbody, fnlocals, fnouters, asm_);

    EMIT(asm_, ret);
    EMIT(asm_, label, end_label);

    /* Take care of outers */
    for (i = fnouters_list->size - 1; i >= 0; i--) {
        ch_list_t *var, *scope_var, *scope_list;
        ch_value_t var_name, scope_var_name;
        int j;

        var = (ch_list_t*)fnouters_list->elems[i].obj->data;
        var_name = var->elems[1];

        /* If var is in outers, it's "outer outer", use oref. */
        scope_list = (ch_list_t*)outers.obj->data;
        for (j = 0; j < scope_list->size; j++) {
            scope_var = (ch_list_t*)scope_list->elems[j].obj->data;
            scope_var_name = scope_var->elems[1];

            if (var_name.sym == scope_var_name.sym)
                break;
        }

        if (j < scope_list->size) {
            EMIT(asm_, oref, ch_fix(j));
            continue;
        }

        /* Else find it in locals, use lget */
        scope_list = (ch_list_t*)locals.obj->data;
        for (j = 0; j < scope_list->size; j++) {
            if (scope_list->elems[j].type == CH_SYMBOL)
                continue;

            scope_var = (ch_list_t*)scope_list->elems[j].obj->data;
            scope_var_name = scope_var->elems[1];

            if (var_name.sym == scope_var_name.sym)
                break;
        }

        /* Not there, wtf? */
        if (j >= scope_list->size)
            ch_raise2(ch_eParseError, "freak scoping accident in fn[]");

        EMIT(asm_, lget, ch_fix(j));
    }

    /* Take care of default values */
    for (i = 0; i < defaults->size; i++)
        c_compile_in_context(defaults->elems[i], locals, outers, asm_);

    EMIT(asm_, fn, fn_label, ch_fix(defaults->size),
         ch_fix(fnouters_list->size));
}

static void c_compile_let(ch_list_t *expr,
                          ch_value_t locals,
                          ch_value_t outers,
                          ch_value_t asm_)
{
    ch_value_t progn;
    ch_list_t *progn_list, *locals_list;
    int i;

    locals_list = (ch_list_t*)locals.obj->data;
    for (i = 1; i < expr->size - 1; i++) {
        ch_value_t binding, var;
        ch_list_t *binding_list;
        int var_index;

        binding = expr->elems[i];
        binding_list = (ch_list_t*)binding.obj->data;
        var = binding_list->elems[1];

        for (var_index = 0; var_index < locals_list->size; var_index++) {
            ch_list_t *l;

            if (locals_list->elems[var_index].type == CH_SYMBOL) {
                if (locals_list->elems[var_index].sym == var.sym)
                    break;
                else
                    continue;
            }

            l = (ch_list_t*)locals_list->elems[var_index].obj->data;
            if (l->elems[1].type == CH_SYMBOL && l->elems[1].sym == var.sym)
                break;
        }

        if (var_index >= locals_list->size)
            ch_raise2(ch_eParseError, "freak scoping accident in let[]");

        EMIT(asm_, rebind, ch_fix(var_index));
    }

    progn = ch_list_new2(expr->size, expr->elems);
    progn_list = (ch_list_t*)progn.obj->data;
    progn_list->elems[0] = SYM__progn;
    for (i = 1; i < progn_list->size - 1; i++) {
        ch_list_t *binding_list = (ch_list_t*)progn_list->elems[i].obj->data;
        progn_list->elems[i] = ch_list_new1(3, SYM_setq,
                                            binding_list->elems[1],
                                            binding_list->elems[2]);
    }

    c_compile_in_context(progn, locals, outers, asm_);
}

static void c_compile_while(ch_list_t *expr,
                            ch_value_t locals,
                            ch_value_t outers,
                            ch_value_t asm_)
{
    ch_value_t test, body, test_label, end_label;

    if (expr->size != 3)
        ch_raise2(ch_eParseError, "while[] expects two arguments");

    test = expr->elems[1];
    body = expr->elems[2];

    test_label = ch_gensym("whiletest");
    end_label = ch_gensym("endwhile");

    EMIT(asm_, label, test_label);
    c_compile_in_context(test, locals, outers, asm_);
    EMIT(asm_, not);
    EMIT(asm_, if, end_label);
    c_compile_in_context(body, locals, outers, asm_);
    EMIT(asm_, pop);
    EMIT(asm_, goto, test_label);
    EMIT(asm_, label, end_label);
    EMIT(asm_, nil);
}

static void c_compile_slot(ch_list_t *expr,
                           ch_value_t locals,
                           ch_value_t outers,
                           ch_value_t asm_)
{
    ch_value_t obj, slotname;

    if (expr->size != 3)
        ch_raise2(ch_eParseError, "slot[] expects two arguments");

    if (expr->elems[2].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "second argument to slot[] not an identifier");

    obj = expr->elems[1];
    slotname = expr->elems[2];

    c_compile_in_context(obj, locals, outers, asm_);
    EMIT(asm_, get_attr, slotname);
}

static void c_compile_unwind(ch_list_t *expr,
                             ch_value_t locals,
                             ch_value_t outers,
                             ch_value_t asm_)
{
    if (expr->size != 1)
        ch_raise2(ch_eParseError, "unwind[] expects no arguments");

    EMIT(asm_, uwnd);
}

static void c_compile_unwind_protect(ch_list_t *expr,
                                     ch_value_t locals,
                                     ch_value_t outers,
                                     ch_value_t asm_)
{
    ch_value_t handler_label, end_label;

    if (expr->size != 3)
        ch_raise2(ch_eParseError, "unwind_protect[] expects two arguments");

    handler_label = ch_gensym("handler");
    end_label = ch_gensym("enduwndp");

    EMIT(asm_, uwndp, handler_label, end_label);
    c_compile_in_context(expr->elems[1], locals, outers, asm_);
    EMIT(asm_, ret);
    EMIT(asm_, label, handler_label);
    c_compile_in_context(expr->elems[2], locals, outers, asm_);
    EMIT(asm_, label, end_label);
}

static void c_compile_not(ch_list_t *expr,
                          ch_value_t locals,
                          ch_value_t outers,
                          ch_value_t asm_)
{
    if (expr->size != 2)
        ch_raise2(ch_eParseError, "not[] expects one argument");

    c_compile_in_context(expr->elems[1], locals, outers, asm_);
    EMIT(asm_, not);
}

static void c_compile__progn(ch_list_t *expr,
                             ch_value_t locals,
                             ch_value_t outers,
                             ch_value_t asm_)
{
    int i;

    if (expr->size == 1)
        EMIT(asm_, nil);

    for (i = 1; i < expr->size; i++) {
        if (i > 1)
            EMIT(asm_, pop);

        c_compile_in_context(expr->elems[i], locals, outers, asm_);
    }
}

static void c_compile_quote(ch_list_t *expr,
                            ch_value_t locals,
                            ch_value_t outers,
                            ch_value_t asm_)
{
    if (expr->size != 2)
        ch_raise2(ch_eParseError, "quote[] expects one argument");

    if (expr->elems[1].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "quote[] is not smart enough for that");

    EMIT(asm_, sym, expr->elems[1]);
}

static void c_compile_list_form(ch_list_t *expr,
                                ch_value_t locals,
                                ch_value_t outers,
                                ch_value_t asm_)
{
    int i;

    for (i = expr->size - 1; i >= 1; i--)
        c_compile_in_context(expr->elems[i], locals, outers, asm_);

    if (expr->size == 1)
        EMIT(asm_, liste);
    else
        EMIT(asm_, list, ch_fix(expr->size - 1));
}

static void c_compile_list(ch_value_t expr_val,
                           ch_value_t locals,
                           ch_value_t outers,
                           ch_value_t asm_)
{
    ch_list_t *expr;
    ch_value_t form;

    expr = (ch_list_t*)expr_val.obj->data;
    if (expr->size == 0)
        ch_raise2(ch_eParseError, "zero-length list in AST");

    form = expr->elems[0];
    if (form.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "list in AST doesn't start with symbol");

#define FORM(name, ...)                                 \
    else if (form.sym == SYM_##name.sym)                \
        c_compile_##name(expr, locals, outers, asm_)

    if (0);
    FORM(send);
    FORM(call);
    FORM(if);
    FORM(fn);
    FORM(while);
    FORM(_progn);
    FORM(quote);
    FORM(slot);
    FORM(let);
    FORM(setq);
    FORM(set_slot);
    FORM(unwind);
    FORM(unwind_protect);
    FORM(not);
    FORM(varslot);
    FORM(set_varslot);
    else if (form.sym == SYM_list.sym)
        c_compile_list_form(expr, locals, outers, asm_);
    else
        ch_raise2(ch_eParseError,
                  "macroexpansion missed %s", form.sym + sizeof (long));

#undef FORM
}

static void c_compile_in_context(ch_value_t expr,
                                 ch_value_t locals,
                                 ch_value_t outers,
                                 ch_value_t asm_)
{
    if (expr.type == CH_NIL)
        ch_raise2(ch_eParseError, "nil AST");

    if (expr.type == CH_SYMBOL)
        c_compile_identifier(expr, locals, outers, asm_);
    else if (expr.type == CH_INT)
        c_compile_int(expr, asm_);
    else if (expr.type == CH_FLOAT)
        EMIT(asm_, float, expr);
    else if (CH_TRUEP(ch_isa(expr, ch_cString)))
        EMIT(asm_, string, expr);
    else if (CH_FALSEP(ch_isa(expr, ch_cList)))
        ch_raise2(ch_eParseError, "AST is none of: symbol, int, float, list");
    else
        c_compile_list(expr, locals, outers, asm_);
}

static ch_value_t c_compile_flattened_expr(ch_value_t expr)
{
    ch_value_t asm_;

    asm_ = ch_list_new();
    c_compile_in_context(expr, kNil, kNil, asm_);
    return asm_;
}

static void c_assemble1(ch_list_t *op,
                        ch_value_t code,
                        ch_value_t strings,
                        ch_ht_t *labels,
                        ch_ht_t *label_uses,
                        ch_value_t label_list)
{
    ch_value_t opcode, n_opcode, align;
    long code_size;

    opcode = op->elems[0];
    code_size = CH_SEND(code, size).i;

    if (opcode.sym == SYM_label.sym) {
        ch_value_t name = op->elems[1];

        if (ch_ht_hasKeyp(labels, name))
            ch_raise2(ch_eParseError, "repeated label in asm: %s",
                      name.sym + sizeof (long));

        ch_ht_set(labels, name, ch_fix(code_size / 2));
        CH_SEND(label_list, push, name);
        return;
    }

    /* It is an actual instruction, not a label */
    if (ch_ht_get(opcodes, opcode, &n_opcode) < 0)
        ch_raise2(ch_eParseError, "unrecognized opcode: %s",
                  opcode.sym + sizeof(long));

    if (ch_ht_get(operand_align, opcode, &align) < 0)
        align = ch_fix(2);

    while ((code_size + 2) % align.i != 0) {
        /* Append nop */
        char nop[] = "\0\0";
        code_size += 2;
        CH_SEND(code, concat$b, ch_string_new1(nop));
    }

    /* Append the opcode */
    CH_SEND(code, concat$b, ch_string_new2(2, (char*)&n_opcode.i));

    /* Now worry about operands */
    if (opcode.sym == SYM_int2.sym) {
        short s = (short)op->elems[1].i;
        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&s));
    } else if (opcode.sym == SYM_int4.sym) {
        int n = (int)op->elems[1].i;
        CH_SEND(code, concat$b, ch_string_new2(4, (char*)&n));
    } else if (opcode.sym == SYM_int8.sym) {
        ch_raise2(ch_eParseError, "unimplemented opcode: int8");
    } else if (opcode.sym == SYM_float.sym) {
        double d = (double)op->elems[1].f;
        CH_SEND(code, concat$b, ch_string_new2(8, (char*)&d));
    } else if (opcode.sym == SYM_goto.sym || opcode.sym == SYM_if.sym) {
        /* Takes a label */
        ch_value_t uses;
        unsigned label_placeholder;

        if (ch_ht_get(label_uses, op->elems[1], &uses) < 0) {
            uses = ch_list_new();
            ch_ht_set(label_uses, op->elems[1], uses);
        }

        label_placeholder = 0;
        CH_SEND(uses, push, CH_SEND2(code, size));
        CH_SEND(code, concat$b, ch_string_new2(4, (char*)&label_placeholder));
    } else if (opcode.sym == SYM_fn.sym) {
        /* Takes a label and 2 numbers */
        ch_value_t uses;
        unsigned label_placeholder;
        unsigned short ndefault_values, nouter_vars;

        if (ch_ht_get(label_uses, op->elems[1], &uses) < 0) {
            uses = ch_list_new();
            ch_ht_set(label_uses, op->elems[1], uses);
        }

        label_placeholder = 0;
        ndefault_values = (unsigned short)op->elems[2].i;
        nouter_vars = (unsigned short)op->elems[3].i;
        CH_SEND(uses, push, CH_SEND2(code, size));
        CH_SEND(code, concat$b, ch_string_new2(4, (char*)&label_placeholder));
        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&ndefault_values));
        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&nouter_vars));
    } else if (opcode.sym == SYM_checkargs.sym) {
        int i;

        for (i = 1; i < 5; i++) {
            unsigned char c = (unsigned char)op->elems[i].i;
            CH_SEND(code, concat$b, ch_string_new2(1, (char*)&c));
        }

        for (; i < op->size; i++) {
            unsigned short s = (unsigned short)op->elems[i].i;
            CH_SEND(code, concat$b, ch_string_new2(2, (char*)&s));
        }
    } else if (opcode.sym == SYM_string.sym
               || opcode.sym == SYM_sym.sym
               || opcode.sym == SYM_gset.sym
               || opcode.sym == SYM_gget.sym
               || opcode.sym == SYM_set_attr.sym
               || opcode.sym == SYM_get_attr.sym
               || opcode.sym == SYM_sendn.sym) {
        unsigned short i;
        ch_value_t as_string;
        ch_list_t *strings_list;

        as_string = op->elems[1].type == CH_SYMBOL
            ? ch_string_new1(op->elems[1].sym + sizeof (long))
            : op->elems[1];

        strings_list = (ch_list_t*)strings.obj->data;
        for (i = 0; i < strings_list->size; i++)
            if (CH_TRUEP(CH_SEND(strings_list->elems[i], $e$e, as_string)))
                break;

        if (i == strings_list->size)
            CH_SEND(strings, push, as_string);

        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&i));
    } else if (opcode.sym == SYM_list.sym
               || opcode.sym == SYM_dict.sym
               || opcode.sym == SYM_call.sym
               || opcode.sym == SYM_lget.sym
               || opcode.sym == SYM_lset.sym
               || opcode.sym == SYM_rget.sym
               || opcode.sym == SYM_rset.sym
               || opcode.sym == SYM_oget.sym
               || opcode.sym == SYM_oset.sym
               || opcode.sym == SYM_oref.sym
               || opcode.sym == SYM_rebind.sym) {
        unsigned short s = (unsigned short)op->elems[1].i;
        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&s));
    } else if (opcode.sym == SYM_send.sym) {
        unsigned short i, s;
        ch_value_t as_string;
        ch_list_t *strings_list;

        s = (unsigned short)op->elems[1].i;
        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&s));

        as_string = op->elems[2].type == CH_SYMBOL
            ? ch_string_new1(op->elems[2].sym + sizeof (long))
            : op->elems[2];

        strings_list = (ch_list_t*)strings.obj->data;
        for (i = 0; i < strings_list->size; i++)
            if (CH_TRUEP(CH_SEND(strings_list->elems[i], $e$e, as_string)))
                break;

        if (i == strings_list->size)
            CH_SEND(strings, push, as_string);

        CH_SEND(code, concat$b, ch_string_new2(2, (char*)&i));
    } else if (opcode.sym == SYM_uwndp.sym) {
        ch_value_t uses;
        unsigned int placeholder = 0;

        if (ch_ht_get(label_uses, op->elems[1], &uses) < 0) {
            uses = ch_list_new();
            ch_ht_set(label_uses, op->elems[1], uses);
        }
        CH_SEND(uses, push, CH_SEND2(code, size));
        CH_SEND(code, concat$b, ch_string_new2(4, (char*)&placeholder));

        if (ch_ht_get(label_uses, op->elems[2], &uses) < 0) {
            uses = ch_list_new();
            ch_ht_set(label_uses, op->elems[2], uses);
        }
        CH_SEND(uses, push, CH_SEND2(code, size));
        CH_SEND(code, concat$b, ch_string_new2(4, (char*)&placeholder));
    } else {
        /* Nothing. No operands */
    }
}

static ch_value_t c_assemble(ch_list_t *asm_)
{
    int i;
    ch_value_t code, strings, label_list_val;
    ch_list_t *label_list, *strings_list;
    ch_ht_t *labels, *label_uses;
    char **bcstrings;
    uint32_t *strlens;
    unsigned short ret_opcode;

    code = ch_string_new();
    strings = ch_list_new();
    labels = ch_ht_create(NULL, NULL);
    label_uses = ch_ht_create(NULL, NULL);
    label_list_val = ch_list_new();
    label_list = (ch_list_t*)label_list_val.obj->data;

    for (i = 0; i < asm_->size; i++)
        c_assemble1((ch_list_t*)asm_->elems[i].obj->data,
                    code, strings, labels, label_uses, label_list_val);

    ch_string_t *codestr;
    codestr = (ch_string_t*)code.obj->data;

    for (i = 0; i < label_list->size; i++) {
        unsigned int label_addr;
        ch_value_t addr_val, uses_val;
        ch_list_t *uses;
        int j;

        ch_ht_get(labels, label_list->elems[i], &addr_val);
        label_addr = (unsigned int)addr_val.i;

        ch_ht_get(label_uses, label_list->elems[i], &uses_val);
        uses = (ch_list_t*)uses_val.obj->data;

        for (j = 0; j < uses->size; j++)
            memcpy(&codestr->text[uses->elems[j].i], &label_addr, 4);
    }

    strings_list = (ch_list_t*)strings.obj->data;
    bcstrings = ch_malloc(strings_list->size * sizeof (char*));
    strlens = ch_malloc(strings_list->size * sizeof (uint32_t));
    for (i = 0; i < strings_list->size; i++) {
        ch_string_t *s = (ch_string_t*)strings_list->elems[i].obj->data;
        bcstrings[i] = s->text;
        strlens[i] = s->size;
    }

    ret_opcode = OP_RET;
    CH_SEND(code, concat$b, ch_string_new2(2, (char*)&ret_opcode));

    return ch_bytecode_new(codestr->size / 2, (uint16_t*)codestr->text,
                           strings_list->size, bcstrings, strlens,
                           ch_current_module);
}

static ch_value_t c_compile_top(compiler_t *cpl)
{
    ch_value_t code, asm_;
    ch_list_t *code_list;
    int i;

    /* Maybe leading terminators */
    if (p_next_is_term(cpl->p))
        p_terminators(cpl->p);

    /* No expressions in the input stream. Strange, but OK. */
    if (p_peek_is(cpl->p, SYM_eof))
        code = ch_list_new1(2, SYM_progn, SYM_nil);
    else {
        /* Read expressions one at a time */
        code = ch_list_new1(2, SYM_progn, ch_macroexpand(p_expr(cpl->p)));
        while (p_next_is_term(cpl->p)) {
            p_terminators(cpl->p);
            if (p_peek_is(cpl->p, SYM_eof)
                || p_peek_is(cpl->p, SYM_rbracket)
                || p_peek_is(cpl->p, SYM_comma))
                break;

            CH_SEND(code, push, ch_macroexpand(p_expr(cpl->p)));
        }
    }

    /* Macroexpand for progn */
    code = ch_macroexpand(code);

    /* Maybe extra trailing terminators */
    if (p_next_is_term(cpl->p))
        p_terminators(cpl->p);

    /* TODO: optimize? */

    code_list = (ch_list_t*)code.obj->data;
    for (i = 1; i < code_list->size; i++)
        code_list->elems[i] = flatten_scope(code_list->elems[i]);

    asm_ = c_compile_flattened_expr(code);
    return c_assemble((ch_list_t*)asm_.obj->data);
}

static ch_value_t compile_from_parser(parser_t *p)
{
    compiler_t cpl;
    cpl.p = p;

    return c_compile_top(&cpl);
}

ch_value_t ch_compile_stdio_file(FILE *fp)
{
    parser_t *p;

    p = p_new((void*)fp,
              (int(*)(void*))fgetc,
              (int(*)(int,void*))ungetc);

    return compile_from_parser(p);
}

ch_value_t ch_compile_file(const char* filename)
{
    FILE *fp;
    ch_value_t result;

    fp = fopen(filename, "r");
    if (!fp)
        return kNil;

    result = ch_compile_stdio_file(fp);

    fclose(fp);
    return result;
}

#define SIO_MAX_UNGOT 16
typedef struct {
    const char *source;
    int len, pos;
    int ungot[SIO_MAX_UNGOT];
    int n_ungot;
} stringio_t;

static int stringio_getc(stringio_t *sio)
{
    if (sio->n_ungot > 0)
        return sio->ungot[--sio->n_ungot];
    else if (sio->pos >= sio->len)
        return EOF;
    else
        return sio->source[sio->pos++];
}

static int stringio_ungetc(int c, stringio_t *sio)
{
    if (sio->n_ungot >= SIO_MAX_UNGOT)
        return EOF;

    sio->ungot[sio->n_ungot++] = c;
    return c;
}

ch_value_t ch_evalString(const char *source)
{
    parser_t *p;
    stringio_t sio;
    ch_value_t bytecode;

    sio.source = source;
    sio.len = strlen(source);
    sio.pos = 0;
    sio.n_ungot = 0;

    p = p_new((void*)&sio, (int(*)(void*))stringio_getc,
              (int(*)(int,void*))stringio_ungetc);

    bytecode = compile_from_parser(p);
    ch_runBytecode(bytecode);
    return ch_pop();
}

ch_value_t ch_loadSharedObject(const char *filename)
{
    void *so = dlopen(filename, RTLD_NOW);
    if (!so)
        ch_raise2(ch_eNameError, "problem opening %s: %s",
                  filename, ch_string_new1(dlerror()));

    void *loadfn = dlsym(so, "chai_load_module");
    if (!loadfn) {
        dlclose(so);
        ch_raise2(ch_eNameError, "%s has no chai_load_module() function",
                  filename);
    }

    void (*f)() = (void(*)())loadfn;
    f();

    /* Can't believe I had a dlclose() here. Modules never get
     * unloaded, so never dlclose() them. */

    return kNil;
}

ch_value_t ch_evalFile(const char *filename)
{
    ch_value_t bytecode = ch_compile_file(filename);
    ch_runBytecode(bytecode);
    return ch_pop();
}

ch_value_t ch_load(const char *name)
{
    char buf[8192];
    ch_value_t path_value;
    ch_list_t *path;
    int i, len;
    struct stat sb;

    path_value = ch_CHAIPATH;

    if (CH_FALSEP(ch_isa(path_value, ch_cList)))
        ch_raise2(ch_eTypeError, "CHAIPATH is not a list");

    len = strlen(name);
    path = (ch_list_t*)path_value.obj->data;
    for (i = 0; i < path->size; i++) {
        const char *s;
        const char *prefixdir;

        if (CH_FALSEP(ch_isa(path->elems[i], ch_cString)))
            ch_raise2(ch_eTypeError, "non-string element in CHAIPATH");

        s = CH_CSTR(path->elems[i]);
        prefixdir = s[0] == '/' ? "" : ch_main_script_dir;
        snprintf(buf, sizeof (buf), "%s/%s/%s", prefixdir, s, name);

        if (!stat(buf, &sb))
            break;

        if (len < 3 || strcmp(&name[len - 3], ".ch")) {
            snprintf(buf, sizeof (buf), "%s/%s/%s.ch", prefixdir, s, name);
            if (!stat(buf, &sb))
                break;
        }

        if (len < 3 || strcmp(&name[len - 3], ".so")) {
            snprintf(buf, sizeof (buf), "%s/%s/%s.so", prefixdir, s, name);
            if (!stat(buf, &sb))
                break;
        }

        if (len < 6 || strcmp(&name[len - 6], ".dylib")) {
            snprintf(buf, sizeof (buf), "%s/%s/%s.dylib", prefixdir, s, name);
            if (!stat(buf, &sb))
                break;
        }
    }

    if (i == path->size)
        ch_raise2(ch_eNameError, "can't find %s in CHAIPATH", name);

    len = strlen(buf);
    if (len > 3 && !strcmp(&buf[len - 3], ".so"))
        return ch_loadSharedObject(buf);
    else if (len > 6 && !strcmp(&buf[len - 6], ".dylib"))
        return ch_loadSharedObject(buf);
    else
        return ch_evalFile(buf);
}

static ch_value_t macroexpand_$e(long argc, ch_value_t *argv)
{
    int i;
    ch_value_t result;

    result = ch_list_new1(1, SYM_setf);
    for (i = 0; i < argc; i++)
        CH_SEND(result, push, argv[i]);

    return result;
}

static ch_value_t macroexpand_or(long argc, ch_value_t *argv)
{
    ch_value_t x;

    if (argc == 0)
        return SYM_false;
    else if (argc == 1)
        return argv[0];

    x = ch_gensym("_");

    return ch_list_new1(
        3,
        SYM_let,
        ch_list_new1(3, SYM_op_$e, x, argv[0]),
        ch_list_new1(
            4,
            SYM_if, x,
            x,
            macroexpand_or(argc - 1, &argv[1])));
}

static ch_value_t macroexpand_and(long argc, ch_value_t *argv)
{
    if (argc == 0)
        return SYM_true;
    else if (argc == 1)
        return argv[0];

    return ch_list_new1(
        4,
        SYM_if, argv[0],
        macroexpand_and(argc - 1, &argv[1]),
        SYM_false);
}

static ch_value_t macroexpand_not(long argc, ch_value_t *argv)
{
    if (argc != 1)
        ch_raise2(ch_eParseError, "op_not[] expects one argument");

    return ch_list_new1(2, SYM_not, argv[0]);
}

static ch_value_t macroexpand_$b$e(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_!=[] expects two arguments");

    return ch_list_new1(2, SYM_not,
                        ch_list_new1(3, SYM_op_$e$e, argv[0], argv[1]));
}

static ch_value_t macroexpand_$D(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_.[] expects two arguments");

    return ch_list_new1(3, SYM_send, argv[0], argv[1]);
}

static ch_value_t macroexpand_at(long argc, ch_value_t *argv)
{
    if (argc != 1)
        ch_raise2(ch_eParseError, "op_@[] expects one argument");

    return ch_list_new1(3, SYM_slot, SYM_self, argv[0]);
}

static ch_value_t macroexpand_$p(long argc, ch_value_t *argv)
{
    if (argc == 1)
        return ch_list_new1(3, SYM_send, argv[0], SYM_unary$p);
    else if (argc == 2)
        return ch_list_new1(4, SYM_send, argv[0], SYM_$p, argv[1]);
    else
        ch_raise2(ch_eParseError, "op_+[] expects one or two arguments");

    return kNil;
}

static ch_value_t macroexpand_$m(long argc, ch_value_t *argv)
{
    if (argc == 1)
        return ch_list_new1(3, SYM_send, argv[0], SYM_unary$m);
    else if (argc == 2)
        return ch_list_new1(4, SYM_send, argv[0], SYM_$m, argv[1]);
    else
        ch_raise2(ch_eParseError, "op_+[] expects one or two arguments");

    return kNil;
}

static ch_value_t macroexpand_$t(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_*[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$t, argv[1]);
}

static ch_value_t macroexpand_$d(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_/[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$d, argv[1]);
}

static ch_value_t macroexpand_$d$d(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_//[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$d$d, argv[1]);
}

static ch_value_t macroexpand_$r(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_%[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$r, argv[1]);
}

static ch_value_t macroexpand_$e$e(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_==[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$e$e, argv[1]);
}

static ch_value_t macroexpand_$l(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_<[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$l, argv[1]);
}

static ch_value_t macroexpand_$l$e(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_<=[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$l$e, argv[1]);
}

static ch_value_t macroexpand_$g(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_>[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$g, argv[1]);
}

static ch_value_t macroexpand_$g$e(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_>=[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$g$e, argv[1]);
}

static ch_value_t macroexpand_$t$t(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_**[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$t$t, argv[1]);
}

static ch_value_t macroexpand_$a(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_&[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$a, argv[1]);
}

static ch_value_t macroexpand_$o(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_|[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$o, argv[1]);
}

static ch_value_t macroexpand_$x(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_^[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$x, argv[1]);
}

static ch_value_t macroexpand_$n(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_~[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$n, argv[1]);
}

static ch_value_t macroexpand_$l$l(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_<<[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$l$l, argv[1]);
}

static ch_value_t macroexpand_$g$g(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_>>[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$g$g, argv[1]);
}

static ch_value_t macroexpand_$D$D(long argc, ch_value_t *argv)
{
    if (argc != 2)
        ch_raise2(ch_eParseError, "op_..[] expects two arguments");

    return ch_list_new1(4, SYM_send, argv[0], SYM_$D$D, argv[1]);
}

static ch_value_t macroexpand_try(long argc, ch_value_t *argv)
{
    ch_value_t body, result;
    ch_list_t *result_list;
    int i;

    if (argc < 3 || argc % 2 != 1)
        ch_raise2(ch_eParseError, "try[] expects an odd no. of args >= 3");

    if (argv[argc - 2].type == CH_SYMBOL
        && argv[argc - 2].sym == SYM_ensure.sym) {
        ch_value_t uwndp_body;

        if (argc == 3)
            uwndp_body = argv[0];
        else {
            uwndp_body = ch_list_new1(1, SYM_try);
            for (i = 0; i < argc - 2; i++)
                CH_SEND(uwndp_body, push, argv[i]);
        }

        return ch_list_new1(
            3,
            SYM_progn,
            ch_list_new1(
                3,
                SYM_unwind_protect,
                uwndp_body,
                ch_list_new1(3,
                             SYM_progn,
                             argv[argc - 1],
                             ch_list_new1(1, SYM_unwind))),
            argv[argc - 1]);
    }

    body = argv[0];
    result = ch_list_new1(3, SYM_unwind_protect, body,
                          ch_list_new1(1, SYM_unwind));
    result_list = (ch_list_t*)result.obj->data;

    for (i = 0; i < argc / 2; i++) {
        int index;
        ch_value_t exc_typ, exc_name;

        index = argc - 2 * (i + 1);
        exc_name = kNil;
        exc_typ = argv[index];

        if (CH_TRUEP(ch_isa(exc_typ, ch_cList))) {
            ch_list_t *exc_typ_list = (ch_list_t*)exc_typ.obj->data;
            if (exc_typ_list->size != 3
                || exc_typ_list->elems[0].type != CH_SYMBOL
                || exc_typ_list->elems[0].sym != SYM_op_$e$g.sym
                || exc_typ_list->elems[2].type != CH_SYMBOL)
                ch_raise2(ch_eParseError, "invalid syntax for try clause, "
                          "expect ExcType or ExcType : name");

            exc_typ = exc_typ_list->elems[1];
            exc_name = exc_typ_list->elems[2];
        }

        result_list->elems[2] = ch_list_new1(
            4,
            SYM_if,
            ch_list_new1(4, SYM_send, SYM_$_exc, SYM_is_a$q, exc_typ),
            exc_name.type == CH_NIL ? argv[index + 1] : ch_list_new1(
                3,
                SYM_let,
                ch_list_new1(3, SYM_op_$e, exc_name, SYM_$_exc),
                argv[index + 1]),
            result_list->elems[2]);
    }

    return result;
}

static ch_value_t macroexpand_raise(long argc, ch_value_t *argv)
{
    if (argc != 1)
        ch_raise2(ch_eParseError, "raise expects one argument");

    return ch_list_new1(
        4,
        SYM_progn,
        ch_list_new1(3, SYM_setq, SYM_$_exc, argv[0]),
        ch_list_new1(3, SYM_set_slot,
                     ch_list_new1(3, SYM_slot, SYM_$_exc, SYM_backtrace),
                     ch_list_new1(2, SYM_call, SYM_backtrace)),
        ch_list_new1(1, SYM_unwind));
}

static ch_value_t macroexpand_import(long argc, ch_value_t *argv)
{
    if (argc < 1 || argc > 2 || argv[0].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "_import expects 1-2 identifiers");

    if (argc == 2 && argv[1].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "_import expects 1-2 identifiers");

    ch_value_t imported = ch_load_module(argv[0].sym + sizeof (long));
    ch_module_import(ch_current_module, imported, argv[argc - 1]);

    return SYM_nil;
}

static ch_value_t macroexpand_import_name(long argc, ch_value_t *argv)
{
    if (argc != 2 || argv[0].type != CH_SYMBOL || argv[1].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "_import_name expects 2 identifiers");

    ch_value_t module = ch_load_module(argv[0].sym + sizeof (long));
    ch_module_import_name_from(ch_current_module, argv[1], module);

    return SYM_nil;
}

static ch_value_t macroexpand_import_all(long argc, ch_value_t *argv)
{
    if (argc != 1 || argv[0].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "_import_all expects 2 identifiers");

    ch_value_t module = ch_load_module(argv[0].sym + sizeof (long));
    ch_module_import_all_from(ch_current_module, module);

    return SYM_nil;
}

static ch_value_t macroexpand_use_ops(long argc, ch_value_t *argv)
{
    int i;

    if (argc < 1 || argv[0].type != CH_SYMBOL)
        ch_raise2(ch_eParseError,
                  "use_ops[] expects one identifier followed by any "
                  "number of strings");

    for (i = 1; i < argc; i++)
        if (CH_FALSEP(ch_isa(argv[i], ch_cString)))
            ch_raise2(ch_eParseError,
                      "use_ops[] expects one identifier followed by any "
                      "number of strings");

    ch_value_t from;
    ch_module_t *cur = (ch_module_t*)ch_current_module.obj->data;
    if (ch_ht_get(&cur->imports, argv[0], &from) < 0)
        ch_raise2(ch_eParseError, "no module %s under module %s",
                  argv[0].sym + sizeof (long),
                  cur->module_name.sym + sizeof (long));

    if (argc == 1)
        ch_module_import_all_ops_from(ch_current_module, from);
    else
        for (i = 1; i < argc; i++) {
            ch_value_t op_name = CH_SEND(argv[i], intern);
            ch_module_import_op_from(ch_current_module, op_name, from);
            ch_module_import_uop_from(ch_current_module, op_name, from);
        }

    return SYM_nil;
}

static ch_value_t macroexpand_setf(long argc, ch_value_t *argv)
{
    ch_value_t name, setf_defn;
    ch_list_t *lval;

    if (argc != 2)
        ch_raise2(ch_eParseError, "setf[] expects two arguments");

    if (argv[0].type == CH_SYMBOL)
        return ch_list_new1(3, SYM_setq, argv[0], argv[1]);

    if (CH_FALSEP(ch_isa(argv[0], ch_cList)))
        ch_raise2(ch_eParseError, "nonsymbol, nonlist AST for lval in setf[]");

    lval = (ch_list_t*)argv[0].obj->data;
    if (lval->size == 0)
        ch_raise2(ch_eParseError, "zero-size AST list as lval in setf[]");

    name = lval->elems[0];
    if (name.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "improperly named form as lval in setf[]");

    if (ch_ht_get(setf_impls, name, &setf_defn) < 0)
        ch_raise2(ch_eParseError, "lval (%s) in setf[] is not defsetfed?()",
                  name.sym + sizeof (long));

    /* They take lval, rval */
    return ch_funcall(setf_defn, 2, argv[0], argv[1]);
}

static ch_value_t setf_slot(ch_value_t lval, ch_value_t rval)
{
    return ch_list_new1(3, SYM_set_slot, lval, rval);
}

static ch_value_t setf_send(ch_value_t lval, ch_value_t rval)
{
    ch_value_t tmp, tgt, msg, assign_send;
    ch_list_t *lval_list;
    char new_msg[1024];

    lval_list = (ch_list_t*)lval.obj->data;
    if (lval_list->size < 3)
        ch_raise2(ch_eParseError, "malformed send[] as lval in setf[]");

    tgt = lval_list->elems[1];
    msg = lval_list->elems[2];
    tmp = ch_gensym("rval");

    if (msg.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "malformed send[] as lval in setf[]: nonsymbol method");

    snprintf(new_msg, sizeof (new_msg), "%s=", msg.sym + sizeof (long));

    assign_send = ch_list_new1(
        3,
        SYM_send,
        tgt,
        ch_internCString(new_msg));

    CH_SEND(assign_send, concat$b,
            ch_list_new2(lval_list->size - 3, &lval_list->elems[3]));
    CH_SEND(assign_send, push, tmp);

    return ch_list_new1(
        3,
        SYM_let,
        ch_list_new1(3, SYM_op_$e, tmp, rval),
        ch_list_new1(
            3,
            SYM_progn,
            assign_send,
            tmp));
}

static ch_value_t setf_op_$D(ch_value_t lval, ch_value_t rval)
{
    return ch_list_new1(3, SYM_setf, ch_list_new1(2, SYM_call, lval), rval);
}

static ch_value_t setf_call(ch_value_t lval, ch_value_t rval)
{
    ch_list_t *lval_list;
    ch_value_t func, call;

    lval_list = (ch_list_t*)lval.obj->data;
    if (lval_list->size < 2)
        ch_raise2(ch_eParseError, "lvalue AST is malformed call[]");

    func = lval_list->elems[1];
    if (CH_TRUEP(ch_defcalledp(func)))
        return ch_list_new1(3, SYM_setf, ch_macroexpand1(lval), rval);

    call = ch_list_new1(
        3,
        SYM_send,
        func,
        SYM__call_);

    CH_SEND(call, concat$b,
            ch_list_new2(lval_list->size - 2, &lval_list->elems[2]));

    return ch_list_new1(
        3,
        SYM_setf,
        call,
        rval);
}

ch_value_t ch_my_bindings_for(ch_value_t expr)
{
    if (expr.type == CH_SYMBOL)
        return ch_list_new1(1, expr);

    ch_list_t *explist = (ch_list_t*)expr.obj->data;
    if (CH_FALSEP(ch_isa(expr, ch_cList))
        || explist->size == 0 || explist->elems[0].type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "bad argument to uop_my[]");

    ch_value_t form = explist->elems[0];
    ch_value_t impl;
    if (ch_ht_get(ch_my_bindings, form, &impl) < 0)
        ch_raise2(ch_eParseError, "no my bindings defined for form %s",
                  form.sym + sizeof (long));

    return ch_vfuncall(impl, 1, &expr);
}

static ch_value_t macroexpand_progn(long argc, ch_value_t *args)
{
    ch_list_t *stmt, *assign;
    ch_value_t my_vars, result_progn, result;
    int i, got_one;

    result_progn = ch_list_new1(1, SYM__progn);
    my_vars = ch_list_new();
    got_one = 0;

    for (i = 0; i < argc; i++) {
        if (CH_FALSEP(ch_isa(args[i], ch_cList))) {
            CH_SEND(result_progn, push, args[i]);
            continue;
        }

        stmt = (ch_list_t*)args[i].obj->data;
        if (stmt->size == 2 && stmt->elems[0].type == CH_SYMBOL
            && stmt->elems[0].sym == SYM_uop_my.sym) {

            got_one = 1;

            if (stmt->elems[1].type == CH_SYMBOL) {
                CH_SEND(my_vars, push, stmt->elems[1]);
                continue;
            }

            assign = (ch_list_t*)stmt->elems[1].obj->data;
            if (CH_FALSEP(ch_isa(stmt->elems[1], ch_cList))
                || assign->elems[0].type != CH_SYMBOL)
                ch_raise2(ch_eParseError, "progn[]: bad argument to uop_my[]");

            if (assign->elems[0].sym == SYM_op_$e.sym) {
                CH_SEND(my_vars, concat$b,
                        ch_my_bindings_for(assign->elems[1]));
                CH_SEND(result_progn, push, stmt->elems[1]);
            } else {
                CH_SEND(my_vars, concat$b, ch_my_bindings_for(stmt->elems[1]));
            }
        } else {
            CH_SEND(result_progn, push, args[i]);
        }
    }

    if (got_one) {
        result = ch_list_new1(1, SYM_let);
        CH_SEND(result, concat$b, my_vars);
        CH_SEND(result, push, result_progn);
        return result;
    }

    return result_progn;
}

void ch_defmacro(ch_value_t name, ch_value_t impl)
{
    ch_module_t *m = (ch_module_t*)ch_current_module.obj->data;

    ch_ht_set(&m->macros, name, impl);

    char macro_name[BUFSIZ];
    snprintf(macro_name, sizeof (macro_name), "%s[]", name.sym + sizeof (long));
    ch_setSlot(impl, SYM_name, ch_internCString(macro_name));
}

void ch_defmacro_cfn(ch_value_t name, ch_cfn_t cfn)
{
    ch_value_t val;
    ch_prim_t *prim;

    val = ch_obj_create(ch_cPrimitive);
    prim = (ch_prim_t*)val.obj->data;
    prim->arity = -1;
    prim->impl = cfn;

    ch_defmacro(name, val);
}

static ch_value_t call_op_$D(ch_value_t f, ch_value_t args)
{
    ch_value_t result;
    ch_list_t *dot_op;

    dot_op = (ch_list_t*)f.obj->data;
    if (dot_op->size < 3)
        ch_raise2(ch_eParseError,
                  "call[] for op_.[] getting too small an op_.[]");

    result = ch_list_new1(3, SYM_send, dot_op->elems[1], dot_op->elems[2]);
    CH_SEND(result, concat$b, args);
    return result;
}

static void ch_call_impls_init()
{
    call_impls = ch_ht_create(NULL, NULL);
    ch_ht_set(call_impls, SYM_op_$D, ch_prim_new(2, (ch_cfn_t)call_op_$D));
}

static void ch_setf_impls_init()
{
    setf_impls = ch_ht_create(NULL, NULL);

    ch_ht_set(setf_impls, SYM_slot, ch_prim_new(2, (ch_cfn_t)setf_slot));
    ch_ht_set(setf_impls, SYM_send, ch_prim_new(2, (ch_cfn_t)setf_send));
    ch_ht_set(setf_impls, SYM_call, ch_prim_new(2, (ch_cfn_t)setf_call));
    ch_ht_set(setf_impls, SYM_op_$D, ch_prim_new(2, (ch_cfn_t)setf_op_$D));
}

static ch_value_t macroexpand_defcall(ch_value_t name, ch_value_t f,
                                      ch_value_t args, ch_value_t body)
{
    ch_value_t fn_ast, asm_, bytecode, fn;

    fn_ast = ch_list_new1(4, SYM_fn, f, args, body);
    fn_ast = ch_macroexpand(fn_ast);
    asm_ = c_compile_flattened_expr(flatten_scope(fn_ast));
    bytecode = c_assemble((ch_list_t*)asm_.obj->data);

    ch_runBytecode(bytecode);
    fn = ch_pop();

    ch_ht_set(call_impls, name, fn);

    return SYM_nil;
}

static ch_value_t macroexpand_defsetf(ch_value_t name, ch_value_t lval,
                                      ch_value_t rval, ch_value_t body)
{
    ch_value_t fn_ast, asm_, bytecode, fn;

    fn_ast = ch_list_new1(4, SYM_fn, lval, rval, body);
    fn_ast = ch_macroexpand(fn_ast);
    asm_ = c_compile_flattened_expr(flatten_scope(fn_ast));
    bytecode = c_assemble((ch_list_t*)asm_.obj->data);

    ch_runBytecode(bytecode);
    fn = ch_pop();

    ch_ht_set(setf_impls, name, fn);

    return SYM_nil;
}

static ch_value_t macroexpand_defmacro(long argc, ch_value_t *argv)
{
    ch_list_t *form;
    ch_value_t fn_ast, asm_, bytecode, fn, name;

    if (argc != 2)
        ch_raise2(ch_eParseError, "defmacro[] expects 2 arguments");

    form = (ch_list_t*)argv[0].obj->data;
    if (CH_FALSEP(ch_isa(argv[0], ch_cList)) || form->size == 0)
        ch_raise2(ch_eParseError, "defmacro[] expects a compound form "
                  "as its first argument");

    name = form->elems[0];
    if (name.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "defmacro[] expects first argument form "
                  "to be named by a symbol");

    fn_ast = ch_obj_dup(argv[0]);
    CH_SEND(fn_ast, push, argv[1]);

    form = (ch_list_t*)fn_ast.obj->data;
    form->elems[0] = SYM_fn;

    fn_ast = ch_macroexpand(fn_ast);
    asm_ = c_compile_flattened_expr(flatten_scope(fn_ast));
    bytecode = c_assemble((ch_list_t*)asm_.obj->data);

    ch_runBytecode(bytecode);
    fn = ch_pop();

    ch_defmacro(name, fn);

    return SYM_nil;
}

void ch_deftagstr(ch_value_t name, ch_value_t f)
{
    ch_ht_set(ch_tagstrs, name, f);
}

/*  */
static ch_value_t macroexpand_deftagstr(long argc, ch_value_t *argv)
{
    ch_value_t fn_ast, asm_, bytecode, fn, name, paramname;

    if (argc != 3)
        ch_raise2(ch_eParseError, "deftagstr[] expects 2 arguments");

    name = argv[0];
    if (name.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "deftagstr[] expects an identifier tag "
                  "as its first argument");

    paramname = argv[1];
    if (paramname.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "deftagstr[] expects an identifier "
                  "parameter name as its second argument");

    fn_ast = ch_list_new1(3, SYM_fn, paramname, argv[2]);
    fn_ast = ch_macroexpand(fn_ast);
    asm_ = c_compile_flattened_expr(flatten_scope(fn_ast));
    bytecode = c_assemble((ch_list_t*)asm_.obj->data);

    ch_runBytecode(bytecode);
    fn = ch_pop();

    ch_deftagstr(name, fn);

    return SYM_nil;
}


/* This is where things get *really* crazy */
static ch_value_t macroexpand_defleftprec(ch_value_t prec)
{
    ch_value_t existing;
    ch_module_t *m;

    if (prec.type != CH_INT)
        ch_raise2(ch_eParseError, "defleftprec[] expects an int");

    m = (ch_module_t*)ch_current_module.obj->data;
    if (!ch_ht_get(&m->assoc_table, prec, &existing)
        && existing.sym != SYM_left.sym)

        ch_raise2(ch_eParseError, "precedence %d is already defined as %s",
                  prec.i, existing.sym + sizeof (long));

    ch_ht_set(&m->assoc_table, prec, SYM_left);
    return SYM_nil;
}

static ch_value_t macroexpand_defrightprec(ch_value_t prec)
{
    ch_value_t existing;
    ch_module_t *m;

    if (prec.type != CH_INT)
        ch_raise2(ch_eParseError, "defrightprec[] expects an int");

    m = (ch_module_t*)ch_current_module.obj->data;
    if (!ch_ht_get(&m->assoc_table, prec, &existing)
        && existing.sym != SYM_right.sym)

        ch_raise2(ch_eParseError, "precedence %d is already defined as %s",
                  prec.i, existing.sym + sizeof (long));

    ch_ht_set(&m->assoc_table, prec, SYM_right);
    return SYM_nil;
}

static ch_value_t macroexpand_defuprec(ch_value_t prec)
{
    ch_value_t existing;
    ch_module_t *m;

    if (prec.type != CH_INT)
        ch_raise2(ch_eParseError, "defuprec[] expects an int");

    m = (ch_module_t*)ch_current_module.obj->data;
    if (!ch_ht_get(&m->assoc_table, prec, &existing)
        && existing.sym != SYM_unary.sym)

        ch_raise2(ch_eParseError, "precedence %d is already defined as %s",
                  prec.i, existing.sym + sizeof (long));

    ch_ht_set(&m->assoc_table, prec, SYM_unary);
    return SYM_nil;
}

static ch_value_t macroexpand_defop(ch_value_t op, ch_value_t prec)
{
    char *op_text;
    ch_value_t existing, op_sym, assoc;
    ch_module_t *m;
    int i;

    m = (ch_module_t*)ch_current_module.obj->data;

    if (CH_FALSEP(ch_isa(op, ch_cString)))
        ch_raise2(ch_eParseError, "defop[] expects a string operator name");

    if (prec.type != CH_INT && CH_FALSEP(ch_isa(prec, ch_cString)))
        ch_raise2(ch_eParseError, "defop[] expects an int or string precedence");

    /* If the given precedence is a string, use the same as that operator */
    if (prec.type != CH_INT)
        ch_ht_get(&m->prec_table, ch_internCString(CH_CSTR(prec)), &prec);

    op_text = CH_CSTR(op);
    op_sym = ch_internCString(op_text);
    if (!ch_ht_get(&m->prec_table, op_sym, &existing))
        ch_raise2(ch_eParseError, "operator %s is already defined with "
                  "precedence %d",
                  CH_CSTR(op), existing.i);

    if (ch_ht_get(&m->assoc_table, prec, &assoc) < 0
        || assoc.sym == SYM_unary.sym)
        ch_raise2(ch_eParseError, "unknown binary precedence: %d", prec.i);

    ch_ht_set(&m->prec_table, op_sym, prec);

    /* Check if parse_op_idents needs to be updated */
    for (i = 0; op_text[i]; i++) {
        if ((op_text[i] < 'A' || op_text[i] > 'Z') &&
            (op_text[i] < 'a' || op_text[i] > 'z') &&
            (op_text[i] < '0' || op_text[i] > '9') &&
            ((op_text[i] != '!' && op_text[i] != '?') || op_text[i+1] != '\0'))
            break;
    }

    if (!op_text[i])
        /* it's an identifier if we fell off */
        ch_ht_set(&m->op_idents, op_sym, kTrue);

    return SYM_nil;
}

static ch_value_t macroexpand_defuop(ch_value_t op, ch_value_t prec)
{
    char *op_text;
    ch_value_t existing, op_sym, assoc;
    ch_module_t *m;
    int i;

    m = (ch_module_t*)ch_current_module.obj->data;

    if (CH_FALSEP(ch_isa(op, ch_cString)))
        ch_raise2(ch_eParseError, "defop[] expects a string operator name");

    if (prec.type != CH_INT && CH_FALSEP(ch_isa(prec, ch_cString)))
        ch_raise2(ch_eParseError, "defop[] expects an int or string precedence");

    /* If the given precedence is a string, use the same as that operator */
    if (prec.type != CH_INT)
        ch_ht_get(&m->uprec_table, ch_internCString(CH_CSTR(prec)), &prec);

    op_text = CH_CSTR(op);
    op_sym = ch_internCString(op_text);
    if (!ch_ht_get(&m->uprec_table, op_sym, &existing))
        ch_raise2(ch_eParseError, "operator %s is already defined with "
                  "precedence %d",
                  CH_CSTR(op), existing.i);

    if (ch_ht_get(&m->assoc_table, prec, &assoc) < 0
        || assoc.sym != SYM_unary.sym)
        ch_raise2(ch_eParseError, "unknown unary precedence: %d", prec.i);

    ch_ht_set(&m->uprec_table, op_sym, prec);

    /* Check if parse_op_idents needs to be updated */
    for (i = 0; op_text[i]; i++) {
        if ((op_text[i] < 'A' || op_text[i] > 'Z') &&
            (op_text[i] < 'a' || op_text[i] > 'z') &&
            (op_text[i] < '0' || op_text[i] > '9') &&
            op_text[i] != '_' &&
            ((op_text[i] != '!' && op_text[i] != '?') || op_text[i+1] != '\0'))
            break;
    }

    if (!op_text[i])
        /* it's an identifier if we fell off */
        ch_ht_set(&m->op_idents, op_sym, kTrue);

    return SYM_nil;
}

static ch_value_t macroexpand_defmybindings(ch_value_t form,
                                            ch_value_t param,
                                            ch_value_t body)
{
    if (form.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "defmybindings[] expects an identifier first");

    if (param.type != CH_SYMBOL)
        ch_raise2(ch_eParseError, "defmybindings[] expects an identifier second");

    ch_value_t fn_ast = ch_list_new1(3, SYM_fn, param, body);
    fn_ast = ch_macroexpand(fn_ast);

    ch_value_t asm_ = c_compile_flattened_expr(flatten_scope(fn_ast));
    ch_value_t bytecode = c_assemble((ch_list_t*)asm_.obj->data);

    ch_runBytecode(bytecode);
    ch_value_t fn = ch_pop();

    ch_ht_set(ch_my_bindings, form, fn);
    return SYM_nil;
}

static ch_value_t macroexpand_cur_module(long argc, ch_value_t *argv)
{
    if (argc != 0)
        ch_raise2(ch_eParseError, "current_module{} wants no arguments");

    ch_value_t name;
    name = ((ch_module_t*)ch_current_module.obj->data)->module_name;

    return ch_list_new1(2, SYM_quote, name);
}

static void ch_macros_init()
{
    ch_value_t old_cur_mod = ch_current_module;
    ch_module_t *m = (ch_module_t*)ch_m_builtins.obj->data;
    ch_current_module = ch_m_builtins;

    ch_defmacro_cfn(SYM_progn, macroexpand_progn);
    ch_my_bindings = ch_ht_create(NULL, NULL);
    ch_ht_set(&m->macros, SYM_defmybindings,
              ch_prim_new(3, (ch_cfn_t)macroexpand_defmybindings));

    ch_defmacro_cfn(SYM_defmacro, macroexpand_defmacro);
    ch_defmacro_cfn(SYM_deftagstr, macroexpand_deftagstr);
    ch_ht_set(&m->macros, SYM_defcall,
              ch_prim_new(4, (ch_cfn_t)macroexpand_defcall));
    ch_ht_set(&m->macros, SYM_defsetf,
              ch_prim_new(4, (ch_cfn_t)macroexpand_defsetf));

    ch_ht_set(&m->macros, SYM_defleftprec,
              ch_prim_new(1, (ch_cfn_t)macroexpand_defleftprec));
    ch_ht_set(&m->macros, SYM_defrightprec,
              ch_prim_new(1, (ch_cfn_t)macroexpand_defrightprec));
    ch_ht_set(&m->macros, SYM_defuprec,
              ch_prim_new(1, (ch_cfn_t)macroexpand_defuprec));

    ch_ht_set(&m->macros, SYM_defop,
              ch_prim_new(2, (ch_cfn_t)macroexpand_defop));
    ch_ht_set(&m->macros, SYM_defuop,
              ch_prim_new(2, (ch_cfn_t)macroexpand_defuop));

    ch_defmacro_cfn(SYM__import, macroexpand_import);
    ch_defmacro_cfn(SYM__import_name, macroexpand_import_name);
    ch_defmacro_cfn(SYM__import_all, macroexpand_import_all);
    ch_defmacro_cfn(SYM_use_ops, macroexpand_use_ops);
    ch_defmacro_cfn(SYM_op_$e, macroexpand_$e);
    ch_defmacro_cfn(SYM_op_or, macroexpand_or);
    ch_defmacro_cfn(SYM_op_and, macroexpand_and);
    ch_defmacro_cfn(SYM_op_$o$o, macroexpand_or);
    ch_defmacro_cfn(SYM_op_$a$a, macroexpand_and);
    ch_defmacro_cfn(SYM_uop_not, macroexpand_not);
    ch_defmacro_cfn(SYM_uop_$b, macroexpand_not);
    ch_defmacro_cfn(SYM_op_$b$e, macroexpand_$b$e);
    ch_defmacro_cfn(SYM_op_$D, macroexpand_$D);
    ch_defmacro_cfn(SYM_op_$A, macroexpand_at);
    ch_defmacro_cfn(SYM_op_$p, macroexpand_$p);
    ch_defmacro_cfn(SYM_op_$m, macroexpand_$m);
    ch_defmacro_cfn(SYM_uop_$p, macroexpand_$p);
    ch_defmacro_cfn(SYM_uop_$m, macroexpand_$m);
    ch_defmacro_cfn(SYM_op_$t, macroexpand_$t);
    ch_defmacro_cfn(SYM_op_$d, macroexpand_$d);
    ch_defmacro_cfn(SYM_op_$d$d, macroexpand_$d$d);
    ch_defmacro_cfn(SYM_op_$r, macroexpand_$r);
    ch_defmacro_cfn(SYM_op_$e$e, macroexpand_$e$e);
    ch_defmacro_cfn(SYM_op_$l, macroexpand_$l);
    ch_defmacro_cfn(SYM_op_$l$e, macroexpand_$l$e);
    ch_defmacro_cfn(SYM_op_$g, macroexpand_$g);
    ch_defmacro_cfn(SYM_op_$g$e, macroexpand_$g$e);
    ch_defmacro_cfn(SYM_op_$t$t, macroexpand_$t$t);
    ch_defmacro_cfn(SYM_op_$a, macroexpand_$a);
    ch_defmacro_cfn(SYM_op_$o, macroexpand_$o);
    ch_defmacro_cfn(SYM_op_$x, macroexpand_$x);
    ch_defmacro_cfn(SYM_op_$n, macroexpand_$n);
    ch_defmacro_cfn(SYM_op_$l$l, macroexpand_$l$l);
    ch_defmacro_cfn(SYM_op_$g$g, macroexpand_$g$g);
    ch_defmacro_cfn(SYM_op_$D$D, macroexpand_$D$D);

    ch_defmacro_cfn(SYM_try, macroexpand_try);
    ch_defmacro_cfn(SYM_raise, macroexpand_raise);
    ch_defmacro_cfn(SYM_setf, macroexpand_setf);

    ch_defmacro_cfn(SYM_current_module, macroexpand_cur_module);

    ch_current_module = old_cur_mod;
}

static void ch_opcodes_init()
{
    int i;

    opcodes = ch_ht_create(NULL, NULL);

#define DEFOP(name)  ch_ht_set(opcodes, SYM_##name, ch_fix(i++))

    i = 0;
    DEFOP(nop);
    DEFOP(pop);
    DEFOP(dup);
    DEFOP(int2);
    DEFOP(int4);
    DEFOP(int8);
    DEFOP(float);
    DEFOP(string);
    DEFOP(sym);
    DEFOP(nil);
    DEFOP(true);
    DEFOP(false);
    DEFOP(liste);
    DEFOP(dicte);
    DEFOP(list);
    DEFOP(listn);
    DEFOP(dict);
    DEFOP(dictn);
    DEFOP(fn);
    DEFOP(goto);
    DEFOP(if);
    DEFOP(gset);
    DEFOP(gget);
    DEFOP(lset);
    DEFOP(lget);
    DEFOP(rset);
    DEFOP(rget);
    DEFOP(oset);
    DEFOP(oget);
    DEFOP(oref);
    DEFOP(get_attr);
    DEFOP(set_attr);
    DEFOP(pop_get_attr);
    DEFOP(pop_set_attr);
    DEFOP(call);
    DEFOP(send);
    DEFOP(calln);
    DEFOP(sendn);
    DEFOP(checkargs);
    DEFOP(rebind);
    DEFOP(ret);
    DEFOP(not);
    DEFOP(uwnd);
    DEFOP(uwndp);

#undef DEFOP

    operand_align = ch_ht_create(NULL, NULL);

    ch_ht_set(operand_align, SYM_int4, ch_fix(4));
    ch_ht_set(operand_align, SYM_int8, ch_fix(8));
    ch_ht_set(operand_align, SYM_float, ch_fix(8));
    ch_ht_set(operand_align, SYM_fn, ch_fix(4));
    ch_ht_set(operand_align, SYM_goto, ch_fix(4));
    ch_ht_set(operand_align, SYM_if, ch_fix(4));
    ch_ht_set(operand_align, SYM_send, ch_fix(4));
    ch_ht_set(operand_align, SYM_uwndp, ch_fix(4));
}

void ch_compiler_init()
{
    ch_macros_init();
    ch_opcodes_init();
    ch_call_impls_init();
    ch_setf_impls_init();

    ch_tagstrs = ch_ht_create(NULL, NULL);
}
