/* ht.c -- chai hash tables
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "ht.h"
#include "mem.h"
#include "obj.h"
#include "sym_literals.h"
#include <stdio.h>
#include <string.h>

#define DEFAULT_HT_CAP 4 /* _always_ a power of two. */
#define HT_MAX_DENSITY 1.333
#define HT_MIN_DENSITY 0.333

/* Definitions. */
void ch_ht_init(ch_ht_t* ht,
                ch_hashFunction_t hfn,
                ch_eqFunction_t eq)
{
    ht->cap = DEFAULT_HT_CAP;
    ht->mask = ht->cap - 1;
    ht->size = 0;
    ht->hfn = hfn ? hfn : ch_hashSym;
    ht->eq = eq ? eq : ch_shallowEq;

    ht->buckets = ch_malloc(ht->cap * sizeof(ch_hashBucket_t*));
}

ch_ht_t* ch_ht_create(ch_hashFunction_t hfn, ch_eqFunction_t eq)
{
    ch_ht_t* ht = NULL;
    ht = ch_malloc(sizeof(ch_ht_t));

    ch_ht_init(ht, hfn, eq);

    return ht;
}

void ch_ht_mergeb(ch_ht_t *ht, ch_ht_t *from)
{
    int i;
    ch_hashBucket_t *bucket;

    for (i = 0; i < from->cap; i++) {
        bucket = from->buckets[i];
        while (bucket) {
            ch_ht_set(ht, bucket->key, bucket->value);
            bucket = bucket->next;
        }
    }
}

/* Rehashing... newCap should _always_ be a power of two. */
void ch_ht_rehash(ch_ht_t *ht, size_t new_cap)
{
    int i;
    size_t new_mask;
    ch_hashBucket_t **new_buckets;

    new_cap = new_cap > DEFAULT_HT_CAP ? new_cap : DEFAULT_HT_CAP;
    if (new_cap == ht->cap)
        return;

    new_mask = new_cap - 1;

    new_buckets = ch_malloc(new_cap * sizeof(ch_hashBucket_t*));

    for (i = 0; i < ht->cap; i++) {
        ch_hashBucket_t *bucket;

        bucket = ht->buckets[i];
        while (bucket) {
            ch_hashBucket_t *new_bucket;
            size_t index;

            new_bucket = ch_malloc(sizeof(ch_hashBucket_t));
            index = ht->hfn(bucket->key) & new_mask;

            new_bucket->key = bucket->key;
            new_bucket->value = bucket->value;
            new_bucket->next = new_buckets[index];

            new_buckets[index] = new_bucket;

            bucket = bucket->next;
        }
    }

    ht->buckets = new_buckets;
    ht->cap = new_cap;
    ht->mask = new_mask;
}

/* Predicates */
int ch_ht_hasKeyp(ch_ht_t* ht, ch_value_t key)
{
    size_t i = ht->hfn(key) & ht->mask;

    ch_hashBucket_t* bucket = ht->buckets[i];

    while(bucket) {
        if(ht->eq(bucket->key, key))
            return 1;

        bucket = bucket->next;
    }

    return 0;
}

int ch_ht_emptyp(ch_ht_t* ht)
{
    return ht->size == 0;
}

/* USING the hash table. */
void ch_ht_clear(ch_ht_t* ht)
{
    ch_hashBucket_t **b;

    b = ch_malloc(DEFAULT_HT_CAP * sizeof(ch_hashBucket_t*));
    ht->buckets = b;
    ht->cap = DEFAULT_HT_CAP;
    ht->mask = ht->cap - 1;
    ht->size = 0;
}

void ch_ht_set(ch_ht_t *ht, ch_value_t key, ch_value_t value)
{
    ch_hashBucket_t *bucket, *new_bucket;
    size_t i;

    i = ht->hfn(key) & ht->mask;
    bucket = ht->buckets[i];

    while (bucket) {
        if (ht->eq(bucket->key, key)) {
            bucket->value = value;
            return;
        }
        bucket = bucket->next;
    }

    /* If we haven't returned yet, we need to create a new entry. */
    new_bucket = ch_malloc(sizeof(ch_hashBucket_t));
    new_bucket->key = key;
    new_bucket->value = value;
    new_bucket->next = ht->buckets[i];
    ht->buckets[i] = new_bucket;

    if (++ht->size > HT_MAX_DENSITY * ht->cap)
        ch_ht_rehash(ht, ht->cap * 2);
}

int ch_ht_get(ch_ht_t *ht, ch_value_t key, ch_value_t* result)
{
    size_t i;
    ch_hashBucket_t *bucket;

    i = ht->hfn(key) & ht->mask;
    bucket = ht->buckets[i];

    while (bucket) {
        if (ht->eq(bucket->key, key)) {
            *result = bucket->value;
            return 0;
        }
        bucket = bucket->next;
    }

    *result = kNil;
    return -1;
}

int ch_ht_delete(ch_ht_t *ht, ch_value_t key)
{
    size_t i;
    ch_hashBucket_t* bucket;

    i = ht->hfn(key) & ht->mask;
    bucket = ht->buckets[i];

    if (!bucket)
        return -1;

    if (ht->eq(bucket->key, key)) {
        ht->buckets[i] = bucket->next;
        if (--ht->size < HT_MIN_DENSITY)
            ch_ht_rehash(ht, ht->cap / 2);

        return 0;
    }

    while (bucket->next) {
        if (ht->eq(bucket->next->key, key)) {
            bucket->next = bucket->next->next;
            if (--ht->size < HT_MIN_DENSITY * ht->cap)
                ch_ht_rehash(ht, ht->cap / 2);

            return 0;
        }

        bucket = bucket->next;
    }

    return -1;
}

ch_value_t ch_ht_keys(ch_ht_t *ht)
{
    ch_value_t result = ch_list_new();

    ch_hashBucket_t *bucket;
    int i;

    for (i = 0; i < ht->cap; i++) {
        bucket = ht->buckets[i];
        while (bucket) {
            CH_SEND(result, push, bucket->key);
            bucket = bucket->next;
        }
    }

    return result;
}

/* Jenkins One-at-a-time hash */
size_t ch_hashCString(const char *s, int len)
{
    size_t hash = 0;

    if (len < 0) {
        const char* c;

        for (c = s; *c; c++) {
            hash += *c;
            hash += (hash << 10);
            hash ^= (hash >> 6);
        }
    } else {
        size_t i;

        for (i = 0; i < len; i++) {
            hash += s[i];
            hash += (hash << 10);
            hash ^= (hash >> 6);
        }
    }

    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);

    return hash;
}
