/* obj.c -- chai object system
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "obj.h"
#include "sym.h"
#include "assorted.h"
#include "Bytecode.h"
#include "Bytecode.h"
#include "Exception.h"
#include "Module.h"
#include "sym_literals.h"
#include <stdio.h>
#include <string.h>

/* For dicts! */
/* Hashing */
size_t ch_hashValue(ch_value_t val)
{
    /* We don't care what type the method returns. We just look at the
     * result's integer component. */
    return (size_t)CH_SEND(val, hash).i;
}
/* equality */
int ch_hashEq(ch_value_t a, ch_value_t b)
{
    ch_value_t resultVal = CH_SEND(a, $e$e, b); // ==
    return CH_TRUEP(resultVal);
}

/* for ch_reprData.entries */
/* object repr */
static inline void obj_initData(ch_obj_t* obj) {}
static inline void obj_dup(void* to, void* from) {}

/* class repr */
static void class_initData(ch_obj_t* obj)
{
    ch_class_t* class = (ch_class_t*)obj->data;

    class->superclass = NULL;
    ch_ht_init(&class->methods, NULL, NULL);

    /* Prevent circular problems in bootstrapping */
    if (ch_cList.type != CH_NIL)
        class->subclasses = ch_list_new();
}
static void class_dup(void* dest, void* src)
{
    ch_class_t* iclass = (ch_class_t*)src;
    ch_class_t* rclass = (ch_class_t*)dest;
    rclass->superclass = iclass->superclass;
    ch_ht_mergeb(&rclass->methods, &iclass->methods);
}

/* dict repr */
static void dict_initData(ch_obj_t* obj)
{
    ch_ht_t* dict = (ch_ht_t*)obj->data;
    ch_ht_init(dict, ch_hashValue, ch_hashEq);
}
static void dict_dup(void* dest, void* src)
{
    ch_ht_t* idict = (ch_ht_t*)src;
    ch_ht_t* rdict = (ch_ht_t*)dest;
    ch_ht_mergeb(rdict, idict);
}

/* list repr */
static void list_initData(ch_obj_t* obj)
{
    ch_list_t* list = (ch_list_t*)obj->data;

    list->cap = CH_DEFAULT_LIST_CAP;
    list->size = 0;
    list->elems = ch_malloc(list->cap * sizeof(ch_value_t));
}
static void list_dup(void *dest, void *src)
{
    ch_list_t *ilist, *rlist;
    ch_value_t *new_elems;

    ilist = (ch_list_t*)src;
    rlist = (ch_list_t*)dest;
    new_elems = ch_malloc(ilist->cap * sizeof(ch_value_t));

    rlist->elems = new_elems;
    rlist->cap = ilist->cap;
    rlist->size = ilist->size;

    memcpy(rlist->elems, ilist->elems, rlist->cap * sizeof(ch_value_t));
}

/* string repr */
static void string_initData(ch_obj_t* obj)
{
    ch_string_t* str = (ch_string_t*)obj->data;
    str->cap = CH_DEFAULT_STRING_CAP;
    str->size = 0;
    str->text = ch_malloc_atomic(CH_DEFAULT_STRING_CAP);
    str->text[0] = '\0';
}
static void string_dup(void *dest, void *src)
{
    ch_string_t *istr, *rstr;
    char* new_text;

    istr = (ch_string_t*)src;
    rstr = (ch_string_t*)dest;
    new_text = ch_malloc_atomic(istr->cap);

    rstr->text = new_text;
    rstr->cap = istr->cap;
    rstr->size = istr->size;

    memcpy(rstr->text, istr->text, rstr->cap);
}

/* prim repr */
static void prim_initData(ch_obj_t* obj) { }
static void prim_dup(void* dest, void* src)
{
    ch_prim_t* iprim = (ch_prim_t*)src;
    ch_prim_t* rprim = (ch_prim_t*)dest;
    rprim->impl = iprim->impl;
    rprim->arity = iprim->arity;
}

/* Create an object! yay! */
ch_value_t ch_obj_create(ch_value_t class)
{
    ch_value_t result_val;
    unsigned int repr_id;
    size_t size;
    ch_obj_t *result;

    repr_id = ((ch_class_t*)class.obj->data)->reprID;
    size = sizeof(ch_obj_t) - 8 + ch_reprData.entries[repr_id].dataSize;
    result = ch_malloc(size);

    result->class = class.obj;
    ch_ht_init(&result->slots, NULL, NULL);

    ch_reprData.entries[repr_id].ctor(result);
    ch_set_finalizer(result, (void(*)(void*))ch_reprData.entries[repr_id].dtor);

    result_val.type = CH_OBJECT;
    result_val.obj = result;

    return result_val;
}

/* Duplicate an object */
ch_value_t ch_obj_dup(ch_value_t value)
{
    if (value.type != CH_OBJECT)
        ch_raise2(ch_eTypeError, "can't dup an immediate value");

    ch_obj_t* obj = value.obj;
    ch_value_t class = { CH_OBJECT };
    class.obj = obj->class;

    ch_value_t result = ch_obj_create(class);
    ch_ht_mergeb(&result.obj->slots, &obj->slots);

    unsigned int reprID = ((ch_class_t*)class.obj->data)->reprID;
    /* It's dup(dest, src) */
    ch_reprData.entries[reprID].dup(result.obj->data, obj->data);
    ch_set_finalizer(result.obj, (void(*)(void*))ch_reprData.entries[reprID].dtor);

    return result;
}

unsigned int ch_genRepr(int size, ch_ctor_t ctor, ch_dupFn_t dup)
{
    ch_reprDataEntry_t *entry;

    if (ch_reprData.nextID >= ch_reprData.allocd) {
        ch_reprData.allocd *= 2;
        ch_reprData.entries =
            ch_realloc(ch_reprData.entries,
                       sizeof(ch_reprDataEntry_t) * ch_reprData.allocd);
    }


    entry = &ch_reprData.entries[ch_reprData.nextID];
    entry->dataSize = size;
    entry->ctor = ctor;
    entry->dup = dup;
    entry->dtor = NULL;

    return ch_reprData.nextID++;
}

void ch_reprDtor(unsigned int id, ch_ctor_t dtor)
{
    ch_reprData.entries[id].dtor = dtor;
}

void ch_reprData_init()
{
    /* Put reprdata in a sane state. */
    ch_globalRoot(&ch_reprData.entries);
    ch_reprData.entries = ch_malloc(sizeof(ch_reprDataEntry_t) * 4);
    ch_reprData.allocd = 4;
    ch_reprData.nextID = 0;
}

void ch_obj_bootstrap()
{
    /* Representation for Class objects */
    unsigned int classRepr = ch_genRepr(sizeof(ch_class_t),
                                        class_initData,
                                        class_dup);

    /* Manually initialize class Class */
    ch_globalRoot(&ch_cClass.obj);
    ch_cClass.obj = ch_malloc(sizeof(ch_obj_t) - 8 + sizeof(ch_class_t));
    ch_cClass.type = CH_OBJECT;
    ch_ht_init(&ch_cClass.obj->slots, NULL, NULL);
    ch_cClass.obj->class = ch_cClass.obj;
    class_initData(ch_cClass.obj);
    ((ch_class_t*)ch_cClass.obj->data)->reprID = classRepr;

    /* Make a representation that doesn't do anything for objects. */
    /* Create class Object */
    ch_globalRoot(&ch_cObject.obj);
    ch_cObject = ch_obj_create(ch_cClass);
    unsigned int objectRepr = ch_genRepr(0,
                                         obj_initData,
                                         obj_dup);
    ((ch_class_t*)ch_cObject.obj->data)->reprID = objectRepr;

    ((ch_class_t*)ch_cClass.obj->data)->superclass = ch_cObject.obj;

    /* List type is needed to create subclass lists */
    ch_globalRoot(&ch_cList.obj);
    ch_cList = ch_obj_create(ch_cClass);
    unsigned int listRepr = ch_genRepr(sizeof(ch_list_t),
                                       list_initData,
                                       list_dup);
    ((ch_class_t*)ch_cList.obj->data)->reprID = listRepr;
    ((ch_class_t*)ch_cList.obj->data)->superclass = ch_cObject.obj;

    /* We need primitives to define methods for List#push */
    ch_globalRoot(&ch_cPrimitive.obj);
    ch_cPrimitive = ch_obj_create(ch_cClass);
    unsigned int primRepr = ch_genRepr(sizeof(ch_prim_t),
                                       prim_initData,
                                       prim_dup);
    ((ch_class_t*)ch_cPrimitive.obj->data)->reprID = primRepr;

    /* We need List#push for ch_list_new1 */
    ch_defineListMethods();

    /* Set the names and subclass lists of Object, Class, and List
     * appropriately, as ch_class_create() would have done. */
    ch_setSlot(ch_cObject, SYM_name, SYM_Object);
    ch_setSlot(ch_cClass, SYM_name, SYM_Class);
    ch_setSlot(ch_cList, SYM_name, SYM_List);
    ch_setSlot(ch_cPrimitive, SYM_name, SYM_Primitive);
    ((ch_class_t*)ch_cClass.obj->data)->subclasses = ch_list_new();
    ((ch_class_t*)ch_cList.obj->data)->subclasses = ch_list_new();
    ((ch_class_t*)ch_cPrimitive.obj->data)->subclasses = ch_list_new();
    ((ch_class_t*)ch_cObject.obj->data)->subclasses =
        ch_list_new1(2, ch_cClass, ch_cList);

    /* Primitive is a subclass of Function. We fix that up later. */

    /* Builtin object types. */
    ch_globalRoot(&ch_cDict.obj);
    ch_cDict = ch_class_create(SYM_Dict, ch_cObject);
    unsigned int dictRepr = ch_genRepr(sizeof(ch_ht_t),
                                       dict_initData,
                                       dict_dup);
    ((ch_class_t*)ch_cDict.obj->data)->reprID = dictRepr;

    ch_globalRoot(&ch_cString.obj);
    ch_cString = ch_class_create(SYM_String, ch_cObject);
    unsigned int stringRepr = ch_genRepr(sizeof(ch_string_t),
                                         string_initData,
                                         string_dup);
    ((ch_class_t*)ch_cString.obj->data)->reprID = stringRepr;

    ch_globalRoot(&ch_cFunction.obj);
    ch_cFunction = ch_class_create(SYM_Function, ch_cObject);

    /* Primitives are Functions */
    ((ch_class_t*)ch_cPrimitive.obj->data)->superclass = ch_cFunction.obj;
    CH_SEND(((ch_class_t*)ch_cFunction.obj->data)->subclasses,
            push, ch_cPrimitive);

    ch_Closure_init();

    /* Builtin immediate value types. */

    /* NilClass */
    size_t i;
    for(i = 0; i < 6; i++) {
        ch_globalRoot(&ch_immediateClasses[i]);
    }

    ch_globalRoot(&ch_cNilClass.obj);
    ch_cNilClass = ch_class_create(SYM_NilClass, ch_cObject);
    ch_immediateClasses[CH_NIL] = ch_cNilClass.obj;

    /* Number */
    ch_globalRoot(&ch_cNumber.obj);
    ch_cNumber = ch_class_create(SYM_Number, ch_cObject);

    /* Integer */
    ch_globalRoot(&ch_cFixnum.obj);
    ch_cFixnum = ch_class_create(SYM_Fixnum, ch_cNumber);
    ch_immediateClasses[CH_INT] = ch_cFixnum.obj;

    /* Float */
    ch_globalRoot(&ch_cFloat.obj);
    ch_cFloat = ch_class_create(SYM_Float, ch_cNumber);
    ch_immediateClasses[CH_FLOAT] = ch_cFloat.obj;

    /* Boolean */
    ch_globalRoot(&ch_cBoolean.obj);
    ch_cBoolean = ch_class_create(SYM_Boolean, ch_cObject);
    ch_immediateClasses[CH_BOOLEAN] = ch_cBoolean.obj;

    /* Symbol */
    ch_globalRoot(&ch_cSymbol.obj);
    ch_cSymbol = ch_class_create(SYM_Symbol, ch_cObject);
    ch_immediateClasses[CH_SYMBOL] = ch_cSymbol.obj;

    /* Define nice methods */
    ch_defineObjectMethods();
    ch_defineClassMethods();
    ch_defineNilClassMethods();
    ch_defineNumberMethods();
    ch_defineIntegerMethods();
    ch_defineFloatMethods();
    ch_defineDictMethods();
    /* List methods already defined because they're needed further up. */
    ch_defineStringMethods();
    ch_defineSymbolMethods();
    ch_defineBooleanMethods();

    /* definePrimitiveMethods() is just this */
    ch_defineMethod(ch_cPrimitive,
                    SYM__call_,
                    -2,
                    (ch_cfn_t)ch_vfuncall);

    ch_defineClosureMethods();

    /* I'm getting sick of creating classes in this file */
    ch_Exceptions_init();
    ch_File_init();
    ch_Range_init();
    ch_Regex_init();
    ch_Bytecode_init();
    ch_Module_init();
}

static void fix_cache(ch_class_t *klass, ch_value_t name,
                      ch_value_t new_val, int root)
{
    ch_list_t *subclasses;
    int i;

    if (!root) {
        ch_value_t meth;

        if (ch_ht_get(&klass->methods, name, &meth) < 0)
            goto recurse;

        if(meth.ref->type != CH_REF)
            return;

        *meth.ref = new_val;
    }

recurse:
    if (klass->subclasses.type == CH_NIL)
        return;

    subclasses = (ch_list_t*)klass->subclasses.obj->data;
    for (i = 0; i < subclasses->size; i++)
        fix_cache((ch_class_t*)subclasses->elems[i].obj->data,
                  name, new_val, 0);
}

static void nameMethod(ch_value_t class, ch_value_t name, ch_value_t f)
{
    if (f.type != CH_OBJECT)
        return;

    ch_setSlot(f, SYM_name, name);
    ch_setSlot(f, SYM_method_of, class);
}

void ch_defineMethodRaw(ch_value_t class, ch_value_t name, ch_value_t func)
{
    ch_value_t result;
    int must_fix_cache = 0;

    nameMethod(class, name, func);

    if (ch_ht_get(&((ch_class_t*)class.obj->data)->methods,
                  name,
                  &result) < 0) {
        result.type = CH_REF;
        result.ref = ch_malloc(sizeof(ch_value_t));

        ch_ht_set(&((ch_class_t*)class.obj->data)->methods,
                  name,
                  result);
    } else if (result.ref->type == CH_REF)
        must_fix_cache = 1;

    if (!must_fix_cache && class.obj != ch_cObject.obj) {
        /* We still might need to. Check */
        ch_class_t *klass = (ch_class_t*)class.obj->data;
        ch_value_t method;

        klass = (ch_class_t*)klass->superclass->data;
        while (klass != (ch_class_t*)((ch_obj_t*)NULL)->data) {
            if (ch_ht_get(&klass->methods, name, &method) == 0) {
                /* There's a superclass somewhere up there with this method */
                must_fix_cache = 1;
                break;
            }
            klass = (ch_class_t*)klass->superclass->data;
        }
    }

    *result.ref = func;

    if (must_fix_cache)
        fix_cache((ch_class_t*)class.obj->data, name, result, 1);
}

void ch_defineMethod(ch_value_t class,
                     ch_value_t name,
                     int arity,
                     ch_cfn_t impl)
{
    ch_value_t methodVal = ch_obj_create(ch_cPrimitive);
    ch_prim_t* prim = (ch_prim_t*)methodVal.obj->data;

    prim->impl = impl;
    prim->arity = arity >= 0 ? arity + 1 : arity;
    if (prim->arity > 4 || prim->arity < -2)
        ch_raise2(ch_eArityError, "invalid method arity %d", arity - 1);

    ch_defineMethodRaw(class, name, methodVal);
}

ch_value_t ch_lookupMethod(ch_class_t *class, ch_value_t method)
{
    ch_class_t *orig_class;
    ch_value_t ret;

    orig_class = class;
    while (ch_ht_get(&class->methods, method, &ret) < 0) {
        if(!class->superclass)
            return kNil;

        class = (ch_class_t*)class->superclass->data;
    }

    if (class != orig_class) {
        ch_value_t ref_to_ref = { CH_REF, { 0 } };

        if (ret.ref->type == CH_REF)
            ret = *ret.ref;

        ref_to_ref.ref = ch_malloc(sizeof(ch_value_t));
        *ref_to_ref.ref = ret;
        ch_ht_set(&orig_class->methods, method, ref_to_ref);
        ret = ref_to_ref;
    }

    return ret;
}

ch_value_t ch_classOf(ch_value_t obj)
{
    ch_obj_t* class = (obj.type == CH_OBJECT
                       ? obj.obj->class
                       : ch_immediateClasses[obj.type]);
    ch_value_t result = { CH_OBJECT };
    result.obj = class;
    return result;
}

ch_value_t ch_respondsTop(ch_value_t tgt, ch_value_t method)
{
    ch_class_t* class = (ch_class_t*)(ch_classOf(tgt).obj->data);
    ch_value_t f = ch_lookupMethod(class, method);

    /* in case */
    while (f.type == CH_REF)
        f = *f.ref;

    return f.type == CH_NIL ? kFalse : kTrue;
}

/* Nice handy varargs-based stuff :) */
ch_value_t ch_send(ch_value_t obj, ch_value_t method, long argc, ...)
{
#ifdef X86_VARARGS
    return ch_vsend(obj, method, argc, (ch_value_t*)(&argc + 1));
#else
    ch_value_t args[argc];

    int i;
    va_list varargs;
    va_start(varargs, argc);
    for(i = 0; i < argc; i++) {
        args[i] = va_arg(varargs, ch_value_t);
    }
    va_end(varargs);

    return ch_vsend(obj, method, argc, args);
#endif
}

ch_value_t ch_funcall(ch_value_t f, long argc, ...)
{
#ifdef X86_VARARGS
    return ch_vfuncall(f, argc, (ch_value_t*)(&argc + 1));
#else
    ch_value_t args[argc];

    int i;
    va_list varargs;
    va_start(varargs, argc);
    for(i = 0; i < argc; i++)
        args[i] = va_arg(varargs, ch_value_t);
    va_end(varargs);

    return ch_vfuncall(f, argc, args);
#endif
}

/* And the nitty-gritty */
ch_value_t ch_vsend(ch_value_t obj,
                    ch_value_t method,
                    long argc,
                    ch_value_t* argv)
{
    if (method.type != CH_SYMBOL)
        ch_raise2(ch_eTypeError, "method names must be symbols");

    ++argc; /* one extra for self */
    ch_value_t args[argc + 1]; /* cheating extra space for method name */

    ch_class_t* class = (ch_class_t*)ch_classOf(obj).obj->data;

    ch_value_t f = ch_lookupMethod(class, method);
    if (f.type == CH_NIL || f.ref->type == CH_NIL) {
        ch_value_t missing = ch_lookupMethod(class, SYM_method_missing);
        if (missing.type == CH_NIL || missing.ref->type == CH_NIL)
            return ch_default_method_missing(obj, method);

        memcpy(&args[2], argv, (argc - 1) * sizeof (ch_value_t));
        args[0] = obj;
        args[1] = method;
        if (f.ref->type == CH_REF)
            return ch_vfuncall(*f.ref->ref, argc + 1, args);
        else
            return ch_vfuncall(*f.ref, argc + 1, args);
    }

    /* Found the method */
    memcpy(&args[1], argv, (argc - 1) * sizeof(ch_value_t));
    args[0] = obj;

    if (f.ref->type == CH_REF)
        return ch_vfuncall(*f.ref->ref, argc, args);
    else
        return ch_vfuncall(*f.ref, argc, args);
}

ch_value_t ch_default_method_missing(ch_value_t tgt, ch_value_t method)
{
    ch_value_t tgtStr;
    ch_value_t inspect =
        ch_lookupMethod((ch_class_t*)ch_classOf(tgt).obj->data, SYM_inspect);

    if (inspect.type == CH_NIL || inspect.ref->type == CH_NIL)
        tgtStr = ch_string_new1("#<uninspectable?>");
    else if (inspect.ref->type == CH_REF)
        tgtStr = ch_vfuncall(*inspect.ref->ref, 1, &tgt);
    else
        tgtStr = ch_vfuncall(*inspect.ref, 1, &tgt);

    ch_value_t className;
    if(ch_getSlot(ch_classOf(tgt), SYM_name, &className) < 0
       || className.type != CH_SYMBOL)
        className = SYM_unnamed_class;

    ch_raise2(ch_eNoMethodError, "No method %s for %s:%s",
              method.sym + sizeof (long), CH_CSTR(tgtStr),
              className.sym + sizeof (long));

    return kNil; /* Humour the compiler */
}

ch_value_t ch_vfuncall(ch_value_t f, long argc, ch_value_t* argv)
{
    if (ch_cPrimitive.obj != ch_classOf(f).obj) {
        if (ch_cClosure.obj == ch_classOf(f).obj) {
            int i;

            for (i = argc - 1; i >= 0; i--)
                ch_push(argv[i]);

            ch_push(ch_fix(argc));
            ch_callClosure(f);
            return ch_pop();
        }

        return ch_vsend(f, SYM__call_, argc, argv);
    }

    ch_prim_t* prim = (ch_prim_t*)f.obj->data;
    ch_value_t result = kNil;

    ch_backtrace_push(f);

    if (prim->arity >= 0 && argc != prim->arity)
        ch_raise2(ch_eArgumentError,
                  "wrong number of arguments to procedure: %d for %d",
                  argc,
                  prim->arity);

    switch(prim->arity) {
    case 0:
        result = ((ch_value_t(*)(void))prim->impl)();
        break;
    case 1:
        result = ((ch_value_t(*)(ch_value_t))prim->impl)(argv[0]);
        break;
    case 2:
        result = ((ch_value_t(*)(ch_value_t,ch_value_t))
                  prim->impl)(argv[0], argv[1]);
        break;
    case 3:
        result = ((ch_value_t(*)(ch_value_t,ch_value_t,ch_value_t))
                  prim->impl)(argv[0], argv[1], argv[2]);
        break;
    case 4:
        result = ((ch_value_t(*)(ch_value_t,ch_value_t,ch_value_t,ch_value_t))
                  prim->impl)(argv[0], argv[1], argv[2], argv[3]);
        break;
    case -1:
        result = prim->impl(argc, argv);
        break;
    case -2:
        result = ((ch_value_t(*)(ch_value_t,long,ch_value_t*))
                  prim->impl)(argv[0], argc - 1, &argv[1]);
        break;
    default:
        ch_raise2(ch_eArityError, "Invalid primitive arity %d", prim->arity);
    }

    ch_backtrace_pop();
    return result;
}

ch_value_t ch_isa(ch_value_t obj, ch_value_t class)
{
    ch_class_t* objClass =
        (ch_class_t*)(obj.type == CH_OBJECT
                      ? obj.obj->class
                      : ch_immediateClasses[obj.type])->data;

    while(1) {
        if(objClass == (ch_class_t*)class.obj->data) {
            return kTrue;
        }
        if(!objClass->superclass) { break; }
        objClass = (ch_class_t*)objClass->superclass->data;
    }

    return kFalse;
}

ch_value_t ch_class_create(ch_value_t name, ch_value_t super)
{
    if(CH_FALSEP(ch_isa(super, ch_cClass))) { super = ch_cObject; }

    ch_value_t result = ch_obj_create(ch_cClass);
    ch_class_t* klass = (ch_class_t*)result.obj->data;

    klass->superclass = super.obj;
    klass->reprID = ((ch_class_t*)super.obj->data)->reprID;
    ch_setSlot(result, SYM_name, name);

    /* Avoid bootstrapping circularity */
    if (ch_cList.type != CH_NIL)
        CH_SEND(((ch_class_t*)klass->superclass->data)->subclasses,
                push, result);

    return result;
}

void ch_checkArgs(long argc,
                  ch_value_t* args,
                  long reqd,
                  long opt,
                  int rest,
                  ...)
{
    if (argc < reqd)
        ch_raise2(ch_eArgumentError, "not enough args: %d for %d", argc, reqd);

    if (!rest && argc > reqd + opt)
        ch_raise2(ch_eArgumentError,
                  "too many args: %d for %d", argc, reqd + opt);

    int i;
    int argn = 0;
    va_list vargs;
    va_start(vargs, rest);

    for (i = 0; i < reqd; i++) {
        ch_value_t* ptr = va_arg(vargs, ch_value_t*);
        *ptr = args[argn++];
    }
    for (i = 0; i < opt && argn < argc; i++) {
        ch_value_t* ptr = va_arg(vargs, ch_value_t*);
        *ptr = args[argn++];
    }

    if (rest) {
        ch_value_t* list = va_arg(vargs, ch_value_t*);
        *list = ch_list_new();
        while (argn < argc)
            CH_SEND(*list, push, args[argn++]);
    }

    va_end(vargs);
}

ch_value_t ch_list_new1(long size, ...)
{
    ch_value_t result = ch_list_new();
    int i;
    va_list args;
    va_start(args, size);

    for (i = 0; i < size; i++)
        CH_SEND(result, push, va_arg(args, ch_value_t));

    va_end(args);
    return result;
}

ch_value_t ch_list_new2(long size, ch_value_t* elems)
{
    ch_value_t result = ch_list_new();
    int i;

    for (i = 0; i < size; i++)
        CH_SEND(result, push, elems[i]);

    return result;
}

ch_value_t ch_string_new1(const char* text)
{
    return ch_string_new2(strlen(text), text);
}

ch_value_t ch_string_new2(size_t size, const char* text)
{
    ch_value_t result;
    ch_string_t *str;
    int new_cap;


    result = ch_string_new();
    str = (ch_string_t*)result.obj->data;
    new_cap = str->cap;

    /* size + 1 because, aw hell, let's keep a terminating NUL around. */
    while (new_cap < size + 1)
        new_cap *= 2;

    if (new_cap != str->cap) {
        str->cap = new_cap;
        str->text = ch_realloc(str->text, new_cap);
    }

    str->size = size;
    memcpy(str->text, text, str->size);
    str->text[str->size] = '\0';

    return result;
}

ch_value_t ch_prim_new(int arity, ch_cfn_t impl)
{
    if (arity > 4 || arity < -2)
        ch_raise2(ch_eArityError, "invalid arity for primitive: %d", arity);

    ch_value_t result = ch_obj_create(ch_cPrimitive);
    ch_prim_t* prim = (ch_prim_t*)result.obj->data;
    prim->arity = arity;
    prim->impl = impl;
    return result;
}
