/* obj.c -- chai object system interface
   September 24, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#ifndef CH_OBJ_H
#define CH_OBJ_H

#include "misc.h"
#include "mem.h"
#include "value.h"
#include "ht.h"
#include "state.h"
#include "Exception.h"
/*#include "sym_literals.h"*/
#include <stdarg.h>

typedef ch_value_t (*ch_cfn_t)(long, ch_value_t*);

typedef struct ch_obj {
    struct ch_obj* class;
    ch_ht_t slots;
    unsigned char data[8];
} ch_obj_t;

typedef struct {
    unsigned int reprID;
    ch_obj_t* superclass;
    ch_ht_t methods;
    ch_value_t subclasses;
} ch_class_t;

typedef struct {
    size_t cap;
    size_t size;
    ch_value_t* elems;
} ch_list_t;

typedef struct {
    size_t cap;
    size_t size;
    char* text;
} ch_string_t;

typedef struct {
    ch_cfn_t impl;
    int arity;
} ch_prim_t;

/* Some constants */
#define CH_DEFAULT_LIST_CAP 4
#define CH_DEFAULT_STRING_CAP 16

/* Initialize object internal hidden data representations. */
void ch_reprData_init(void);

/* Bootstraps the object system. */
extern void ch_obj_bootstrap(void);

/* Create an underlying representation and get its ID. */
extern unsigned int ch_genRepr(int size, ch_ctor_t ctor, ch_dupFn_t dup);
extern void ch_reprDtor(unsigned int id, ch_ctor_t dtor);

/* Only creates an instance, does not inititialize it. */
extern ch_value_t ch_obj_create(ch_value_t class);

/* Creates an instance which is a duplicate of another instance,
 * including any hidden data. Pointed-to data in hidden data are
 * reallocated as appropriate. */
extern ch_value_t ch_obj_dup(ch_value_t obj);

/* ch_obj_scan : for use from custom scan methods in types with hidden
 * data. */
extern void ch_obj_scan(void* ptr);

/* define a method
 *
 * arity works like this:
 *
 * - a "concrete" (non-negative) arity means that it's implemented by
 *   a C function taking that many ch_value_t arguments. These go up
 *   to 4. Concrete arities for methods are implicitly increased by 1
 *   in order to make room for a self parameter.
 *
 * - an arity of -1 means that the function is implemented by a C
 *   function whose signature matches:
 *       ch_value_t cFunc(long argc, ch_value_t* args)
 *
 * - an arity of -2 means that the function is implemented by a C
 *   function whose signature matches:
 *       ch_value_t cFunc(ch_value_t self, long argc, ch_value_t* args);
 *   in order to facilitate writing methods that take variadic
 *   arguments, optional arguments, or more than self + 3 arguments.
 */
extern void ch_defineMethodRaw(ch_value_t class,
                               ch_value_t name, ch_value_t func);

extern void ch_defineMethod(ch_value_t class,
                            ch_value_t name,
                            int arity,
                            ch_cfn_t impl);

/* Determine the class of an object, even if it's an immediate
 * value. */
extern ch_value_t ch_classOf(ch_value_t obj);
extern ch_value_t ch_default_method_missing(ch_value_t tgt, ch_value_t msg);

/* message sending */

/* Nice macro using crazy gcc extensions. */
#define CH_SEND(tgt, method, ...)                                       \
    ({                                                                  \
        static ch_value_t __ch__impl={ 0, { 0 } };                      \
        static ch_class_t *__ch__class=NULL;                            \
        ch_value_t __ch__args[] = {                                     \
            { 0, { 0 } }, tgt, ## __VA_ARGS__                           \
        };                                                              \
        if ((ch_class_t*)ch_classOf(__ch__args[1]).obj->data != __ch__class \
            || __ch__impl.type == CH_NIL) {                             \
            __ch__class = (ch_class_t*)ch_classOf(__ch__args[1]).obj->data; \
            __ch__impl = ch_lookupMethod(__ch__class, SYM_##method);    \
        }                                                               \
        (__ch__impl.type != CH_REF || __ch__impl.ref->type == CH_NIL)   \
        ? /* not found */                                               \
        ({                                                              \
            ch_value_t __ch__miss =                                     \
                ch_lookupMethod(__ch__class, SYM_method_missing);       \
            (__ch__miss.type != CH_REF || __ch__miss.ref->type == CH_NIL) \
                ? /* method_missing not found */                        \
                ch_default_method_missing(__ch__args[1], SYM_##method)  \
                : /* method_missing found */                            \
            ({                                                          \
            __ch__args[0] = __ch__args[1];                              \
            __ch__args[1] = SYM_##method;                               \
            __ch__miss.ref->type == CH_REF                              \
                ?                                                       \
                ch_vfuncall(*__ch__miss.ref->ref,                       \
                            sizeof (__ch__args) / sizeof (ch_value_t),  \
                            &__ch__args[0])                             \
                :                                                       \
                ch_vfuncall(*__ch__miss.ref,                            \
                            sizeof (__ch__args) / sizeof (ch_value_t),  \
                            &__ch__args[0]);                            \
            });                                                         \
        })                                                              \
            : /* method found */                                        \
            __ch__impl.ref->type == CH_REF ?                            \
            ch_vfuncall(*__ch__impl.ref->ref,                           \
                        sizeof (__ch__args) / sizeof (ch_value_t) - 1,  \
                        &__ch__args[1])                                 \
            :                                                           \
            ch_vfuncall(*__ch__impl.ref,                                \
                        sizeof (__ch__args) / sizeof (ch_value_t) - 1,  \
                        &__ch__args[1]);                                \
    })

/* Because we're too cool for the preprocessor. */
#define CH_SEND2(tgt, method, ...)                                       \
    ({                                                                  \
        static ch_value_t __ch__impl={ 0, { 0 } };                      \
        static ch_class_t *__ch__class=NULL;                            \
        ch_value_t __ch__args[] = {                                     \
            { 0, { 0 } }, tgt, ## __VA_ARGS__                           \
        };                                                              \
        if ((ch_class_t*)ch_classOf(__ch__args[1]).obj->data != __ch__class \
            || __ch__impl.type == CH_NIL) {                             \
            __ch__class = (ch_class_t*)ch_classOf(__ch__args[1]).obj->data; \
            __ch__impl = ch_lookupMethod(__ch__class, SYM_##method);    \
        }                                                               \
        (__ch__impl.type != CH_REF || __ch__impl.ref->type == CH_NIL)   \
        ? /* not found */                                               \
        ({                                                              \
            ch_value_t __ch__miss =                                     \
                ch_lookupMethod(__ch__class, SYM_method_missing);       \
            (__ch__miss.type != CH_REF || __ch__miss.ref->type == CH_NIL) \
                ? /* method_missing not found */                        \
                ch_default_method_missing(__ch__args[1], SYM_##method)  \
                : /* method_missing found */                            \
            ({                                                          \
            __ch__args[0] = __ch__args[1];                              \
            __ch__args[1] = SYM_##method;                               \
            __ch__miss.ref->type == CH_REF                              \
                ?                                                       \
                ch_vfuncall(*__ch__miss.ref->ref,                       \
                            sizeof (__ch__args) / sizeof (ch_value_t),  \
                            &__ch__args[0])                             \
                :                                                       \
                ch_vfuncall(*__ch__miss.ref,                            \
                            sizeof (__ch__args) / sizeof (ch_value_t),  \
                            &__ch__args[0]);                            \
            });                                                         \
        })                                                              \
            : /* method found */                                        \
            __ch__impl.ref->type == CH_REF ?                            \
            ch_vfuncall(*__ch__impl.ref->ref,                           \
                        sizeof (__ch__args) / sizeof (ch_value_t) - 1,  \
                        &__ch__args[1])                                 \
            :                                                           \
            ch_vfuncall(*__ch__impl.ref,                                \
                        sizeof (__ch__args) / sizeof (ch_value_t) - 1,  \
                        &__ch__args[1]);                                \
    })

extern ch_value_t ch_vsend(ch_value_t tgt,
                           ch_value_t method,
                           long argc,
                           ch_value_t* argv);
extern ch_value_t ch_send(ch_value_t tgt, ch_value_t method, long argc, ...);
extern ch_value_t ch_lookupMethod(ch_class_t *klass, ch_value_t method);
extern ch_value_t ch_respondsTop(ch_value_t tgt, ch_value_t method);

/* function calling */
extern ch_value_t ch_vfuncall(ch_value_t f, long argc, ch_value_t* args);
extern ch_value_t ch_funcall(ch_value_t f, long argc, ...);

/* Accessing slots */
static inline void ch_setSlot(ch_value_t obj, ch_value_t slot, ch_value_t value)
{
    if(obj.type != CH_OBJECT)
        ch_raise2(ch_eTypeError, "can't set slot on immutable value");

    if(slot.type != CH_SYMBOL)
        ch_raise2(ch_eTypeError, "slot names are symbols");

    ch_ht_set(&obj.obj->slots, slot, value);
}

static inline int ch_getSlot(ch_value_t obj, ch_value_t slot, ch_value_t* ret)
{
    if(obj.type != CH_OBJECT) { return -1; }

    if(slot.type != CH_SYMBOL)
        ch_raise2(ch_eTypeError, "slot names are symbols");

    return ch_ht_get(&obj.obj->slots, slot, ret);
}

/* Convenience */
extern ch_value_t ch_isa(ch_value_t obj, ch_value_t class);
extern ch_value_t ch_class_create(ch_value_t name, ch_value_t super);

/* This expects optional arguments to be already initialized. It frees
 * them if it overwrites them. */
extern void ch_checkArgs(long argc,
                         ch_value_t* args,
                         long reqd,
                         long opt,
                         int rest,
                         ...);

static inline ch_value_t ch_dict_new(void)
{
    return ch_obj_create(ch_cDict);
}
static inline ch_value_t ch_list_new(void)
{
    return ch_obj_create(ch_cList);
}
static inline ch_value_t ch_string_new(void)
{
    return ch_obj_create(ch_cString);
}
extern ch_value_t ch_list_new1(long size, ...);
extern ch_value_t ch_list_new2(long size, ch_value_t*);
extern ch_value_t ch_string_new1(const char* text);
extern ch_value_t ch_string_new2(size_t size, const char* text);
extern ch_value_t ch_prim_new(int arity, ch_cfn_t impl);

/* Test for truth */
/* This doesn't really work in ch_value.h, because of circular
 * dependencies with ch_sym_literals.h. */
static inline int CH_FALSEP(ch_value_t val)
{
    return val.type == CH_NIL || (val.type == CH_BOOLEAN && !val.i);
}
static inline int CH_TRUEP(ch_value_t val)
{
    return !CH_FALSEP(val);
}

/* Get a C string out of a string. */
static inline char* CH_CSTR(ch_value_t str)
{
    if (CH_FALSEP(ch_isa(str, ch_cString)))
        ch_raise2(ch_eTypeError, "not a string");

    ch_string_t* string = (ch_string_t*)str.obj->data;
    return string->text;
}

/* For "normal" hash tables, not the special case of symbol-indexed ones. */
size_t ch_hashValue(ch_value_t val);
int ch_hashEq(ch_value_t a, ch_value_t b);

/* Object identity test */
static inline int ch_eq(ch_value_t a, ch_value_t b)
{
    return a.type == b.type && a.obj == b.obj;
}

#endif /* CH_OBJ_H */
