/* state.c -- (hopefully most?) chai global state and global chai functions
   October 3, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "config.h"
#include "state.h"
#include "obj.h"
#include "value.h"
#include "assorted.h"
#include "Bytecode.h"
#include "Exception.h"
#include "sym.h"
#include "compile.h"
#include "Module.h"
#include "ch_signal.h"
#include "sym_literals.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>

/* Initial search path */
ch_value_t ch_CHAIPATH = { 0, { 0 } };
char *ch_main_script_dir = NULL;
ch_value_t ch_argv = { 0, { 0 } };

/* These aren't worth value.c */
ch_value_t kNil = { CH_NIL, { 0 } };
ch_value_t kFalse = { CH_BOOLEAN, { 0 } };
ch_value_t kTrue = { CH_BOOLEAN, { 1 } };

/* Those globals ought to actually exist. */
ch_value_t ch_m_builtins = { 0, { 0 } };
ch_value_t ch_m_root = { 0, { 0 } };
ch_value_t ch_cObject = { 0, { 0 } };
ch_value_t ch_cClass = { 0, { 0 } };
ch_value_t ch_cDict = { 0, { 0 } };
ch_value_t ch_cList = { 0, { 0 } };
ch_value_t ch_cString = { 0, { 0 } };
ch_value_t ch_cFunction = { 0, { 0 } };
ch_value_t ch_cPrimitive = { 0, { 0 } };
ch_value_t ch_cClosure = { 0, { 0 } };
ch_value_t ch_cNilClass = { 0, { 0 } };
ch_value_t ch_cNumber = { 0, { 0 } };
ch_value_t ch_cFixnum = { 0, { 0 } };
ch_value_t ch_cFloat = { 0, { 0 } };
ch_value_t ch_cBoolean = { 0, { 0 } };
ch_value_t ch_cSymbol = { 0, { 0 } };
ch_value_t ch_cFile = { 0, { 0 } };
ch_value_t ch_stdin = { 0, { 0 } };
ch_value_t ch_stdout = { 0, { 0 } };
ch_value_t ch_stderr = { 0, { 0 } };
ch_value_t ch_cBytecode = { 0, { 0 } };
struct ch_obj* ch_immediateClasses[6] = { 0, 0, 0, 0, 0, 0 };
ch_reprData_t ch_reprData;
ch_ht_t ch_opFuncs;
ch_value_t ch_entireAST = { 0, { 0 } };
/* End globals. */

static ch_value_t ch_puts(long argc, ch_value_t* args)
{
    return ch_vsend(ch_stdout, SYM_puts, argc, args);
}
static ch_value_t ch_print(long argc, ch_value_t* args)
{
    return ch_vsend(ch_stdout, SYM_print, argc, args);
}
static ch_value_t ch_gets()
{
    return CH_SEND(ch_stdin, gets);
}

static ch_value_t ch_public_gensym(long argc, ch_value_t *args)
{
    ch_value_t argstr;
    char *str;

    argstr = kNil;
    str = NULL;
    ch_checkArgs(argc, args, 0, 1, 0, &argstr);

    if (argstr.type == CH_NIL)
        str = "";
    else if (CH_FALSEP(ch_isa(argstr, ch_cString)))
        ch_raise2(ch_eTypeError, "gensym() expects a string argument");
    else
        str = ((ch_string_t*)argstr.obj->data)->text;

    return ch_gensym(str);
}

static ch_value_t ch_public_eval(ch_value_t s)
{
    if (CH_FALSEP(ch_isa(s, ch_cString)))
        ch_raise2(ch_eTypeError, "eval() expects a string");

    return ch_evalString(CH_CSTR(s));
}

static ch_value_t ch_public_eval_file(ch_value_t s)
{
    if (CH_FALSEP(ch_isa(s, ch_cString)))
        ch_raise2(ch_eTypeError, "eval_file() expects a string");

    return ch_evalFile(CH_CSTR(s));
}

static ch_value_t ch_public_load(ch_value_t s)
{
    if (CH_FALSEP(ch_isa(s, ch_cString)))
        ch_raise2(ch_eTypeError, "eval_file() expects a string");

    return ch_load(CH_CSTR(s));
}

static void setup_main_script_dir(const char *program_name)
{
    char program_dir[8192];
    int i;

    for (i = strlen(program_name); i >= 0 && program_name[i] != '/'; i--);
    snprintf(program_dir, sizeof (program_dir), "%.*s", i+1, program_name);

    if (program_dir[0] != '/') {
        char buf[8192], cwd[8192];

        if(!getcwd(cwd, sizeof (cwd)))
            fatal("getcwd: %s", strerror(errno));

        snprintf(buf, sizeof (buf), "%s/%s", cwd, program_dir);
        ch_main_script_dir = ch_strdup(buf);
    } else {
        ch_main_script_dir = ch_strdup(program_dir);
    }
}

static void append_path_dirs(const char *path)
{
    const char *s;

    while ((s = strstr(path, ":")) != NULL) {
        if (s != path)
            CH_SEND(ch_CHAIPATH, push, ch_string_new2(s - path, path));

        path = &s[1];
    }

    if (*path)
        CH_SEND(ch_CHAIPATH, push, ch_string_new1(path));
}

static void setup_search_path(void)
{
    char *env_CHAIPATH;

    ch_CHAIPATH = ch_list_new();

    env_CHAIPATH = getenv("CHAIPATH");
    if (env_CHAIPATH && strcmp(env_CHAIPATH, ""))
        append_path_dirs(env_CHAIPATH);

    /* Don't let $CHAIPATH remove the default dirs. */
    append_path_dirs(DEFAULT_CHAIPATH);
}

static ch_value_t backtrace()
{
    /* subtracting one to skip this function's frame */
    return ch_list_new2(ch_backtrace_top - 1, ch_backtrace);
}

static ch_value_t ch_srand(long argc, ch_value_t *argv)
{
    ch_value_t seed = kNil;
    ch_checkArgs(argc, argv, 0, 1, 0, &seed);

    if (seed.type == CH_NIL)
        seed = ch_fix(time(NULL));

    if (seed.type != CH_INT)
        ch_raise2(ch_eTypeError, "srand() expects an integer");

    srand(seed.i);
    return kNil;
}

static ch_value_t ch_rand(long argc, ch_value_t *argv)
{
    ch_value_t lim = kNil;
    float frand;
    ch_checkArgs(argc, argv, 0, 1, 0, &lim);

    frand = (float)rand() / (float)RAND_MAX;

    if (lim.type == CH_NIL)
        return ch_float(frand);
    else if (lim.type == CH_INT)
        return ch_fix((int)(frand * lim.i));
    else if (lim.type == CH_FLOAT)
        return ch_float(frand * lim.f);
    else {
        ch_raise2(ch_eTypeError, "rand() expects an integer or float");
        return kNil; /* appease compiler */
    }
}

static void init_m_builtins_parser(void)
{
    ch_module_t *m = (ch_module_t*)ch_m_builtins.obj->data;

    ch_ht_set(&m->op_idents, SYM_and, kTrue);
    ch_ht_set(&m->op_idents, SYM_or, kTrue);
    ch_ht_set(&m->op_idents, SYM_not, kTrue);
    ch_ht_set(&m->op_idents, SYM_my, kTrue);

    ch_ht_set(&m->prec_table, SYM_or, ch_fix(100));
    ch_ht_set(&m->prec_table, SYM_and, ch_fix(200));
    ch_ht_set(&m->prec_table, SYM_$e, ch_fix(400));
    ch_ht_set(&m->prec_table, SYM_$e$g, ch_fix(400));
    ch_ht_set(&m->prec_table, SYM_$o$o, ch_fix(500));
    ch_ht_set(&m->prec_table, SYM_$a$a, ch_fix(600));
    ch_ht_set(&m->prec_table, SYM_$o, ch_fix(700));
    ch_ht_set(&m->prec_table, SYM_$x, ch_fix(800));
    ch_ht_set(&m->prec_table, SYM_$a, ch_fix(900));
    ch_ht_set(&m->prec_table, SYM_$e$e, ch_fix(1000));
    ch_ht_set(&m->prec_table, SYM_$b$e, ch_fix(1000));
    ch_ht_set(&m->prec_table, SYM_$l, ch_fix(1100));
    ch_ht_set(&m->prec_table, SYM_$g, ch_fix(1100));
    ch_ht_set(&m->prec_table, SYM_$l$e, ch_fix(1100));
    ch_ht_set(&m->prec_table, SYM_$g$e, ch_fix(1100));
    ch_ht_set(&m->prec_table, SYM_$D$D, ch_fix(1200));
    ch_ht_set(&m->prec_table, SYM_$l$l, ch_fix(1300));
    ch_ht_set(&m->prec_table, SYM_$g$g, ch_fix(1300));
    ch_ht_set(&m->prec_table, SYM_$p, ch_fix(1400));
    ch_ht_set(&m->prec_table, SYM_$m, ch_fix(1400));
    ch_ht_set(&m->prec_table, SYM_$t, ch_fix(1500));
    ch_ht_set(&m->prec_table, SYM_$d, ch_fix(1500));
    ch_ht_set(&m->prec_table, SYM_$d$d, ch_fix(1500));
    ch_ht_set(&m->prec_table, SYM_$r, ch_fix(1500));
    ch_ht_set(&m->prec_table, SYM_$t$t, ch_fix(1700));
    ch_ht_set(&m->prec_table, SYM_$D, ch_fix(3000));
    ch_ht_set(&m->prec_table, SYM___call_op__, ch_fix(3000));
    ch_ht_set(&m->prec_table, SYM_$c, ch_fix(3000));

    ch_ht_set(&m->uprec_table, SYM_not, ch_fix(300));
    ch_ht_set(&m->uprec_table, SYM_my, ch_fix(350));
    ch_ht_set(&m->uprec_table, SYM_$t, ch_fix(1350));
    ch_ht_set(&m->uprec_table, SYM_$m, ch_fix(1600));
    ch_ht_set(&m->uprec_table, SYM_$p, ch_fix(1600));
    ch_ht_set(&m->uprec_table, SYM_$b, ch_fix(1600));
    ch_ht_set(&m->uprec_table, SYM_$n, ch_fix(1600));
    /* higher than the binary op because associativity is tied to precedence */
    ch_ht_set(&m->uprec_table, SYM_$D, ch_fix(3100));
    ch_ht_set(&m->uprec_table, SYM_$A, ch_fix(3200));

    ch_ht_set(&m->assoc_table, ch_fix(100), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(200), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(300), SYM_unary);
    ch_ht_set(&m->assoc_table, ch_fix(350), SYM_unary);
    ch_ht_set(&m->assoc_table, ch_fix(400), SYM_right);
    ch_ht_set(&m->assoc_table, ch_fix(500), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(600), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(700), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(800), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(900), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1000), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1100), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1200), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1300), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1350), SYM_unary);
    ch_ht_set(&m->assoc_table, ch_fix(1400), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1500), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(1600), SYM_unary);
    ch_ht_set(&m->assoc_table, ch_fix(1700), SYM_right);
    ch_ht_set(&m->assoc_table, ch_fix(3000), SYM_left);
    ch_ht_set(&m->assoc_table, ch_fix(3100), SYM_unary);
    ch_ht_set(&m->assoc_table, ch_fix(3200), SYM_unary);
}

static ch_value_t prec_of(ch_value_t name, ch_value_t op, ch_value_t u)
{
    ch_value_t m = ch_get_module(name);
    if (m.type == CH_NIL)
        ch_raise2(ch_eNameError, "no module %s", name.sym + sizeof (long));

    return ch_module_precedence_of(m, op, u);
}

void ch_state_init(const char *program_name)
{
    /* Set up stacks */
    ch_stack_size = 64;
    ch_stack_top = 0;
    ch_stack = ch_malloc(sizeof (ch_value_t) * ch_stack_size);
    ch_globalRoot(&ch_stack);

    ch_backtrace_size = 1024;
    ch_backtrace_top = 0;
    ch_backtrace = ch_malloc(2 * sizeof (ch_value_t) * ch_backtrace_size);
    ch_globalRoot(&ch_backtrace);

    ch_handlers_init();

    /* Bootstrap the object system. */
    ch_obj_bootstrap();

    /* Setup the global environment. */
    ch_m_builtins = ch_obj_create(ch_cModule);
    ch_m_root = ch_obj_create(ch_cModule);
    ((ch_module_t*)ch_m_builtins.obj->data)->module_name = SYM__builtins;
    ((ch_module_t*)ch_m_root.obj->data)->module_name = SYM__r;

    ch_module_import(ch_m_builtins, ch_m_root, SYM__r);
    ch_module_import(ch_m_builtins, ch_m_builtins, SYM__builtins);
    ch_module_import(ch_m_root, ch_m_builtins, SYM__builtins);

    ch_current_module = ch_m_builtins;

    ch_globalRoot(&ch_m_builtins.obj);
    ch_globalRoot(&ch_m_root.obj);
    init_m_builtins_parser();

    ch_defn(puts, -1, ch_puts);
    ch_defn(print, -1, ch_print);
    ch_defn(gets, 0, ch_gets);
    ch_defn(rand, -1, ch_rand);
    ch_defn(srand, -1, ch_srand);

    /* Macros and eval */
    ch_defn(eval, 1, ch_public_eval);
    ch_defn(eval_file, 1, ch_public_eval_file);
    ch_defn(load, 1, ch_public_load);
    ch_defn(gensym, -1, ch_public_gensym);
    ch_defn(macroexpand1, 1, ch_macroexpand1);
    ch_defn(macroexpand, 1, ch_macroexpand);
    ch_defn(defcalled$q, 1, ch_defcalledp);
    ch_defn(defsetfed$q, 1, ch_defsetfedp);
    ch_defn(my_bindings_for, 1, ch_my_bindings_for);
    ch_defn(precedence_of, 3, prec_of);

    /* Builtin classes */
    ch_set_global(ch_m_builtins, SYM_Object, ch_cObject);
    ch_set_global(ch_m_builtins, SYM_Class, ch_cClass);
    ch_set_global(ch_m_builtins, SYM_Dict, ch_cDict);
    ch_set_global(ch_m_builtins, SYM_List, ch_cList);
    ch_set_global(ch_m_builtins, SYM_String, ch_cString);

    ch_set_global(ch_m_builtins, SYM_Function, ch_cFunction);
    ch_set_global(ch_m_builtins, SYM_Primitive, ch_cPrimitive);
    ch_set_global(ch_m_builtins, SYM_Closure, ch_cClosure);

    ch_set_global(ch_m_builtins, SYM_NilClass, ch_cNilClass);
    ch_set_global(ch_m_builtins, SYM_Number, ch_cNumber);
    ch_set_global(ch_m_builtins, SYM_Fixnum, ch_cFixnum);
    ch_set_global(ch_m_builtins, SYM_Float, ch_cFloat);
    ch_set_global(ch_m_builtins, SYM_Boolean, ch_cBoolean);
    ch_set_global(ch_m_builtins, SYM_Symbol, ch_cSymbol);

    ch_set_global(ch_m_builtins, SYM_Module, ch_cModule);

    ch_set_global(ch_m_builtins, SYM_Range, ch_cRange);
    ch_set_global(ch_m_builtins, SYM_Regex, ch_cRegex);
    ch_set_global(ch_m_builtins, SYM_RegexMatchData, ch_cRegexMatchData);
    ch_set_global(ch_m_builtins, SYM_RegexCompilationError,
               ch_eRegexCompilationError);

    /* File class and standard streams */
    ch_set_global(ch_m_builtins, SYM_File, ch_cFile);
    ch_defvar(stdout, ch_stdout);
    ch_defvar(stderr, ch_stderr);
    ch_defvar(stdin, ch_stdin);

    /* Exceptions */
    ch_defn(backtrace, 0, backtrace);

    ch_set_global(ch_m_builtins, SYM_Exception, ch_eException);
    ch_set_global(ch_m_builtins, SYM_TypeError, ch_eTypeError);
    ch_set_global(ch_m_builtins, SYM_NameError, ch_eNameError);
    ch_set_global(ch_m_builtins, SYM_ArgumentError, ch_eArgumentError);
    ch_set_global(ch_m_builtins, SYM_NoMethodError, ch_eNoMethodError);
    ch_set_global(ch_m_builtins, SYM_UnimplementedError, ch_eUnimplementedError);
    ch_set_global(ch_m_builtins, SYM_BytecodeFormatError, ch_eBytecodeFormatError);
    ch_set_global(ch_m_builtins, SYM_FileModeError, ch_eFileModeError);
    ch_set_global(ch_m_builtins, SYM_OutOfBoundsError, ch_eOutOfBoundsError);
    ch_set_global(ch_m_builtins, SYM_ArityError, ch_eArityError);
    ch_set_global(ch_m_builtins, SYM_ParseError, ch_eParseError);
    ch_set_global(ch_m_builtins, SYM_StackDepthError, ch_eStackDepthError);
    ch_set_global(ch_m_builtins, SYM_FileOpenError, ch_eFileOpenError);
    ch_set_global(ch_m_builtins, SYM_DivideByZeroError, ch_eDivideByZeroError);

    /* Setup search path */
    setup_main_script_dir(program_name);
    setup_search_path();
}
