/* sys.c -- implementation of the chai sys module
   October 3, 2008

  Copyright (C) 2007-2008 Devin Papineau

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Devin Papineau jdmpapin@uwaterloo.ca
*/

#include "chai.h"
#include <stdlib.h>

/* for Sleep() */
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#else
#include <unistd.h>
#endif

/* functions */
static ch_value_t ch_exit(long argc, ch_value_t* args)
{
    ch_value_t retcode = kNil;
    ch_checkArgs(argc, args, 0, 1, 0, &retcode);

    exit(retcode.type == CH_INT ? retcode.i : 0);
    return kNil; /* Make compiler happy? */
}

static ch_value_t ch_sleep(ch_value_t duration)
{
    if (duration.type != CH_INT && duration.type != CH_FLOAT)
        ch_raise2(ch_eTypeError,
                  "sleep() expects an integer or floating-point duration");

    float seconds = duration.type == CH_INT ? (float)duration.i : duration.f;

#ifdef WIN32
    /* use Sleep() */
    int millis = (int)(seconds * 1000.0);
    Sleep(millis);
#else
    /* Use sleep() and usleep() */
    int wholeSecs = (int)seconds;
    int uSecs = (int)((seconds - wholeSecs) * 1000000.0);
    sleep(wholeSecs);
    usleep(uSecs);
#endif

    return kNil;
}

static ch_value_t ch_public_signal(long argc, ch_value_t *argv)
{
    ch_value_t signal, handler = kNil;

    ch_checkArgs(argc, argv, 1, 1, 0, &signal, &handler);

    if (signal.type == CH_SYMBOL)
        signal = ch_fix(ch_signal_from_name(signal.sym + sizeof (long)));
    else if (CH_TRUEP(ch_isa(signal, ch_cString)))
        signal = ch_fix(ch_signal_from_name(CH_CSTR(signal)));

    if (signal.type != CH_INT)
        ch_raise2(ch_eTypeError,
                  "expect a symbol, string, or int to identify the signal");

    return ch_signal_set_handler(signal.i, handler);
}

/* symbols */
ch_value_t SYM_name, SYM_containing_module, SYM_exit,
    SYM_sleep, SYM_signal, SYM_argv, SYM_CHAIPATH;

/* initialize the module */
void chai_load_module(void)
{
    SYM_name = ch_internCString("name");
    SYM_containing_module = ch_internCString("containing_module");
    SYM_exit = ch_internCString("exit");
    SYM_sleep = ch_internCString("sleep");
    SYM_signal = ch_internCString("signal");
    SYM_argv = ch_internCString("argv");
    SYM_CHAIPATH = ch_internCString("CHAIPATH");

    ch_defn(exit, -1, ch_exit);
    ch_defn(sleep, 1, ch_sleep);
    ch_defn(signal, -1, ch_public_signal);

    ch_defvar(argv, ch_argv);
    ch_defvar(CHAIPATH, ch_CHAIPATH);
}
